<?php

// Namespace declaration
namespace Raptus\WC\BexioConnector;

// Exit if accessed directly
defined('ABSPATH') or die();

// Dependencies
use WC_Tax;
use WC_Countries;

/**
 * The admin module of this plugin.
 */
class Admin {

    /**
     * Use Singleton trait to disallow multiple instances of this class.
     */
    use Singleton;

    /**
     * The prefix for the option to store the hash of the last confirmed notification.
     * @var String
     * */
    private $notification_option_prefix = 'wcba_last_notification__';

    /**
     * The prefix for the option to store the timestamp of the last notification check.
     * @var String
     * */
    private $notification_option_timestamp = 'wcba_last_notification_request';

    /**
     * The interval in which to check for notifications.
     * @var int
     * */
    private $notification_request_interval = 60 * 5; // 5 Minutes

    /**
     * Constructs a new instance of the Admin class.
     * Registers all required filter and action hooks.
     */
    protected function __construct() {
        // Custom CSS
        add_action('admin_head', array($this, 'enqueue_admin_css'));

        add_action('admin_notices', array($this, 'general_notices'), 9999);

        if (Options::get('isauthorized', false)) {
            // Trigger admin actions if requested
            add_action('admin_init', array($this, 'admin_actions'));

            // Custom JS
            add_action('admin_enqueue_scripts', array($this, 'enqueue_admin_js'));

            // Display status messages
            add_action('admin_notices', array($this, 'check_connection_to_api'), 9998);
            add_action('admin_notices', array($this, 'check_sync'), 9999);

            // Render a new meta section in the admin order detail view
            add_action('woocommerce_admin_order_data_after_order_details', array($this, 'admin_order_meta'), 10, 1);

            // Render custom column in shop order admin table
            add_filter('manage_edit-shop_order_columns', array($this, 'show_shop_order_columns'), 999);
            add_action('manage_shop_order_posts_custom_column', array($this, 'render_shop_order_columns'), 10, 2);

            if (Stock::instance()->is_stock_sync_enabled()) {
                // Render custom column in shop product admin table
                add_filter('manage_edit-product_columns', array($this, 'show_product_columns'), 999);
                add_action('manage_product_posts_custom_column', array($this, 'render_product_columns'), 10, 2);

                add_filter('bulk_actions-edit-product', array($this, 'add_product_bulk_actions'), 10, 1);
                add_filter('handle_bulk_actions-edit-product', array($this, 'handle_product_bulk_actions'), 10, 3);
                add_action('admin_notices', array($this, 'check_product_stock_sync'), 10);
            }
        }

        // Add custom plugin links in plugin table
        add_filter('plugin_action_links_' . PLUGIN_FILE, array($this, 'add_action_links'), 10, 1);
    }

    public function add_product_bulk_actions($actions) {
        $actions['product_stock_bexio_sync'] = __('Synchronize stock with Bexio', 'raptus-wc-bexio-connector');
        return $actions;
    }

    public function handle_product_bulk_actions($redirect_to, $doaction, $post_ids) {
        if ($doaction == 'product_stock_bexio_sync' ) {
            $result = Stock::instance()->sync_all($post_ids, true);
            $prefix = 'product_stock_bexio_sync';
            $redirect_to = add_query_arg(array(
                "{$prefix}_complete" => count($result['updated']),
                "{$prefix}_incomplete" => count($result['not_updated']),
                "{$prefix}_skipped" => count($result['skipped']),
            ), $redirect_to);
        }
        return $redirect_to;
    }

    /**
     * Checks the Bexio synchronization status of the current order after a synchronization has been performed.
     *
     * - Prints a success message if the status is now synchronized.
     * - Prints a warning message if the status is only partially synchronized.
     * - Prints an error message if the status is still unsynchronized.
     */
    public function check_product_stock_sync($type = '', $msg = '') {
        if (isset($_GET['product_stock_bexio_sync_complete'])) {
            if ($_GET['product_stock_bexio_sync_complete'] > 0) {
                $type = 'success';
                $msg = sprintf(__('The stock of %d products was successfully synchronized with Bexio.', 'raptus-wc-bexio-connector'), $_GET['product_stock_bexio_sync_complete']);
                if (isset($_GET['product_stock_bexio_sync_incomplete']) && $_GET['product_stock_bexio_sync_incomplete'] > 0) {
                    $type = 'warning';
                    $msg .= '<br/>' . sprintf(__('However, the stock synchronization was incomplete, %d products could not be synchronized.', 'raptus-wc-bexio-connector'), $_GET['product_stock_bexio_sync_incomplete']);
                }
            } else {
                $type = 'error';
                $msg = __('The stock synchronization failed, the stock of no prouducts was updated.', 'raptus-wc-bexio-connector');
            }
            if (isset($_GET['product_stock_bexio_sync_skipped'])) {
                $skipped = $_GET['product_stock_bexio_sync_skipped'];
                $msg .= '<br/><em>' . sprintf(_n('%d product without stock was skipped.', '%d products without stock were skipped.', $skipped, 'raptus-wc-bexio-connector'), $skipped) . '</em>';
            }
            Utils::print_notice($type, $msg);
        }
    }

    /**
     * Add custom plugin action links.
     *
     * @param array $links The current plugin action links.
     * @return array The modified action links array, with additional actions added.
     */
    public function add_action_links($links) {
        array_unshift($links, '<a href="' . FAQ_URL . '" target="_blank">' . __('Find help', 'raptus-wc-bexio-connector') . '</a>');
        array_unshift($links, '<a href="' . Settings::instance()->get_settings_page_url() . '">' . __('Settings') . '</a>');
        return $links;
    }

    /**
     * Add a new column header for displaying the Bexio order number.
     *
     * @param  array $columns The current columns.
     * @return array The modified columns array, with additional column added.
     */
    public function show_shop_order_columns($columns) {
       $columns['bexio_order_nr'] = __('Bexio Details', 'raptus-wc-bexio-connector');
       return $columns;
    }

    /**
     * Render the Bexio order number if the current column the Bexio order number column.
     *
     * @param string $column The identifier of the current colum.
     * @param int $post_id The id of the current post.
     */
    public function render_shop_order_columns($column, $post_id) {
        if ($column == 'bexio_order_nr') {
            $links = array();
            foreach (Data::get_sync_types() as $type => $label) {
                $links[] = $this->print_bexio_edit_link($post_id, $type, false);
            }
            echo '<p style="line-height: 1.5em;">';
            echo join('<br/>', array_filter($links));
            echo '</div>';
        }
    }

    /**
     * Display the direct link to the Bexio order if available.
     *
     * @param int $order_id The id of the current WC order.
     */
    private function print_bexio_edit_link($order_id, $type = 'order', $echo = true, $default = false) {
        $html = '';
        if (metadata_exists('post', $order_id, '_bexio_sync_processing')) {
            $html .= __('Processing', 'raptus-wc-bexio-connector');
            $html .= ' <span class="dashicons dashicons-admin-generic rotating"></span>';
        } else {
            $bexio_id = get_post_meta($order_id, '_bexio_' . $type . '_id', true);
            if ($bexio_id) {
                $html .= '<a href="' . BEXIO_BASE_URL . 'kb_' . $type . '/show/id/' . $bexio_id . '" target="_blank">';
            }
            $document_nr = get_post_meta($order_id, '_bexio_' . $type . '_nr', true);
            if ($document_nr) {
                $html .= $document_nr;
            } else if ($bexio_id) {
                $html .= $default ?: Data::get_sync_types($type);
            }
            if ($bexio_id) {
                $html .= '</a>';
            }
            $html = apply_filters('raptus_wp_bexio_connector_edit_link_' . $type, $html, $order_id);
        }
        if (!$echo) {
            return $html;
        }
        echo $html;
    }

    /**
     * Add a new column header for displaying the last stock synchronization time.
     *
     * @param  array $columns The current columns.
     * @return array The modified columns array, with additional column added.
     */
    public function show_product_columns($columns) {
       $columns['stock_status'] = __('Last synchronized', 'raptus-wc-bexio-connector');
       return $columns;
    }

    /**
     * Render the last stock synchronization time if the current column is the stock status column.
     *
     * @param string $column The identifier of the current colum.
     * @param int $post_id The id of the current post.
     */
    public function render_product_columns($column, $post_id) {
        if ($column == 'stock_status') {
            $cache_time = Stock::instance()->get_cache_time_by_id($post_id);
            $format = get_option('date_format') . ' ' . get_option('time_format');
            $status = $cache_time ? date_i18n($format, $cache_time) : '-';
            echo apply_filters('raptus-wc-bexio-connector-stock-status', $status, $post_id);
        }
    }

    /**
     * Replace all option key placeholders with their respective option value.
     *
     * @param string $message The message to repalce.
     * @return string The modified message.
     */
    protected function replace_option_placeholders($message) {
        return preg_replace_callback('/{option:(.*?)}/', function($matches) {
            return get_option($matches[1], true);
        }, $message);
    }

    /**
     * Show general notices in the backend (not only in plugin settings).
     * */
    public function general_notices() {
        $show_bexio_authorization_notice = get_option('raptus_wp_bexio_connector_needs_reauthorization');
        if ($show_bexio_authorization_notice) {
            $msg = sprintf(__('Migration to the new plugin version of WooCommerce Bexio Automator is complete. You will need to re-authorize with Bexio to continue synchronizing. <a href="%s">Go to settings</a>.', 'raptus-wc-bexio-connector'),
                Settings::instance()->get_settings_page_url()
            );
            Utils::print_notice('info', $msg, true, 'p', true);
        }

        if (isset($_GET['wcba_notification_confirmed']) && isset($_GET['wcba_lang'])) {
            $msg_hash = $_GET['wcba_notification_confirmed'];
            $lang = $_GET['wcba_lang'];
            update_option($this->notification_option_prefix . $lang, $msg_hash, false);
            $redirect_url = remove_query_arg([
                'wcba_notification_confirmed',
                'wcba_lang'
            ]);
            wp_redirect($redirect_url);
        } else {
            $this->check_for_public_notifications();
        }
    }

    /**
     * Check if there are any important public notifications.
     * If there are, display the notification as an admin notice, until the user has confirmed it.
     * Confirmed notifications will not be shown again.
     *
     * If the response code is 404 (no file was found), retry with fallback language.
     *
     * @param string $lang The language to search notifications for (defaults to current user locale).
     * @param boolean $retry Whether to retry in fallback language if no file was found (response code 404).
     * @return boolean Returns true if a notification was found, false otherwise.
     * */
    private function check_for_public_notifications($lang = null, $retry = true) {
        // Check if the last request was within the request-interval
        $last_requested = get_option($this->notification_option_timestamp, false);
        if ($last_requested) {
            if (time() - $last_requested <= $this->notification_request_interval) {
                // If the last request took place within the interval, skip the notification check
                return false;
            }
        }

        $lang = $lang ?: get_user_locale();
        $lang = substr($lang, 0, 2); // shorten to 2-digit language code (e.g. 'en_EN' to 'en')
        update_option($this->notification_option_timestamp, time(), false);
        $response = wp_remote_request(PUBLIC_NOTIFICATION_URL . $lang . '.html', [
            'method' => 'GET',
            'timeout' => 45
        ]);
        $response_code = wp_remote_retrieve_response_code($response);
        if ($response_code == 404 && $retry) {
            // No notification found in given language, retry with fallback 'en' (no further retries)
            return $this->check_for_public_notifications('en', false);
        }

        if ($response_code == 200) {
            // A notification was found, retrieve message
            $msg = wp_remote_retrieve_body($response);
            if ($msg === '') {
                // Allow an empty message to not show anything (saves the need for fallback request vs. if there were no file at all)
                return false;
            }
            $msg_hash = md5($msg);

            // Check if notification was already confirmed
            $has_option = get_option($this->notification_option_prefix . $lang, false);
            if ($has_option && $has_option == $msg_hash) {
                // Same notification was already shown to and confirmed by the user, do not show again
                return false;
            }

            $msg = $this->replace_option_placeholders($msg);
            $msg .= sprintf('<br/><a href="%s" class="button button-secondary">%s</a>',
                add_query_arg([
                    'wcba_notification_confirmed' => $msg_hash,
                    'wcba_lang' => $lang
                ]),
                _x('Acknowledge', 'Confirmation button for public plugin notifications', 'raptus-wc-bexio-connector')
            );
            Utils::print_notice('info', $msg, false, 'p', true);
            return true;
        }

        return false;
    }

    /**
     * Check for the constants WP_HTTP_BLOCK_EXTERNAL and WC_ACCESSIBLE_HOSTS and show error notice if WP_HTTP_BLOCK_EXTERNAL ist true and
     * WP_ACCESSIBLE_HOSTS is set to a value different than bexio-shop.ch API. See config.php where WP_ACCESSIBLE_HOSTS is defined
     *
     */
    public function check_connection_to_api() {
        if ( defined('WP_HTTP_BLOCK_EXTERNAL') && WP_HTTP_BLOCK_EXTERNAL === true && defined('WP_ACCESSIBLE_HOSTS') && !strpos(WP_ACCESSIBLE_HOSTS, 'api.bexio-shop.ch')) {
            $type = 'error';
            $msg  = sprintf(__('Your WordPress configuration blocks http requests to the Bexio Automator API through WP_HTTP_BLOCK_EXTERNAL and WP_ACCESSIBLE_HOSTS (%s)', 'raptus-wc-bexio-connector'), WP_ACCESSIBLE_HOSTS);
            Utils::print_notice($type, $msg);
        }
    }


    /**
     * Checks the Bexio synchronization status of the current order after a synchronization has been performed.
     * Prints a success message if the status is now synchronized.
     * Prints an error message if the status is still unsynchronized.
     * Prints a warning message if the ordre was only partially synchronized.
     */
    public function check_sync() {
        if (isset($_GET['order_bexio_sync']) && isset($_GET['post'])) {
            $order_id = $_GET['post'];
            $order = wc_get_order($order_id);
            $sync_type = $_GET['order_bexio_sync'];
            if (!$this->check_processing($order_id)) {
                $message_type = '';
                $messages = ['success' => [], 'warning' => [], 'error' => []];
                foreach (Data::get_sync_types() as $type => $name) {
                    if ($sync_type != 1 && $sync_type != $type) {
                        continue;
                    }
                    if (SyncRules::can_sync($type, $order) || $this->is_force_sync()) {
                        if (metadata_exists('post', $order_id, "_bexio_{$type}_id")) {
                            $messages['success'][$type] = $name;
                        } else if ($type == 'payment' && metadata_exists('post', $order_id, '_bexio_invoice_paid')) {
                            $messages['success'][$type] = $name;
                        } else {
                            $messages['error'][$type] = $name;
                        }
                    }
                    if ($msg = $this->get_sync_status_message($order_id, $type)) {
                        $messages['warning'][$type] = $msg;
                    }
                }
                $message = [];
                foreach ($messages as $_message_type => $_messages) {
                    if (!empty($_messages)) {
                        $message_type = $_message_type;
                        if ($_message_type == 'success') {
                            $message[] = __('The following items were successfully synchronized with Bexio:', 'raptus-wc-bexio-connector');
                        } else if ($_message_type == 'error') {
                            $message[] = __('The following items could not be synchronized with Bexio:', 'raptus-wc-bexio-connector');
                        }
                        if ($_message_type == 'success' || $_message_type == 'error') {
                            $message[] = '<ul style="list-style: square; padding-left: 20px; margin: 5px 0;"><li style="margin-bottom: 0;">' . implode('</li><li style="margin-bottom: 0;">', $_messages) . '</li></ul>';
                        } else {
                            $message = array_merge($message, array_values($_messages));
                        }
                    }
                }
                if (!empty(Api::instance()->get_error())) {
                    $message_type = 'error';
                    $message[] = '<code style="margin-top: 10px; display: inline-block;">'
                        . __('API error ', 'raptus-wc-bexio-connector') . Api::instance()->get_error_message() . '</code>';
                }
                if (!empty($message)) {
                    $message_type = 'warning';
                    if (!empty($messages['success']) && empty($messages['warning']) && empty($messages['error'])) {
                        $message_type = 'success';
                    } else if (!empty($messages['error']) && empty($messages['warning']) && empty($messages['success'])) {
                        $message_type = 'error';
                    }
                    Utils::print_notice($message_type, $message, true, 'div');
                } else {
                    Utils::print_notice('warning', [
                        __('No data was synchronized with Bexio', 'raptus-wc-bexio-connector'),
                        sprintf(__('Check the <a href="%s">synchronization settings</a> if this is unexpected or trigger a synchronization forcefully using the actions below.', 'raptus-wc-bexio-connector'), Settings::instance()->get_settings_page_url()),
                    ]);
                }
            }
        } else if (isset($_GET['order_bexio_unlink']) && $_GET['order_bexio_unlink'] == 'done' && isset($_GET['post'])) {
            $order_id = $_GET['post'];
            $result = true;
            foreach (Data::get_sync_types() as $type => $label) {
                foreach (['id', 'nr'] as $item) {
                    if (get_post_meta($order_id, "_bexio_{$type}_{$item}")) {
                        $result = false;
                    }
                }
            }
            if ($result) {
                Utils::print_notice('success', __('All bexio documents have been unlinked.', 'raptus-wc-bexio-connector'));
            } else {
                Utils::print_notice('warning', __('The bexio documents could not be unlinked, please try again.', 'raptus-wc-bexio-connector'));
            }
        }
    }

    /**
     * Check if a synchronization is already in progress.
     */
    protected function check_processing($order_id) {
        if (metadata_exists('post', $order_id, '_bexio_sync_processing')) {
            $url = add_query_arg(array(
                'order_bexio_sync' => 1,
                'order_bexio_force_sync' => 1,
            ), $_SERVER['REQUEST_URI']); // Redirect to self
            Utils::print_notice('warning',
                __('This order is already being processed, please be patient.', 'raptus-wc-bexio-connector') . '<br/>' .
                sprintf(__('Click %s if you wish to force a synchroniatzion now anyway, note that this might create a double entry in Bexio.', 'raptus-wc-bexio-connector'),
                '<a href="' . $url . '">' . __('here', 'raptus-wc-bexio-connector') . '</a>')
            );
            return true;
        }
        return false;
    }

    /**
     * Test if force sync is activated.
     */
    protected function is_force_sync() {
        return isset($_GET['order_bexio_force_sync']) && $_GET['order_bexio_force_sync'];
    }

    /**
     * Add Bexio information to oder detail admin page.
     *
     * @param \WC_Order $order The current order.
     */
    public function admin_order_meta($order) {
        echo '<h4 style="float:left;">' . __('Bexio Details: ', 'raptus-wc-bexio-connector') . '</h4>';
        echo '<p class="form-field form-field-wide wc-bexio-order-nr">';
        $order_id = method_exists($order, 'get_id') ? $order->get_id() : $order->id; // WC2+ compatible
        if (metadata_exists('post', $order_id, '_bexio_order_id')) { // WC3+ if ($order->meta_exists('_bexio_order_id') ) {
            foreach(Data::get_sync_types() as $type => $label) {
                if (metadata_exists('post', $order_id, '_bexio_' . $type . '_id')) { // WC3+ if ($order->meta_exists('_bexio_order_id') ) {
                    echo $label . ': ';
                    $this->print_bexio_edit_link($order_id, $type, true, __('without number', 'raptus-wc-bexio-connector'));
                    echo '<br/>';
                }
                $msg = $this->get_sync_status_message($_GET['post'], $type);
                if ($msg) {
                    echo '<br/><span class="notice notice-warning" style="display:inline-block;margin-bottom:2px;">' . $msg . '</span>';
                }
            }
            $unlink_url = $this->get_unlink_url();
            $confirm_msg = __('Do you really want to unlink this order from the connected bexio docments? Note that you have to manually delete any previously created documents in bexio.', 'raptus-wc-bexio-connector');
            printf('<small><a href="%s" class="button button-link button-link-delete" style="margin-right: 5px; opacity: 0.75;" onclick="if (!confirm(\'' . $confirm_msg. '\')) return false;">%s</a></small>', $unlink_url, __('Unlink', 'raptus-wc-bexio-connector'));
        } else if (metadata_exists('post', $order_id, '_bexio_sync_processing')) { // WC3+ if ($order->meta_exists('_bexio_order_nr') ) {
            echo __('This order is already being processed, please be patient.', 'raptus-wc-bexio-connector') . '<br/>';
            echo '<small><em>'.sprintf(__('Click %s if you wish to force a synchroniatzion now anyway, note that this might create a double entry in Bexio.', 'raptus-wc-bexio-connector'), __('the button below', 'raptus-wc-bexio-connector')).'</em></small><br/>';
            $url = $this->get_sync_url(true); // Redirect to self
            echo '<a href="' . $url . '" class="button" onclick="if(jQuery(this).hasClass(\'processing\')) return false; jQuery(this).addClass(\'disabled processing\')">' . __('Force synchronize now', 'raptus-wc-bexio-connector') . '</a>';
        } else {
            echo __('This order is not yet synchronized with Bexio.', 'raptus-wc-bexio-connector') . '<br/>';
            $url = $this->get_sync_url();
            echo '<a href="' . $url . '" class="button" onclick="if(jQuery(this).hasClass(\'processing\')) return false; jQuery(this).addClass(\'disabled processing\')">' . __('Synchronize now', 'raptus-wc-bexio-connector') . '</a>';
        }
        echo '<span style="display:block;">';
        foreach (Data::get_sync_types() as $type => $label) {
            if ($this->show_sync_url($type, $order_id, $order)) {
                $url = $this->get_sync_url(true, $type);
                printf('<a href="%s" class="button button-link" style="margin-right: 5px; opacity: 0.75;" onclick="if(jQuery(this).hasClass(\'processing\')) return false; jQuery(this).addClass(\'disabled processing\')">%s</a>', $url, sprintf(__('Create Bexio %s', 'raptus-wc-bexio-connector'), $label));
            }
        }
        echo '</span>';
        echo '</p>';
    }

    /**
     * Test if the sync button should be shown.
     */
    private function show_sync_url($type, $order_id, $order) {
        if (metadata_exists('post', $order_id, '_bexio_' . $type . '_id')) {
            return false;
        }
        if ($type != 'order' && !metadata_exists('post', $order_id, '_bexio_order_id')) {
            return false;
        }
        if ($type == 'payment' && (!metadata_exists('post', $order_id, '_bexio_invoice_id') || !$order->is_paid() || metadata_exists('post', $order_id, '_bexio_invoice_paid'))) {
            return false;
        }
        return true;
    }

    /**
     * Get the correct synchronization URL for the given sync type.
     */
    private function get_sync_url($force = false, $type = 1) {
        $args = array('order_bexio_sync' => $type);
        if ($force) {
            $args['order_bexio_force_sync'] = 1;
        }
        return add_query_arg($args, $_SERVER['REQUEST_URI']); // Redirect to self
    }

    /**
     * Get the URL for unlinking a WooCommerce order with the created Bexio documents.
     */
    private function get_unlink_url($done = false) {
        $args = array('order_bexio_unlink' => $done ? 'done' : 'requested');
        return add_query_arg($args, $_SERVER['REQUEST_URI']); // Redirect to self
    }

    /**
     * Get the appropriate status message for the given order id and status type.
     *
     * @param  int $order_id The current order ID.
     * @param  string $type The status type (delivery or invoice).
     * @return string The found status message or an empty string is none is available.
     */
    private function get_sync_status_message($order_id, $type) {
        $status = get_post_meta($order_id, "_bexio_{$type}_status", true);
        if ($status) {
            $types = Data::get_sync_types();
            if ($status == 1) {
                return sprintf(__('%s could not be issued.', 'raptus-wc-bexio-connector'), $types[$type]);
            } else if ($status == 2) {
                return sprintf(__('%s could not be paid.', 'raptus-wc-bexio-connector'), $types[$type]);
            } else if ($status == 3) {
                return sprintf(__('%s could not be created.', 'raptus-wc-bexio-connector'), $types[$type]);
            }
        }
        return false;
    }

    /**
     * Trigger a Bexio synchronization, forcefully if requested.
     */
    public function admin_actions() {
        if (!is_admin() || !isset($_GET['post'])) {
            return;
        }
        $order_id = $_GET['post'];
        if (isset($_GET['order_bexio_sync'])) {
            $order = wc_get_order($order_id);
            $sync_type = $_GET['order_bexio_sync'];
            $force_sync = $this->is_force_sync();
            if ($sync_type == 1) {
                Order::instance()->sync_all($order_id);
            } else if ($sync_type == 'order' && (SyncRules::can_sync_order($order) || $force_sync)) {
                Order::instance()->set_order($order_id);
            } else if ($sync_type == 'delivery' && (SyncRules::can_sync_delivery($order) || $force_sync)) {
                Order::instance()->create_delivery($order_id);
            } else if ($sync_type == 'invoice' && (SyncRules::can_sync_invoice($order) || $force_sync)) {
                Order::instance()->create_invoice($order_id);
            } else if ($sync_type == 'payment' && (SyncRules::can_sync_payment($order) || $force_sync)) {
                Order::instance()->pay_order($order_id);
            }
        } else if (isset($_GET['order_bexio_data_dump'])) {
            $order_data = Order::instance()->get_order_data($order_id);
            echo '<pre>';
            print_r($order_data);
            echo '</pre>';
            die();
        } else if (isset($_GET['order_bexio_unlink']) && $_GET['order_bexio_unlink'] == 'requested') {
            foreach(Data::get_sync_types() as $type => $label) {
                $items = array_merge(['id', 'nr', 'status', 'issued', 'paid'], $type == 'invoice' ? ['pdf_url'] : []);
                foreach ($items as $item) {
                    delete_post_meta($order_id, "_bexio_{$type}_{$item}");
                }
            }
            if (wp_redirect($this->get_unlink_url(true))) {
                exit;
            }
        }
    }

    /**
     * Enqueue custom CSS.
     */
    public function enqueue_admin_css() {
        wp_enqueue_style('wc-bexio-automator-admin-styles', \Raptus\WC\BexioConnector\ASSETS_URL . 'css/admin.css');
    }

    /**
     * Enqueue custom JS
     * */
    public function enqueue_admin_js() {
        if (is_admin() && isset($_REQUEST['manage-account'])) {
            wp_enqueue_script('wc-bexio-automator-manage-account', \Raptus\WC\BexioConnector\ASSETS_URL . 'js/manage-account.js', ['jquery'], \Raptus\WC\BexioConnector\PLUGIN_VERSION);
        }
    }
}
