<?php

// Namespace declaration
namespace Raptus\WC\BexioConnector;

// Exit if accessed directly
defined('ABSPATH') or die();

/**
 * The API class module of this plugin.
 */
class Api {

    /**
     * Holds the default remote URL.
     * @var string
     */
    const REMOTE_URL = 'https://api2.bexio-shop.ch/api';

    /**
     * Holds the actual remote URL.
     * @var string
     */
    private $remote_url;

    /**
     * The response code that are considered successful.
     * All other response codes can be considered an error.
     * @var array
     */
    private $successful_http_request_codes = array(200, 201, 202);

    /**
     * The base64 encoded basic authorization string.
     * This string is injected to the header of every request.
     * @var string
     */
    private $authorization;

    /**
     * The last receive error from the remote API.
     * @var string
     */
    private $error;

    /**
     * Use Singleton trait to disallow multiple instances of this class.
     */
    use Singleton;

    /**
     * Constructs a new instance of this class.
     * Registers all required filter and action hooks.
     */
    protected function __construct() {
        $this->remote_url = defined('BEXIO_CONNECTOR_REMOTE_URL') ? BEXIO_CONNECTOR_REMOTE_URL : self::REMOTE_URL;
        $this->set_authorization(Options::get('authorization'));
    }

    /**
     * Get the last occurred error while processing requests.
     *
     * @return array The latest error.
     */
    public function get_error() {
        return $this->error;
    }

    /**
     * Get the latest error message.
     *
     * @return string The latest error message.
     */
    public function get_error_message() {
        if ($this->error) {
            $msg = $this->error['status'] . ': ' . $this->error['message'];
            if ($this->error['type'] == 'oauth_refresh_token_expired') {
                $url = menu_page_url(Settings::SETTINGS_PAGE_ID, false);
                $msg .= ' <a href="'. $url .'" class="button">' . __('Refresh') . '</a>';
            }
            return $msg;
        }
    }

    /**
     * Set the basic authorization hash to inject into the header of every request.
     * The hash must be created as follows: `base64_encode("$user:$password)`
     * If no hash is set, not authorization will be injected.
     *
     * @param string The basic authorization hash
     */
    public function set_authorization($auth) {
        if ($auth) {
            $this->authorization = $auth;
        }
    }

    /**
     * Get the current client status.
     *
     * @return array The response from the remote API.
     */
    public function client_status() {
        $status = $this->request('GET', '/client');
        if ($status && $status['expires'] != Options::get('expires')) {
            Options::set('expires', $status['expires']);
        }
        return $status;
    }

    /**
     * Get the currently connected company information.
     *
     * @see    https://api.bexio-shop.ch/doc/#api-Connection-GetConnectedCompany
     * @return array The response from the remote API.
     */
    public function client_company() {
        return $this->request('GET', '/connection/company');
    }

    /**
     * Register a new client using the given client data.
     *
     * @see    https://api.bexio-shop.ch/doc/#api-Subscription-RegisterClient
     * @param  array $data The new client data.
     * @return array The response from the remote API.
     */
    public function client_register($data) {
        return $this->request('POST', '/client', $data);
    }

    /**
     * Update an existing client using the given client data.
     *
     * @see    https://api.bexio-shop.ch/doc/#api-Subscription-UpdateClient
     * @param  array $data The modified client data.
     * @return array The response from the remote API.
     */
    public function client_update($data) {
        return $this->request('PUT', '/client', $data);
    }

    /**
     * Get the authentication URL using the given authentication data.
     *
     * @see    https://api.bexio-shop.ch/doc/#api-Connection-GetOAuth2AuthenticationURL
     * @param  array $data The authentication data.
     * @return array The response from the remote API.
     */
    public function client_authenticate($data) {
        return $this->request('GET', '/connection/authenticate', $data);
    }

    /**
     * Get a list of Bexio users.
     *
     * @see    https://api.bexio-shop.ch/doc/#api-Settings-GetUserList
     * @return array The response from the remote API.
     */
    public function get_users() {
        return $this->request('GET', '/settings/user') ?: array();
    }

    /**
     * Get a list of Bexio taxes.
     *
     * @see    https://api.bexio-shop.ch/doc/#api-Settings-GetTaxListByType
     * @return array The response from the remote API.
     */
    public function get_taxes() {
        $taxes = $this->request('GET', '/settings/tax/sales_tax,net_tax,not_taxable_turnover');
        if (empty($taxes)) {
            $taxes = $this->request('GET', '/settings/tax');
        }
        return $taxes ?: array();
    }

    /**
     * Get a list of Bexio accounts.
     *
     * @see    https://api.bexio-shop.ch/doc/#api-Settings-GetAccountListByType
     * @return array The response from the remote API.
     */
    public function get_accounts() {
        return $this->request('GET', '/settings/account/1') ?: array();
    }

    /**
     * Get a list of Bexio contact groups.
     *
     * @see    https://api.bexio-shop.ch/doc/#api-Settings-GetContactGroups
     * @return array The response from the remote API.
     */
    public function get_contact_groups() {
        return $this->request('GET', '/settings/contact_group') ?: array();
    }

    /**
     * Get a list of Bexio bank accounts.
     *
     * @see    https://api.bexio-shop.ch/doc/#api-Settings-GetBankAccounts
     * @return array The response from the remote API.
     */
    public function get_bank_accounts() {
        return $this->request('GET', '/settings/bank_account') ?: array();
    }

    /**
     * Get a list of Bexio payment types.
     *
     * @see    https://api.bexio-shop.ch/doc/#api-Settings-GetPaymentTypes
     * @return array The response from the remote API.
     */
    public function get_payment_types() {
        return $this->request('GET', '/settings/payment_type') ?: array();
    }

    /**
     * Get a list of Bexio document templates.
     *
     * @see    https://api.bexio-shop.ch/doc/#api-Settings-GetDocumentTemplates
     * @return array The response from the remote API.
     */
    public function get_document_templates() {
        return $this->request('GET', '/settings/document_template') ?: array();
    }

    /**
     * Get a list of Bexio document settings.
     *
     * @see    https://api.bexio-shop.ch/doc/#api-Settings-GetDocumentSettings
     * @return array The response from the remote API.
     */
    public function get_document_settings() {
        return $this->request('GET', '/settings/document_setting') ?: array();
    }

    /**
     * Get a list of Bexio products.
     *
     * @see    https://api.bexio-shop.ch/doc/#api-Products-GetProductList
     * @return array The response from the remote API.
     */
    public function get_products($skus = '') {
        return $this->request('GET', '/product', $skus == '' ? array() : array('skus' => $skus)) ?: array();
    }

    /**
     * Get a list of Bexio product stocks.
     *
     * @see    https://api.bexio-shop.ch/doc/#api-Products-GetProductStock
     * @return array The response from the remote API.
     */
    public function get_stocks() {
        return $this->request('GET', '/product/stock') ?: array();
    }

    /**
     * Get a list of Bexio product stocks filtered by SKU numbers.
     *
     * @see    https://api.bexio-shop.ch/doc/#api-Products-GetProductStockBySKU
     * @param  array $skus The SKU numbers of all the products to fetch the current stock for.
     * @return array The response from the remote API.
     */
    public function get_stock($skus, $mode = null, $sku_modes = array()) {
        return $this->request('POST', '/product/stock', array('sku' => $skus, 'fetch_mode' => $mode, 'sku_modes' => $sku_modes)) ?: array();
    }

    /**
     * Get a list of Bexio product prices.
     *
     * @see    https://api.bexio-shop.ch/doc/#api-Products-GetProductPriceBySKU
     * @return array The response from the remote API.
     */
    public function get_prices($skus = array()) {
        return $this->request('POST', '/product/price', array('sku' => $skus)) ?: array();
    }

    /**
     * Get the current order states.
     *
     * @see    https://api.bexio-shop.ch/doc/#api-Orders-GetOrderStatesById
     * @param  array $order_ids The order ids of all the orders to fetch the current state for.
     * @return array The response from the remote API.
     */
    public function get_order_states($order_ids) {
        return $this->request('POST', '/order/state', array('ids' => $order_ids)) ?: array();
    }

    /**
     * Create a new order with the given order data.
     *
     * @see    https://api.bexio-shop.ch/doc/#api-Orders-CreateOrder
     * @param  array $data The order data acorrding to specfication.
     * @return array The response from the remote API.
     */
    public function set_order($data) {
        return $this->request('POST', '/order', $data);
    }

    /**
     * Pay the order with the given order id.
     *
     * @see    https://api.bexio-shop.ch/doc/#api-Orders-PayOrder
     * @param  int $order_id The id of the the order to pay.
     * @param  string $date_paid The date on which the payment has been issued. Optional. Default is order date.
     * @return array The response from the remote API.
     */
    public function pay_order($order_id, $date_paid = false) {
        return $this->request('POST', "/order/{$order_id}/pay", array('date_paid' => $date_paid));
    }

    /**
     * Create a remote invoice for the order with the given id.
     *
     * @see    https://api.bexio-shop.ch/doc/#api-Orders-CreateInvoice
     * @param  int $order_id The id of the the order to create an invoice for.
     * @param  array $data The additional data for creating the invoice.
     * @return array The response from the remote API.
     */
    public function create_invoice($order_id, $data = array()) {
        return $this->request('POST', "/order/{$order_id}/invoice", $data);
    }

    /**
     * Pay the invoice with the given invoice id.
     * No payment will be issued, if the invoice is already paid.
     *
     * @see    https://api.bexio-shop.ch/doc/#api-Invoices-PayInvoice
     * @param  int $invoice_id The id of the the invoice to pay.
     * @param  string $date_paid The date on which the payment has been issued. Optional. Default is invoice date.
     * @return array The response from the remote API.
     */
    public function pay_invoice($invoice_id, $date_paid = false) {
        return $this->request('POST', "/invoice/{$invoice_id}/pay", array('date_paid' => $date_paid));
    }

    /**
     * Get the invoice PDF with the given invoice id.
     * No payment will be issued, if the invoice is already paid.
     *
     * @see    https://api.bexio-shop.ch/doc/#api-Invoices-PayInvoice
     * @param  int $invoice_id The id of the the invoice to pay.
     * @param  string $date_paid The date on which the payment has been issued. Optional. Default is invoice date.
     * @return array The response from the remote API.
     */
    public function get_invoice_pdf($invoice_id) {
        return $this->request('GET', "/invoice/{$invoice_id}/pdf");
    }

    /**
     * Get the current invoice states.
     *
     * @see    https://api.bexio-shop.ch/doc/#api-Invoices-GetInvoiceStates
     * @param  array $invoice_ids The invoice ids of all the invoices to fetch the current state for.
     * @return array The response from the remote API.
     */
    public function get_invoice_states($invoice_ids) {
        return $this->request('POST', '/invoice/state', $invoice_ids );
    }

    /**
     * Create a remote delivery for the order with the given id.
     *
     * @see    https://api.bexio-shop.ch/doc/#api-Orders-CreateDelivery
     * @param  int $order_id The id of the the order to create a delivery for.
     * @param  array $data The additional data for creating the delivery.
     * @return array The response from the remote API.
     */
    public function create_delivery($order_id, $data = array()) {
        return $this->request('POST', "/order/{$order_id}/delivery", $data);
    }

    /**
     * Issue the delivery with the given delivery id.
     * Nothing will happen, if the delivery is already issued.
     *
     * @see    https://api.bexio-shop.ch/doc/#api-Deliveries-IssueDelivery
     * @param  int $delivery_id The id of the the delivery to issue.
     * @return array The response from the remote API.
     */
    public function issue_delivery($delivery_id) {
        return $this->request('POST', "/delivery/{$delivery_id}/issue");
    }

    /**
     * Request a resend of the verification mail.
     * This will also get a new salt, if no verification is already pending.
     *
     * @see    https://api.bexio-shop.ch/doc/#api-Deliveries-IssueDelivery
     * @return array The response from the remote API.
     */
    public function resend_verification_mail() {
        return $this->request('GET', '/client/confirmation/send');
    }

    /**
     * Send a log message to cloud logging.
     * This logging aims to aid the technical support in helping customers with problems.
     *
     * @see    https://api.bexio-shop.ch/doc/#api-Log-LogMessage
     * @param  array $data The message to log to the cloud logging.
     * @return array The response from the remote API.
     */
    public function log_to_remote($message, $level = 'info', $additional_data = array()) {
        $current_user = wp_get_current_user();
        if($current_user) {
            // Add the originator to the log (eg. from data / authorization resets)
            $additional_data['shop_user'] = $current_user->user_login;
        }
        $additional_data['sent_from'] = 'shop';
        if(is_array($message)) {
            // Ensure cloud log shows the whole message instead of only the first part
            $message = implode("\n", $message);
        }
        $data = array(
            'message' => $message,
            'level' => $level,
            'additional_data' => $additional_data
        );
        // Check if shop-side logging has been disabled
        if(apply_filters('raptus_wc_bexio_connector_remote_log', true)) {
            return $this->request('POST', '/log', $data);
        }
    }

    /**
     * Get the latest plugin version.
     *
     * @see    https://api.bexio-shop.ch/doc/#api-Plugins-GetLatestPlugin
     * @return array The response from the remote API.
     */
    public function get_latest_plugin() {
        return $this->request('GET', '/plugin/latest') ?: array();
    }

    /**
     * Handles requests to the WooCommerce Bexio Automator Service using the native WordPress HTTP API.
     *
     * @param  string $method The requst method (`POST`, `PUT`, `GET`, `DELETE`). Defaults to `GET`.
     * @param  string $route The relative route to send the request to. Leading slash required.
     * @param  array $data The request data array to send. GET request data will be encoded as query string.
     *         POST and PUT request data will be encoded as JSON in the body. Optional. Default to an empty array.
     * @return array The response from the remote API. Or false if communication is not possible.
     */
    protected function request($method = 'GET', $route = '', $data = array()) {

        // Reset latest error
        $this->error = null;

        // Check if authentication is required
        $headers = array('Content-type' => 'application/json');

        if (isset($this->authorization)) {
            $auth = array('Authorization' => 'Basic ' . $this->authorization);
            $headers = array_merge($headers, $auth);
        }

        $accept_language = isset($data['language']) ? $data['language'] : Utils::get_language_code();
        // Verify language is an actual string before setting it as header, to prevent strange values from Bexio
        if ($accept_language != '' && is_string($accept_language)) {
            $headers = array_merge($headers, array(
                'Accept-Language' => $accept_language
            ));
        }

        // Set additional headers
        $headers = array_merge($headers, array(
            'Plugin-version' => PLUGIN_VERSION,
            'WooCommerce-version' => Utils::get_woocommerce_version(),
            'WordPress-version' => $GLOBALS['wp_version'],
            'Real-domain' => Utils::get_domain()
        ));

        // Set default params
        $args = array(
            'method' => $method,
            'timeout' => 45,
            'sslverify' => defined('WP_DEBUG') && WP_DEBUG ? 0 : 1,
            'headers' => $headers,
        );

        // Allow filtering the basic request route.
        $route = apply_filters('raptus_wc_bexio_connector_request_route', $route, $method, $data);

        // Setting data as either body payload or query string dependent of the request method
        if ($method != 'GET') {
            $data = wp_json_encode($data);
            // If json_encode failed, show an explicit error message and do not send a request
            if ($data === false) {
                $last_error = json_last_error();
                $message = '';
                switch($last_error) {
                    case JSON_ERROR_SYNTAX:
                        $message = 'json_encode failed - Syntax error, malformed JSON';
                        break;

                    case JSON_ERROR_UTF8:
                        $message = 'json_encode failed - Malformed UTF-8 characters';
                        break;

                    default:
                        $message = 'json_encode failed - Data could not be encoded';
                        break;
                }
                $this->error = array(
                    'type' => 'data_error',
                    'status' => 400,
                    'message' => $message
                );
                return false;
            }
            $args['body'] = $data;
        } else {
            $route = add_query_arg(array_map(function($p) {
                return urlencode($p);
            }, $data), $route);
        }

        $response = wp_remote_request($this->remote_url . $route, $args);
        if (is_wp_error($response)) {
            $this->error = array(
                'type' => 'connection_error',
                'status' => $response->get_error_code(),
                'message' => __($response->get_error_message(), 'raptus-wc-bexio-connector')
            );
            error_log('WCBA - ' . print_r($this->error, true));
            return false;
        } else {
            $responseCode = wp_remote_retrieve_response_code($response);
            $responseBody = wp_remote_retrieve_body($response);
            $responseArray = json_decode($responseBody, true);

            if (!$this->validate_error($responseCode, $responseArray)) {
                return false;
            }

            return $responseArray;
        }
    }

    /**
     * Checks if the given response is an error and takes according measures.
     * The error case will be indicated be an according HTTP response status code.
     *
     * @param  int $code The response code. Any non 20x number is considered an error.
     * @param  array $response The response data.
     */
    protected function validate_error($code, $res) {
        if (in_array($code, $this->successful_http_request_codes)) {
            if (!empty($res['status']) && Options::get('isblocked')) {
                if (($res['status'] != 'blocked' && $res['status'] != 'cancelled') || ($res['status'] == 'cancelled' && isset($res['is_expired']) && !$res['is_expired'])) {
                    Options::remove('isblocked');
                }
            }
            return true;
        }

        if (isset($res['type'])) {
            switch ($res['type']) {
                case 'unauthorized_access':
                case 'subscription_expired':
                case 'invalid_salt':
                case 'end_of_life':
                    Options::set('isblocked', $res['type']);
                    break;

                case 'subscription_not_found':
                case 'credentials_not_found':
                    Options::remove('authorization');
                    Options::remove('isauthorized');
                    Options::remove('isblocked');
                    break;

                case 'oauth_error':
                    $message = @json_decode($res['message'], true);
                    if (is_array($message) && $message['error'] == 'invalid_grant' && $message['error_description'] == 'Invalid refresh token') {
                        Options::remove('isauthorized');
                    }
                    if (is_array($message) && $message['error'] == 'invalid_client') {
                        Options::remove('isauthorized');
                    }
                    break;

                case 'oauth_refresh_token_expired':
                    Options::remove('isauthorized');
                    break;
                default:
                    // Handle authorization failures
                    // if ($res['status'] == 401 ||
                    //     $res['status'] == 403) {
                    //     Options::remove('isauthorized');
                    // }
                    break;
            }
        }
        if (Options::get('isblocked')) {
            // do not render error since this will be handled by blocked template
        } else if ($res) {
            $this->error = $res;
        } else {
            $this->error = array(
                'type' => 'no_response',
                'status' => $code,
                'message' => __('No response from remote API', 'raptus-wc-bexio-connector')
            );
        }
        return false;
    }

}

?>