<?php

// Namespace declaration
namespace Raptus\WC\BexioConnector;

// Exit if accessed directly
defined('ABSPATH') or die();

/**
 * Data cache to avoid duplicate requests to the service.
 */
class Data {

    /**
     * The static cache for data that is not .
     * @var array
     */
    private static $cache;

    /**
     * The static file cache.
     * @var FileCache
     */
    private static $file_cache;

    /**
     * Get the request data from the cache.
     *
     * @return array the requested data or `null` if not found.
     */
    public static function __callStatic($name, $arguments) {
        $property = $name;
        if (($pos = strpos($name, '_')) !== false) {
            $property = substr($name, $pos + 1);
        }
        $cache = static::file_cache();
        if (!isset($cache[$property]) || !$cache[$property]) {
            $api = Api::instance();
            $data = null;
            if (method_exists($api, $property)) {
                $data = call_user_func([$api, $property]);
            } else if (method_exists($api, "get_{$property}")) {
                $data = call_user_func([$api, "get_{$property}"]);
            }
            $cache[$property] = $data;
            return $data;
        }
        return $cache[$property];
    }

    /**
     * Get the file cache instance.
     *
     * @return FileCache
     */
    private static function file_cache() {
        if (!static::$file_cache) {
            static::$file_cache = new FileCache('wcba-data');
        }
        return static::$file_cache;
    }

    /**
     * Remove the file cache with the specified key.
     * */
    public static function clear_cache($key) {
        $cache = static::file_cache();
        unset($cache[$key]);
    }

    /**
     * Get the Bexio subscription types.
     *
     * @return array the Bexio subscription types.
     */
    public static function get_subscription_types() {
        if (!isset(static::$cache['subscription_types']) || !static::$cache['subscription_types']) {
            static::$cache['subscription_types'] = array(
                'standard' => sprintf(__('%s subscription', 'raptus-wc-bexio-connector'), __('Standard', 'raptus-wc-bexio-connector')),
                'premium' => sprintf(__('%s subscription', 'raptus-wc-bexio-connector'), __('Pro', 'raptus-wc-bexio-connector')),
            );
        }
        return static::$cache['subscription_types'];
    }

    /**
     * Get the Bexio subscription states.
     *
     * @return array the Bexio subscription states.
     */
    public static function get_subscription_states() {
        if (!isset(static::$cache['subscription_states']) || !static::$cache['subscription_states']) {
            static::$cache['subscription_states'] = array(
                'inactive' => __('Inactive', 'raptus-wc-bexio-connector'),
                'active' => __('Active', 'raptus-wc-bexio-connector'),
                'blocked' => __('Blocked', 'raptus-wc-bexio-connector'),
                'cancelled' => __('Cancelled', 'raptus-wc-bexio-connector'),
                'open' => __('Open', 'raptus-wc-bexio-connector'),
            );
        }
        return static::$cache['subscription_states'];
    }

    /**
     * Get the contact creation modes.
     *
     * @return array The contact creation modes.
     */
    public static function get_contact_creation_modes() {
        if (!isset(static::$cache['contact_creation_modes']) || !static::$cache['contact_creation_modes']) {
            static::$cache['contact_creation_modes'] = array(
                array(
                    'id' => 'flat',
                    'name' => __('Flat (default)', 'raptus-wc-bexio-connector'),
                ), array(
                    'id' => 'nested',
                    'name' => __('Nested (recommended)', 'raptus-wc-bexio-connector')
                )
            );
        }
        return static::$cache['contact_creation_modes'];
    }

    /**
     * Get the shipping address modes.
     *
     * @return array The shipping address modes.
     */
    public static function get_shipping_address_modes() {
        if (!isset(static::$cache['bexio_shipping_address_mode']) || !static::$cache['bexio_shipping_address_mode']) {
            static::$cache['bexio_shipping_address_mode'] = array(
                array(
                    'id' => 'default',
                    'name' => _x('Default', 'Shipping address synchronization mode', 'raptus-wc-bexio-connector'),
                ), array(
                    'id' => 'force_additional_address',
                    'name' => _x('Force additional address', 'Shipping address synchronization mode', 'raptus-wc-bexio-connector')
                )
            );
        }
        return static::$cache['bexio_shipping_address_mode'];
    }

    /**
     * Get the synchroniatzion types.
     *
     * @return array The synchroniatzion types.
     */
    public static function get_sync_types($type = false) {
        if (!isset(static::$cache['sync_types']) || !static::$cache['sync_types']) {
            static::$cache['sync_types'] = array(
                'order'    => __('Order', 'raptus-wc-bexio-connector'),
                'delivery' => __('Delivery', 'raptus-wc-bexio-connector'),
                'invoice'  => __('Invoice', 'raptus-wc-bexio-connector'),
                'payment'  => __('Payment', 'raptus-wc-bexio-connector'),
            );
        }
        if ($type && isset(static::$cache['sync_types'][$type])) {
            return static::$cache['sync_types'][$type];
        }
        return static::$cache['sync_types'];
    }

    /**
     * Get the WooCommerce order states.
     *
     * @return array The WooCommerce order states.
     */
    public static function get_order_states() {
        if (!isset(static::$cache['order_states']) || !static::$cache['order_states']) {
            static::$cache['order_states'] = array();
            if (function_exists('wc_get_order_statuses')) {
                if ($states = wc_get_order_statuses()) {
                    foreach ($states as $status => $name) {
                        $status = preg_replace('/wc-(.*)/', '$1', $status);
                        static::$cache['order_states'][$status] = $name;
                    }
                }
            }
        }
        return static::$cache['order_states'];
    }

    /**
     * Get all enabled WooCommerce payment gateways.
     *
     * @return array The WooCommerce payment gateways.
     */
    public static function get_payment_gateways() {
        if (!isset(static::$cache['payment_gateways']) || !static::$cache['payment_gateways']) {
            static::$cache['payment_gateways'] = array();
            $gateways = \WC()->payment_gateways->payment_gateways();
            foreach ($gateways as $id => $gateway) {
                if ($gateway->enabled === 'yes') {
                    static::$cache['payment_gateways'][$id] = $gateway;
                }
            }
        }
        return static::$cache['payment_gateways'];
    }

    /**
     * Get all possible tax types.
     *
     * @return array The WooCommerce payment gateways.
     */
    public static function get_tax_types() {
        if (!isset(static::$cache['tax_types']) || !static::$cache['tax_types']) {
            static::$cache['tax_types'] = array(
                'pre_regards_tax_investment' => __('Reverse charge for investments and remaining expenses', 'raptus-wc-bexio-connector'),
                'pre_regards_tax_material' => __('Reverse charge for Material, Products and Services', 'raptus-wc-bexio-connector'),
                'pre_tax_investment' => __('Pre-tax for VAT: investments, other operating expenses', 'raptus-wc-bexio-connector'),
                'pre_tax_material' => __('Pre-tax for VAT: material, wares, services, energy', 'raptus-wc-bexio-connector'),
                'pre_customs_tax_investment' => __('Customs tax VAT Investments, other operating expenses', 'raptus-wc-bexio-connector'),
                'pre_customs_tax_material' => __('Customs tax VAT Material, goods, services, energy', 'raptus-wc-bexio-connector'),
                'sales_tax' => __('VAT due (sales tax)', 'raptus-wc-bexio-connector'),
                'non_consideration_sales_tax' => __('Non consideration', 'raptus-wc-bexio-connector'),
                'not_taxable_turnover' => __('Not taxable turnover', 'raptus-wc-bexio-connector'),
                'net_tax' => __('Net tax rate', 'raptus-wc-bexio-connector'),
                'opted_net_tax' => __('Opted net tax', 'raptus-wc-bexio-connector'),
            );
        }
        return static::$cache['tax_types'];
    }

    /**
     * Gets all WooCommerce tax classes.
     *
     * @return array The WooCommerce tax classes.
     */
    public static function get_tax_classes() {
        $tax_classes = \WC_Tax::get_tax_classes();
        array_unshift($tax_classes, __('Standard', 'woocommerce'));

        $taxes = array();
        foreach ($tax_classes as $tax) {
            $key = sanitize_title($tax);
            $taxes[$key] = $tax;
        }

        return $taxes;
    }

    /**
     * Get the stock fetch modes.
     *
     * @return array The contact creation modes.
     */
    public static function get_stock_fetch_modes() {
        if (!isset(static::$cache['stock_fetch_modes']) || !static::$cache['stock_fetch_modes']) {
            static::$cache['stock_fetch_modes'] = array(
                'inactive' => array(
                    'name' => __('Disabled', 'raptus-wc-bexio-connector'),
                    'desc' => __('Don\'t synchronize any stock with Bexio, disable the synchronization.', 'raptus-wc-bexio-connector'),
                ),
                'effective' => array(
                    'name' => __('Effective', 'raptus-wc-bexio-connector'),
                    'desc' => __('Represents the actual products in the warehouse.', 'raptus-wc-bexio-connector'),
                ),
                'reserved'  => array(
                    'name' => __('Reserved', 'raptus-wc-bexio-connector'),
                    'desc' => __('Stock of products that are reserved but no yet picked.', 'raptus-wc-bexio-connector'),
                ),
                'picked' => array(
                    'name' => __('Picked', 'raptus-wc-bexio-connector'),
                    'desc' => __('Stock of products that are picked (packed) but not yet shipped.', 'raptus-wc-bexio-connector'),
                ),
                'unpicked' => array(
                    'name' => __('Unpicked', 'raptus-wc-bexio-connector'),
                    'desc' => __('Effective stock without already picked products.', 'raptus-wc-bexio-connector'),
                ),
                'available' => array(
                    'name' => __('Available (default)', 'raptus-wc-bexio-connector'),
                    'desc' => __('Products that are neither reserved nor picked.', 'raptus-wc-bexio-connector'),
                    'default' => true,
                ),
            );
        }
        return static::$cache['stock_fetch_modes'];
    }

    /**
     * Get the order creation modes.
     *
     * @return array The order creation modes.
     */
    public static function get_order_creation_modes() {
        if (!isset(static::$cache['order_creation_modes']) || !static::$cache['order_creation_modes']) {
            static::$cache['order_creation_modes'] = array(
                array(
                    'id' => 'regular',
                    'name' => __('Regular (default)', 'raptus-wc-bexio-connector'),
                ), array(
                    'id' => 'async',
                    'name' => __('Asynchronous', 'raptus-wc-bexio-connector')
                )
            );
        }
        return static::$cache['order_creation_modes'];
    }

    /**
     * Get the stock synchronization modes.
     *
     * @return array The stock synchronization modes.
     */
    public static function get_stock_synchronization_modes() {
        if (!isset(static::$cache['stock_synchronization_modes']) || !static::$cache['stock_synchronization_modes']) {
            static::$cache['stock_synchronization_modes'] = array(
                array(
                    'id' => 'regular',
                    'name' => __('Regular (default)', 'raptus-wc-bexio-connector'),
                ), array(
                    'id' => 'async',
                    'name' => __('Asynchronous', 'raptus-wc-bexio-connector')
                )
            );
        }
        return static::$cache['stock_synchronization_modes'];
    }

    public static function delete_wcba_data_contents() {
        // Delete wcba-data contents
        $upload_dir = wp_upload_dir();
        $wcba_data_dir = $upload_dir['basedir'] . '/wcba-data';
        array_map('unlink', glob($wcba_data_dir . '/*'));
        @unlink($wcba_data_dir . '/.htaccess');
        @rmdir($wcba_data_dir);
    }

    /**
     * Get the order comment synchronization modes.
     * */
    public static function get_comment_sync_modes() {
        if (!isset(static::$cache['comment_sync_modes']) || !static::$cache['comment_sync_modes']) {
            static::$cache['comment_sync_modes'] = array(
                array(
                    'id' => 'default',
                    'name' => _x('Default', 'Comment synchronization mode', 'raptus-wc-bexio-connector'),
                ), array(
                    'id' => 'position',
                    'name' => _x('As position', 'Comment synchronization mode', 'raptus-wc-bexio-connector')
                )
            );
        }
        return static::$cache['comment_sync_modes'];
    }

    /**
     * Private constructor to avoid instances of this static utility class.
     */
    private function __construct() {}
}