<?php

// Namespace declaration
namespace Raptus\WC\BexioConnector;

// Exit if accessed directly
defined('ABSPATH') or die();

/**
 * Static class to provide default option values.
 */
class Defaults {

    /**
     * The default values.
     *
     * @var array
     */
    protected static $defaults;

    /**
     * The default sync rules.
     *
     * @var array
     */
    protected static $sync_rules;

    /**
     * Get the default value for the given key.
     * If no default value with that key is available, `null` is returned.
     *
     * @param  string $key The key of the default value to get.
     * @return mixed The found default value or `null` if not found.
     */
    public static function get($key) {
        $defaults = static::get_defaults();
        if (isset($defaults[$key])) {
            return $defaults[$key];
        }
        if (strpos($key, 'sync_rules') !== false) {
            $sync_rules = static::sync_rules_defaults();
            if (isset($sync_rules[$key])) {
                return $sync_rules[$key];
            }
        }
        return null;
    }

    /**
     * Get the default values array, lazily initialized.
     *
     * @return array The default values.
     */
    protected static function get_defaults($sync_rules = false) {
        if (empty(static::$defaults)) {
            static::$defaults = array(
                'isauthorized' => false,
                'authorize_uri' => null,
                'authorization' => null,
                'expires' => null,
                'isblocked' => null,
                'version' => false,

                'name' => '',
                'company' => '',
                'email' => get_option('admin_email', ''),
                'address' => '',
                'zip' => '',
                'city' => '',
                'country' => null,

                'bexio_order_title_pattern' => __('WC Order {order_id}', 'raptus-wc-bexio-connector'),
                'bexio_failed_order_sync_notification_email' => get_option('admin_email', ''),
                'bexio_contact_group' => null, // default is no contact group
                'bexio_contact_creation_mode' => 'nested', // new
                'bexio_shipping_address_mode' => 'default',

                'bexio_user_creator' => null, // first user from list, default behavior
                'bexio_user_owner' => null, // first user from list, default behavior
                'bexio_document_template' => null, // default is bexio document template default
                'bexio_account_products' => 101, // 3200
                'bexio_account_shipping' => 151, // 3600
                'bexio_account_discount' => 0, // none
                'bexio_tax_shipping' => 16, // 7.7%
                'bexio_tax_standard' => 16, // 7.7%
                'bexio_tax_reduced-rate' => 17, // 2.5%
                'bexio_tax_reduzierter-preis' => 17, // 2.5%
                'bexio_tax_zero-rate' => 3, // 0.0%
                'bexio_disable_taxes' => false,
                'bexio_override_invoice_due_date' => false,
                'bexio_auto_optimize_address_data' => false,
                'bexio_enable_price_rounding' => true,

                'bexio_display_product_excerpt' => false,
                'bexio_comments_sync_mode' => 'default',
                'bexio_create_immediate_delivery' => false,
                'bexio_issue_immediate_delivery' => false,
                'bexio_create_immediate_invoice' => false,
                'bexio_issue_immediate_invoice' => true,
                'bexio_pay_immediate_invoice' => false,
                'bexio_disable_stock_synchronization' => true,
                'bexio_sync_order_states' => false,
                'bexio_async_order_creation' => 'regular',
                'bexio_append_state_to_city' => false,
                'bexio_sync_free_shipping' => false,
                'bexio_sync_free_order' => true,
                'bexio_sync_free_delivery' => true,
                'bexio_sync_free_invoice' => false,
                'bexio_sync_free_payment' => false
            );

            //static::$defaults = array_merge(self::sync_rules_defaults(), static::$defaults);
        }
        return apply_filters('raptus_wc_bexio_connector_defaults', static::$defaults);
    }

    /**
     * Get the default statuses for the all synchroniatzion types.
     *
     * @return array The default values.
     */
    protected static function sync_rules_defaults() {
        if (empty(static::$sync_rules)) {
            static::$sync_rules = array();
            foreach (Data::get_payment_gateways() as $gateway_id => $gateway) {
                foreach (Data::get_sync_types() as $type_id => $type_name) {
                    static::$sync_rules['bexio_sync_rules_' . $gateway_id . '_' . $type_id] = self::get_sync_type_default($type_id);
                }
            }
            foreach (Data::get_sync_types() as $type_id => $type_name) {
                // Rules for virtual orders
                static::$sync_rules['bexio_sync_rules_virtual_' . $type_id] = 'default';
                // Fallback if no gateway is set (usually free orders)
                static::$sync_rules['bexio_sync_rules_' . $type_id] = self::get_sync_type_default($type_id);
            }
        }
        return static::$sync_rules;
    }

    /**
     * Get the default status for the given synchroniatzion type.
     *
     * @return string The default value.
     */
    protected static function get_sync_type_default($type) {
        if ($type == 'delivery' || $type == 'payment') {
            return 'processing';
        }
        return 'on-hold';
    }

    /**
     * Private constructor to disallow instances of this static class.
     */
    private function __construct() {}
}

?>