<?php

// Namespace declaration
namespace Raptus\WC\BexioConnector;

// Exit if accessed directly
defined('ABSPATH') or die();

/**
 * The email module for sending specialized email messages.
 */
class Email {

    /**
     * Use Singleton trait to disallow multiple instances of this class.
     */
    use Singleton;

    /**
     * Send the order synchronization failed email.
     *
     * @param  int $order_id The ID of the order where the synchronization failed.
     * @return boolean Whether the email contents were sent successfully.
     */
    public function send_order_sync_failed_email($order_id, $error_message) {
        $recipient = Options::get('bexio_failed_order_sync_notification_email', '');
        if (!$recipient) {
            return;
        }

        $order = wc_get_order($order_id);
        $subject = __('Order Synchronization with Bexio Failed', 'raptus-wc-bexio-connector');
        $msg = '<p>' . __('The synchronization with Bexio for the following order has failed', 'raptus-wc-bexio-connector') . ':</p>';

        $date_time_format = get_option('date_format') . ' ' . get_option('time_format');
        $data = array(
            _x('Order', 'email', 'raptus-wc-bexio-connector') => '#' . $order->get_order_number(),
            __('Error', 'raptus-wc-bexio-connector') => $error_message,
            __('Date', 'raptus-wc-bexio-connector') => date_i18n($date_time_format, time()),
        );
        foreach ($data as $key => $value) {
            $msg .= "$key: $value" . '<br/>';
        }

        $msg .= '<p>'. __('This might be a one-time error, we recommend that you perform a manual synchronization for this order.', 'raptus-wc-bexio-connector');
        $msg .= '<br/>' . sprintf(__('If the error persists, take a look at the %s for solutions to known issues.', 'raptus-wc-bexio-connector'),
                      '<a href="' . FAQ_TROUBLESHOOTING_URL . '">' . __('FAQ', 'raptus-wc-bexio-connector') . '</a>');
        $msg .= '</p>';

        $url = get_edit_post_link($order_id);
        if ($url) {
            $msg .= '<p>' . __('Direct link to the order', 'raptus-wc-bexio-connector') . ': ';
            $msg .= sprintf('<a href="%s">%s</a>', $url, $url) . '</p>';
        }

        return $this->send_email($subject, $msg, $recipient, true);
    }

    /**
     * Send the PDF download failed email.
     *
     * @param  int $order_id The ID of the order where the download failed.
     * @param  int $order_id The error message encountered in the API.
     * @return boolean Whether the email contents were sent successfully.
     */
    public function send_pdf_download_failed_email($order_id, $error_message) {
        $recipient = Options::get('bexio_failed_order_sync_notification_email', '');
        if (!$recipient) {
            return;
        }

        $order = wc_get_order($order_id);
        $subject = __('Download of Bexio Invoice PDF Failed', 'raptus-wc-bexio-connector');
        $msg = '<p>' . __('The download of the PDF for the invoice from Bexio for the following order has failed', 'raptus-wc-bexio-connector') . ':</p>';

        $date_time_format = get_option('date_format') . ' ' . get_option('time_format');
        $data = array(
            _x('Order', 'email', 'raptus-wc-bexio-connector') => '#' . $order->get_order_number(),
            __('Error', 'raptus-wc-bexio-connector') => $error_message,
            __('Date', 'raptus-wc-bexio-connector') => date_i18n($date_time_format, time()),
        );
        foreach ($data as $key => $value) {
            $msg .= "$key: $value" . '<br/>';
        }

        $msg .= '<p>'. __('This might be a one-time error, we recommend that you manually download and link the PDF for this invoice.', 'raptus-wc-bexio-connector');
        $msg .= '<br/>' . sprintf(__('If the error persists, take a look at the %s for solutions to known issues.', 'raptus-wc-bexio-connector'),
                      '<a href="' . FAQ_TROUBLESHOOTING_URL . '">' . __('FAQ', 'raptus-wc-bexio-connector') . '</a>');
        $msg .= '</p>';

        $url = get_edit_post_link($order_id);
        if ($url) {
            $msg .= '<p>' . __('Direct link to the order', 'raptus-wc-bexio-connector') . ': ';
            $msg .= sprintf('<a href="%s">%s</a>', $url, $url) . '</p>';
        }

        return $this->send_email($subject, $msg, $recipient, true);
    }

    /**
     * Send the upgrade account email.
     *
     * @param  array $upgrade_data The relevant upgrade data to include in the message.
     * @return boolean Whether the email contents were sent successfully.
     */
    public function send_upgrade_account_email($upgrade_data) {
        $subject = __('Upgrade account request', 'raptus-wc-bexio-connector');
        $message = __('The following WooCommerce Bexio Automator customer wishes to upgrade his account:', 'raptus-wc-bexio-connector') . PHP_EOL;

        if ($upgrade_data) {
            $message .= PHP_EOL .  strtoupper(__('Subscription', 'raptus-wc-bexio-connector')) . PHP_EOL;
            foreach ($upgrade_data as $key => $value) {
                $message .= "$key: $value" . PHP_EOL;
            }
        }
        $message = $this->append_customer_data($message);
        return $this->send_email($subject, $message, false, false, Options::get('email'));
    }

    /**
    * Send the account cancellation email.
    *
    * @return boolean Whether the email contents were sent successfully.
    */
    public function send_cancel_account_email($cancellation_reason = array()) {
        $subject = __('Account cancellation request', 'raptus-wc-bexio-connector');
        $message = __('The following WooCommerce Bexio Automator customer wishes to cancel his account:', 'raptus-wc-bexio-connector') . PHP_EOL;

        $message  = $this->append_customer_data($message);

        if ($cancellation_reason) {
            $message .= PHP_EOL . strtoupper(_x('Reason', 'Cancel account email', 'raptus-wc-bexio-connector')) . PHP_EOL;
            foreach ($cancellation_reason as $key => $value) {
                $message .= "$key: $value" . PHP_EOL;
            }
        }

        return $this->send_email($subject, $message, false, false, Options::get('email'));
    }

    /**
     * Get the relevant customer data from the option table.
     *
     * @return array The customer data.
     */
    private function get_customer_data() {
        $fields = array('name', 'company', 'email', 'address', 'zip', 'city', 'country', 'domain');
        $customer = array_intersect_key(Options::get(), array_flip($fields));
        $customer['domain'] = Utils::get_domain();
        return $customer;
    }

    /**
     * Append customer data to the given message.
     *
     * @param  string $message The message to append to.
     * @return string The modified message.
     */
    private function append_customer_data($message) {
        $customer_data = $this->get_customer_data();
        if ($customer_data) {
            $message .= PHP_EOL . strtoupper(__('Customer', 'raptus-wc-bexio-connector')) . PHP_EOL;
            foreach ($customer_data as $key => $value) {
                $message .= __(ucfirst($key), 'raptus-wc-bexio-connector') . ': ' . $value . PHP_EOL;
            }
        }
        return $message;
    }

    /**
     * Send the email.
     *
     * @param  string $subject The subject of the message to send.
     * @param  string $message The message to send.
     * @return boolean Whether the email contents were sent successfully.
     */
    private function send_email($subject, $message, $recipient = false, $html = false, $sender = false) {
        $recipient = $recipient ?: BEXIO_SHOP_EMAIL;
        $headers = array();
        if ($html) {
            $headers[] = 'Content-Type: text/html; charset=UTF-8';
        }
        $headers[] = 'From: WooCommerce Bexio Automator <' . get_option('admin_email') . '>';
        if ($sender) {
            $headers[] = 'Reply-To: ' . $sender;
        }
        $attachments = array();
        return wp_mail($recipient, $subject, $message, $headers, $attachments);
    }
}
?>