<?php

// Namespace declaration
namespace Raptus\WC\BexioConnector;

// Exit if accessed directly
defined('ABSPATH') or die();

use ArrayAccess;

/**
 * Data cache to avoid duplicate requests to the service.
 */
class FileCache implements ArrayAccess {

    /**
     * The default number of seconds to cache data, which equals to 15 minutes.
     *
     * @var integer
     */
    const DEFAULT_EXPIRATION_TIME = 60 * 15;

    /**
     * The number of seconds to cache data.
     *
     * @var integer
     */
    private $cache_expiration_time;

    /**
     * The base cache directory.
     * @var string
     */
    private $cache_dir;

    /**
     * Create a new instance of the file cache with the given target destination folder.
     *
     * @param bool|string The target directory to store cached data. The directory will be created if it doesn't exist. The default is the plugin slug.
     * @param bool|int The default cached expiration time in seconds for data storage. The default is 15 minutes.
     */
    public function __construct($target = false, $cache_expiration_time = false) {
        $upload_dir = wp_upload_dir();
        $this->cache_dir = $upload_dir['basedir'] . '/' . $target ?: PLUGIN_SLUG;
        $this->cache_expiration_time = $cache_expiration_time ?: self::DEFAULT_EXPIRATION_TIME;
        $this->maybe_create_htaccess();
    }

    /**
     * Check if there is a htaccess file in the cache directory and create one if needed.
     * */
    private function maybe_create_htaccess() {
        if (!file_exists($this->cache_dir)) {
            mkdir($this->cache_dir);
        }
        $htaccess_file = $this->cache_dir . '/.htaccess';
        if (!file_exists($htaccess_file)) {
            file_put_contents($htaccess_file, 'deny from all');
            file_put_contents($this->cache_dir . '/index.html', '');
        }
    }

    /**
     * Check whether an offset exists.
     * False is returned if no cached data is found or the cached data is expired.
     * @see https://www.php.net/manual/en/class.arrayaccess.php
     */
    #[\ReturnTypeWillChange]
    public function offsetExists($offset) {
        return file_exists($this->cache_file($offset)) && !$this->is_expired($offset, $this->get_expiration_time());
    }

    /**
     * Offset to retrieve.
     * @see https://www.php.net/manual/en/class.arrayaccess.php
     */
    #[\ReturnTypeWillChange]
    public function offsetGet($offset) {
        return $this->get($offset);
    }

    /**
     * Assign a value to the specified offset.
     * @see https://www.php.net/manual/en/class.arrayaccess.php
     */
    #[\ReturnTypeWillChange]
    public function offsetSet($offset, $value) {
        return $this->set($offset, $value);
    }

    /**
     *  Unset an offset.
     * @see https://www.php.net/manual/en/class.arrayaccess.php
     */
    #[\ReturnTypeWillChange]
    public function offsetUnset($offset) {
        return $this->remove($offset);
    }

    /**
     * Store the given data in the file cache using the provided key.
     *
     * @param  string $key The key identifier.
     * @return int|boolean The number of bytes that were written to the file cache, or `false` on failure.
     */
    public function set($key, $data) {
        return $this->write($key, $data);
    }

    /**
     * Get the data with the given key from the file cache.
     * If the cache dta is expired an empty value will be returned.
     *
     * @param  string $key The key identifier.
     * @param  bool|int The expiration time to use. Default is the object expiration time.
     * @return array the requested data or `null` if not found.
     */
    public function get($key, $expires = false) {
        $expiration_time = $this->get_expiration_time($expires);
        return $this->read($key, $expiration_time);
    }

    /**
     * Remove the data with the given key from the file cache.
     *
     * @return bool Returns true on success or false on failure.
     */
    public function remove($key) {
        return @unlink($this->cache_file($key));
    }

    /**
     * Write data to the according caching file.
     * Data is always stored in serialized format.
     *
     * @param  string $key The key identifier.
     * @param  mixed $data The data to write.
     * @return int|boolean The number of bytes that were written to the file, or `false` on failure.
     */
    private function write($key, $data) {
        if (!file_exists($this->cache_dir)) {
            mkdir($this->cache_dir);
        }
        $cache_file = $this->cache_file($key);
        $file_contents = file_put_contents($cache_file, serialize($data));
        if (function_exists('chmod')) {
            @chmod($cache_file, 0700);
        }
        return $file_contents;
    }

    /**
     * Read data from the according caching file.
     * Data is automatically unserialized.
     *
     * @param  string $key The key identifier.
     * @return array The data that was read or an empty value if not found.
     */
    private function read($key, $expiration_time) {
        if (!$this->is_expired($key, $expiration_time)) {
            $cache_file = $this->cache_file($key);
            $file_data = file_get_contents($cache_file);
            return unserialize($file_data);
        }
    }

    /**
     * Get the file name for the given identifier based on the underliyng cache configuration.
     */
    private function cache_file($key) {
        return $this->cache_dir . '/' . $key . '.data';
    }

    /**
     * Get the expiration time of this file cache.
     * If no value is given, the default expiration time is used, which is 15 minutes.
     * If a string is given it is converted to a timestamp using PHP string to time conversion.
     *
     * @return int The expiration time in seconds.
     */
    private function get_expiration_time($expiration_time = false) {
        if (is_string($expiration_time)) {
            $expiration_time = strtotime($expiration_time);
        }
        if (!$expiration_time) {
            $expiration_time = time() - $this->cache_expiration_time;
        }
        return $expiration_time;
    }

    /**
     * Check if the cached data is expired or still valid.
     *
     * @return boolean `true` if the data is expired, `false` otherwise.
     */
    private function is_expired($key, $expiration_time) {
        $cache_file = $this->cache_file($key);
        return !file_exists($this->cache_dir) || !file_exists($cache_file) || filemtime($cache_file) < $expiration_time;
    }
}