<?php

// Namespace declaration
namespace Raptus\WC\BexioConnector;

// Exit if accessed directly
defined('ABSPATH') or die();

/**
 * The migration class of this plugin.
 */
class Migrations {

    const MIGRATION_VERSION_OPTION_KEY = 'raptus_wp_bexio_connector_migration_version';

    /**
     * Use Singleton trait to disallow multiple instances of this class.
     */
    use Singleton;

    /**
     * Constructs a new instance of this class.
     * Registers all required filter and action hooks.
     */
    protected function __construct() {
        $this->check_migrations();
    }

    /**
     * Check current data migration progress and execute new migrations.
     */
    public function check_migrations() {
        $current_data_version = get_option(self::MIGRATION_VERSION_OPTION_KEY);
        // Append newer migrations to the bottom of this array
        $migrations = [
            'migration_replace_old_options',
            'migration_to_wcba2',
            'migration_stock_cache_table_structure',
            'migration_stock_cache_table_data',
        ];

        foreach ($migrations as $key => $migration) {
            $data_version = $key + 1;
            if ($current_data_version && $data_version <= $current_data_version) {
                // Skip migrations that have already been done
                continue;
            }
            if (method_exists($this, $migration)) {
                // Do new migration and update data version
                try {
                    call_user_func_array([$this, $migration], []);
                    update_option(self::MIGRATION_VERSION_OPTION_KEY, $data_version);
                } catch(\Exception $ex) {
                    $message = 'Could not execute the migration to data version ' . $data_version;
                    error_log($message);
                    error_log($ex->getMessage());
                    API::instance()->log_to_remote($message, 'error', array('error_message' => $ex->getMessage()));
                    break;
                }
            }
        }
    }

    /**
     * Migrate old options to newer settings (replaces 'raptus_wc_bexio_connector_old_options' filter).
     * These included:
     * - sync rules (as of v1.8.0) - migrated here
     * - stocksync fetch mode (from stocksync module) (as of v1.9.0) - adjusted in module defaults
     *
     * @since v2.0.0
     * */
    public function migration_replace_old_options() {
        // Update syncrule options
        $gateways = Data::get_payment_gateways();
        $this->update_sync_rule('bexio_create_immediate_invoice', $gateways, 'invoice', 'on-hold');
        $this->update_sync_rule('bexio_create_immediate_delivery', $gateways, 'delivery', 'on-hold');
        $this->update_sync_rule('bexio_pay_immediate_invoice', $gateways, 'payment', 'on-hold');
    }

    /**
     * Check if an old sync option key exists. If it does,  update the sync rule
     * of the given type for the specified gateways with the given value.
     * */
    private function update_sync_rule($old_key, $gateways, $type, $value) {
        if (Options::get($old_key)) {
            foreach ($gateways as $gateway_id => $gateway) {
                Options::set('bexio_sync_rules_' . $gateway_id . '_' . $type, $value);
            }
            Options::remove($old_key);
        }
    }

    /**
     * Migrate data to WCBA 2 (as of v2.0.0). This includes:
     * - deleting current authorization (re-authorizing is mandatory for new scopes!)
     * - clearing the file-cache to get up to date data (especially new tax display names)
     *
     * @since v2.0.0
     * */
    public function migration_to_wcba2() {
        // Remove authorization
        Options::remove('isauthorized');
        update_option('raptus_wp_bexio_connector_needs_reauthorization', true);

        // Clear file-cache
        Data::clear_cache('accounts');
        Data::clear_cache('bank_accounts');
        Data::clear_cache('cache');
        Data::clear_cache('client_company');
        Data::clear_cache('client_status');
        Data::clear_cache('contact_groups');
        Data::clear_cache('document_templates');
        Data::clear_cache('payment_types');
        Data::clear_cache('taxes');
        Data::clear_cache('users');
    }

    /**
     * Migrate stock cache table by adding new columns and an additional index.
     *
     * @since v2.3.0
     * */
    public function migration_stock_cache_table_structure() {
        if (get_option(Stock::STOCK_CACHE_TABLE_EXISTS)) {
            global $wpdb;
            $table_name = $wpdb->prefix . Stock::CACHE_TABLE_NAME;
            $col_check = $wpdb->get_results("SELECT COLUMN_NAME FROM INFORMATION_SCHEMA.COLUMNS
                WHERE table_name = '{$table_name}'
                AND column_name = 'stock_cachetime'");
            if (empty($col_check)) {
                error_log(sprintf('WCBA stock cache migration - adding column `stock_cachetime` to table %s', $table_name));
                $result = $wpdb->query("ALTER TABLE {$table_name} ADD `stock_cachetime` INT DEFAULT 0");
            } else {
                error_log('WCBA stock cache migration - table already has column `stock_cachetime`');
            }
        }
    }

    /**
     * Migrations for the plugin's stock cache table data. This includes:
     * - get product id and cache time from postmeta table and update existing entries
     * - remove entries with empty SKU or no found product id
     *
     * @since v2.3.0
     * */
    public function migration_stock_cache_table_data() {
        if (get_option(Stock::STOCK_CACHE_TABLE_EXISTS)) {
            global $wpdb;
            $table_name = $wpdb->prefix . Stock::CACHE_TABLE_NAME;

            // Get cachetime for existing entries based on SKU (ignore empty SKU's)
            $result = $wpdb->query("REPLACE INTO {$table_name} (SELECT c.sku, c.stock, IF(m.meta_value != '', m.meta_value, 0) FROM {$table_name} AS c INNER JOIN {$wpdb->wc_product_meta_lookup} AS w ON (w.sku = c.sku) INNER join {$wpdb->postmeta} as m on (m.post_id = w.product_id and m.meta_key = '_stock_cachetime') and c.sku <> '')");
            error_log(sprintf('WCBA - migrating stock cache data, copied data for %s products', $result ?? 0));

            // Delete entries with empty SKU
            $result = $wpdb->query("DELETE FROM {$table_name} WHERE sku = ''");
            error_log(sprintf('WCBA - migrating stock cache data, deleted %s entries with no SKU', $result ?? 0));
        }
    }

}
