<?php

// Namespace declaration
namespace Raptus\WC\BexioConnector;

// Exit if accessed directly
defined('ABSPATH') or die();

/**
 * Static class for handling plugin options.
 */
class Options {

    /**
     * The admin options key for storing settings in the database.
     *
     * @var string
     */
    const ADMIN_OPTION_KEY = 'raptus_wp_bexio_connector_options';

    /**
     * The keys which will be used for the second options called "raptus_wc_bexio-connector_options_connection" and
     * removed from the first options in database called "raptus_wc_bexio-connector_options".
     *     */
    private static $connection_options_keys = array( 'state', 'code', 'redirect_uri', 'authorize_uri', 'authorization', 'isauthorized', 'isblocked', 'expires', 'wizard', 'modules');

    /**
     * The static options array.
     * @var array
     */
    private static $options;

    /**
     * Get the option with the given key from the options table.
     * If no option with that key is available, the given default value is returned.
     * If no key is specified all options are returned.
     *
     * @param  string $key The key of the option to get. Optional.
     * @param  boolean $default The default value. Optional. Defaults to `false`.
     * @return mixed The found option value or the default value if not found.
     */
    public static function get($key = null, $default = false) {
        $options = static::get_options();
        if (is_null($key)) {
            return $options;
        }
        return isset($options[$key]) ? $options[$key] : static::get_default($key);
    }

    /**
     * Get get default value for the option with the given key.
     *
     * @param  string $key The key of the option default to get.
     * @return mixed The default option value.
     */
    private static function get_default($key) {
        return Defaults::get($key);
    }

    /**
     * Set the option with the given key from the options table to the given value.
     * Options are immediately saved to the database upon setting a new value.
     *
     * @param  string $key The key of the option to set.
     * @param  mixed $value The new value to set.
     */
    public static function set($key, $value) {
        if ($key) {
            $options = static::get_options();
            $options[$key] = $value;
            static::set_options($options);
        }
    }

    /**
     * Unset the option with the given key from the options table.
     * Options are immediately saved to the database upon unsetting the value.
     *
     * @param  string $key The key of the option to unset.
     */
    public static function remove($key) {
        if ($key) {
            $options = static::get_options();
            if (isset($options[$key])) {
                unset($options[$key]);
            }
            static::set_options($options);
        }
    }

    /**
     * Remove all option data from the options table.
     */
    public static function reset() {
        delete_option(static::ADMIN_OPTION_KEY);
        delete_option(static::ADMIN_OPTION_KEY . '_connection');
        // Reset static options to prevent deleted options being immediately overwritten
        static::$options = array();
        // Delete all other options
        global $wpdb;
        $wpdb->query($wpdb->prepare("DELETE FROM {$wpdb->options} WHERE option_name LIKE 'wcba_last_notification_%' OR option_name LIKE 'raptus_wp_bexio_connector_%'"));
    }

    /**
     * Get the options from the database.
     * The options are lazily initialized upon the first invocation for this method.
     * TODO describe why there is two options saved.
     *
     * @return array The options array from the database.
     */
    protected static function get_options() {
        if (!static::$options) {
            $options = get_option(static::ADMIN_OPTION_KEY, array());
            $connection_options = get_option(static::ADMIN_OPTION_KEY . '_connection', array());
            if ($connection_options) {
                $options = array_merge($connection_options, $options);
            }
            static::$options = $options;
        }
        return static::$options;
    }

    /**
     * Set and save the options.
     * TODO describe why there is two options saved.
     * The loop splits the options into the two options database fields, so they do not
     * contain the same keys.
     *
     * @param array $options The new options to save.
     */
    protected static function set_options($options = array()) {
        $options = $options ?: static::get_options();
        static::$options = $options;
        $options_connection = array();
        foreach (static::$connection_options_keys as $option){
            if (array_key_exists($option, $options)){
                $options_connection[$option] = $options[$option];
                unset($options[$option]);
            }
        }
        update_option(static::ADMIN_OPTION_KEY . '_connection', $options_connection, false);
        return update_option(static::ADMIN_OPTION_KEY, $options, false);
    }

    /**
     * Private constructor to disallow instances of this static class.
     */
    private function __construct() {}
}
