<?php

// Namespace declaration
namespace Raptus\WC\BexioConnector;

// Exit if accessed directly
defined('ABSPATH') or die();

use Exception;
use WC_Product;
use WC_Order;
use WC_Order_Item_Meta;
use WC_Product_Factory;

/**
 * The order module of this plugin.
 */
class Order {

    /**
     * The maximum length of Bexio order titles.
     * @var int
     */
    const BEXIO_MAX_TITLE_LENGTH = 80;

    /**
     * Use Singleton trait to disallow multiple instances of this class.
     */
    use Singleton;

    /**
     * Synchronize everything that can be synchronized.
     * If the order has already been synchronized, it won't be synchronized again.
     *
     * @param int $order_id The ID for the order to synchronize.
     */
    public function sync_all($order_id) {
        $order = wc_get_order($order_id);
        if ($order) {
            if (SyncRules::can_sync_order($order)) {
                $this->set_order($order_id);
            }
            if (SyncRules::can_sync_invoice($order)) {
                $this->create_invoice($order_id);
            }
            if (SyncRules::can_sync_delivery($order)) {
                $this->create_delivery($order_id);
            }
            if (SyncRules::can_sync_payment($order)) {
                $this->pay_order($order_id);
            }
            if (SyncRules::can_deliver_order($order)) {
                $this->deliver_order($order_id);
            }
        }
	}

	/**
     * Create an invoice for the given order.
	 *
     * All of the following requirements must be fullfilled:
     * - the order has already been synchronized
     * - no invoice was synchronized yet
     *
     * @param int $order_id The ID for the order to create an invoice for.
     */
	public function create_invoice($order_id) {
        $bexio_order_id = get_post_meta($order_id, '_bexio_order_id', true);
        $bexio_invoice_id = get_post_meta($order_id, '_bexio_invoice_id', true);

        if ($bexio_order_id && !$bexio_invoice_id && !$this->is_synchronizing($order_id)) {
            $this->start_sync_progress($order_id);
            $order = wc_get_order($order_id);
            $data = array(
                'issue' => Options::get('bexio_issue_immediate_invoice'),
                'date' => $this->get_order_date($order_id, $order),
                'date_paid' => $this->get_date_paid($order_id, $order),
                'payment_terms' => $this->get_payment_terms($order),
                'invoice_due_date' => Options::get('bexio_override_invoice_due_date') ? $this->get_date_paid($order_id, $order) : null,
            );
            $data = apply_filters('raptus_wc_bexio_connector_invoice_data', $data, $order);
            $result = Api::instance()->create_invoice($bexio_order_id, array_filter($data));
            $this->set_invoice_meta($order_id, $result);
            $this->stop_sync_progress($order_id);
        }
    }

	/**
     * Create a delivery for the given order.
	 *
     * All of the following requirements must be fullfilled:
     * - the order has already been synchronized
     * - no delivery was synchronized yet
     *
     * @param int $order_id The ID for the order to create a delivery for.
     */
	public function create_delivery($order_id) {
        $bexio_order_id = get_post_meta($order_id, '_bexio_order_id', true);
        $bexio_delivery_id = get_post_meta($order_id, '_bexio_delivery_id', true);

        if ($bexio_order_id && !$bexio_delivery_id && !$this->is_synchronizing($order_id)) {
            $this->start_sync_progress($order_id);
            $order = wc_get_order($order_id);
            $data = array(
                'issue' => Options::get('bexio_issue_immediate_delivery'),
            );
            $data = apply_filters('raptus_wc_bexio_connector_delivery_data', $data, $order);
            $result = Api::instance()->create_delivery($bexio_order_id, $data);
            $this->set_delivery_meta($order_id, $result);
            $this->stop_sync_progress($order_id);
        }
    }

	/**
     * Issue a payment for the given order.
	 *
     * All of the following requirements must be fullfilled:
     * - order synchronization was successful or the order has already been synchronized
     * - the invoice payment is enabled via the settings
     * - the order is marked paid in WooComemrce
     * - the payment is not yet synchronized with Bexio
     *
     * @param int $order_id The ID for the order to synchronize and pay.
     */
	public function pay_order($order_id) {
        $bexio_order_id = get_post_meta($order_id, '_bexio_order_id', true);
        $order = wc_get_order($order_id);
        if ($bexio_order_id && $order->is_paid()
            && !get_post_meta($order_id, '_bexio_invoice_paid', true)
            && !$this->is_synchronizing($order_id)) {

            $result = array();
            $date_paid = $this->get_date_paid($order_id, $order);
            $bexio_invoice_id = get_post_meta($order_id, '_bexio_invoice_id', true);

            $this->start_sync_progress($order_id);

            if ($bexio_invoice_id) {
                $result['invoice_id'] = $bexio_invoice_id;
                $result['invoice_paid'] = Api::instance()->pay_invoice($bexio_invoice_id, $date_paid);
                $result['invoice_issued'] = true;
            } else {
                $result = Api::instance()->pay_order($bexio_order_id, $date_paid);
            }
            $this->set_invoice_meta($order_id, $result);
            $this->stop_sync_progress($order_id);
        }
    }

    /**
     * Mark the given order as delivered.
	 *
     * All of the following requirements must be fullfilled:
     * - order synchronization was successful or the order has already been synchronized
     * - the invoice payment is enabled via the settings
     * - the order is marked paid in WooComemrce
     * - the payment is not yet synchronized with Bexio
     *
     * @param int $order_id The ID for the order to synchronize and pay.
     */
    public function deliver_order($order_id) {
        $bexio_order_id = get_post_meta($order_id, '_bexio_order_id', true);
        $order = wc_get_order($order_id);
        if (apply_filters('raptus_wc_bexio_connector_sync_and_deliver', $bexio_order_id
            && $order->is_paid()
            && get_post_meta($order_id, '_bexio_delivery_issued', true) != 1, $order)
            && !$this->is_synchronizing($order_id)
        ) {
            $result = array();
            $this->start_sync_progress($order_id);
            if (metadata_exists('post', $order_id, '_bexio_delivery_id')) {
                $delivery_id = get_post_meta($order_id, '_bexio_delivery_id', true);
                $result = Api::instance()->issue_delivery($delivery_id);
                if ($result) {
                    update_post_meta($order_id, '_bexio_delivery_issued', 1);
                } else {
                    update_post_meta($order_id, '_bexio_delivery_status', 1);
                }
            }
            $this->stop_sync_progress($order_id);
        }
    }

    /**
     * Add order information to Bexio and store Bexio order number on success.
     * Abort if a Bexio order number is already available.
     */
    public function set_order($order_id) {
        // Abort if order is already synchronized
        if (($this->is_synchronizing($order_id) || $this->is_synchronized($order_id)) && !isset($_GET['order_bexio_force_sync'])) {
            return;
        }

        // Synchronize the order with Bexio
        $success = false;
        $order_data = $this->get_order_data($order_id);
        if ($order_data) {
            $this->start_sync_progress($order_id);
            $result = Api::instance()->set_order($order_data);
            if ($result && $result['status']) {
                $this->set_order_meta($order_id, $result);
                $success = true;
            } else {
                $error = Api::instance()->get_error_message();
                Email::instance()->send_order_sync_failed_email($order_id, $error);
                $message = 'WCBA - Order synchronization failed for order ' . $order_id;
                error_log($message);
                error_log($error);
                Api::instance()->log_to_remote($message, 'error', array('error_message' => $error));
            }
            $this->stop_sync_progress($order_id);
        }
        return $success;
    }

    /**
     * Set order meta based on the order creation response.
     *
     * @param int $order_id The order id to set the meta data for.
     * @param array $result The remote response data.
     */
    protected function set_order_meta($order_id, $result) {
        update_post_meta($order_id, '_bexio_order_id', $result['order_id']);
        update_post_meta($order_id, '_bexio_order_nr', $result['order_nr']);

        $this->set_delivery_meta($order_id, $result);
        $this->set_invoice_meta($order_id, $result);
    }

    /**
     * Set delivery meta based on the remote response.
     *
     * @param int $order_id The order id to set the meta data for.
     * @param array $result The remote response data.
     */
    protected function set_delivery_meta($order_id, $result) {
        $order = wc_get_order($order_id);
        if ($result && isset($result['delivery_id'])) {
            update_post_meta($order_id, '_bexio_delivery_id', $result['delivery_id']);
            if (isset($result['delivery_nr']) && !empty($result['delivery_nr'])) {
                update_post_meta($order_id, '_bexio_delivery_nr', $result['delivery_nr']);
            }
            if (Options::get('bexio_issue_immediate_delivery')) {
                if (!isset($result['delivery_issued']) || !$result['delivery_issued']) {
                    update_post_meta($order_id, '_bexio_delivery_status', 1);
                } else {
                    update_post_meta($order_id, '_bexio_delivery_issued', 1);
                }
            }
        } else if (SyncRules::can_sync_delivery($order)) {
            update_post_meta($order_id, '_bexio_delivery_status', 3);
        }
    }

    /**
     * Set invoice meta based on the remote response.
     *
     * @param int $order_id The order id to set the meta data for.
     * @param array $result The remote response data.
     */
    protected function set_invoice_meta($order_id, $result) {
        $order = wc_get_order($order_id);
        if ($result && isset($result['invoice_id'])) {
            update_post_meta($order_id, '_bexio_invoice_id', $result['invoice_id']);
            if (isset($result['invoice_nr']) && !empty($result['invoice_nr'])) {
                update_post_meta($order_id, '_bexio_invoice_nr', $result['invoice_nr']);
            }
            if (Options::get('bexio_issue_immediate_invoice')) {
                if (!isset($result['invoice_issued']) || !$result['invoice_issued']) {
                    update_post_meta($order_id, '_bexio_invoice_status', 1);
                } else {
                    update_post_meta($order_id, '_bexio_invoice_issued', 1);
                }
            }
            if ($order->is_paid()) {
                if (isset($result['invoice_paid']) && $result['invoice_paid']) {
                    update_post_meta($order_id, '_bexio_invoice_paid', 1);
                    delete_post_meta($order_id, '_bexio_invoice_status');
                } else if (SyncRules::can_sync_payment($order)) {
                    update_post_meta($order_id, '_bexio_invoice_status', 2);
                }
            }
        } else if (SyncRules::can_sync_invoice($order)) {
            update_post_meta($order_id, '_bexio_invoice_status', 3);
        }
    }

    /**
     * Get an array of order data ready for synchronization with the WooCommerce Bexio Automator Service.
     *
     * @param  int $order_id The order id.
     * @return array The order data array.
     */
    public function get_order_data($order_id) {
        // Get order
        $order = wc_get_order($order_id);

        // Set basic order and customer data
        $order_data = array(
            'title' => $this->get_order_title($order_id, $order), //'WC Order #' . $order_id,
            'currency' => method_exists($order, 'get_currency') ? $order->get_currency() : $order->get_order_currency(), // WC2+3 compatible
            'owner_id' => Options::get('bexio_user_owner'),
            'user_id' => Options::get('bexio_user_creator'),
            'account_id' => Options::get('bexio_account_products'), // Fall-back if products do not exist in Bexio
            'mwst_type' => 0,
            'mwst_is_net' => get_option('woocommerce_tax_display_cart') == 'incl' ? 'false' : 'true',
            'reference' => $order_id,
            'customer' => array(
                'billing' => $this->get_customer_address($order, 'billing'),
                'shipping' => $this->get_customer_address($order, 'shipping'),
            ),
            'create_delivery' => SyncRules::can_sync_delivery($order),
            'issue_delivery' => Options::get('bexio_issue_immediate_delivery'),
            'create_invoice' => SyncRules::can_sync_invoice($order),
            'issue_invoice' => Options::get('bexio_issue_immediate_invoice'),
            'pay_invoice' => SyncRules::can_sync_payment($order),
            'contact_creation_mode' => Options::get('bexio_contact_creation_mode'),
            'language' => Utils::get_language_code($order_id, 'shop_order'),
            'date' => $this->get_order_date($order_id, $order),
        );

        // Set the date on which the order was paid
        if ($date_paid = $this->get_date_paid($order_id, $order)) {
            $order_data['date_paid'] = $date_paid;
        }

        // Set Bexio contact group
        if (Options::get('bexio_contact_group', false)) {
            $order_data['contact_group_id'] = Options::get('bexio_contact_group', 1);
        }

        // Set Bexio document template
        if (Options::get('bexio_document_template') && Options::get('bexio_document_template') != 'default') {
            $order_data['document_template'] = Options::get('bexio_document_template');
        }

        // Set the `mwst_type` type 2, which stands for "exempt from taxes" if taxes are disabled
        // See https://docs.bexio.com/resources/kb_order/#create-order
        if (Options::get('bexio_disable_taxes')) {
            $order_data['mwst_type'] = 2;
        }

        // Set Bexio bank account id
        if ($bank_account_id = $this->get_bank_account_id($order)) {
            $order_data['bank_account_id'] = $bank_account_id;
        }

        // Set Bexio payment type id
        if ($payment_type_id = $this->get_payment_type_id($order)) {
            $order_data['payment_type_id'] = $payment_type_id;
        }

        // Set payment method title
        if ($payment_method = $this->get_payment_method_title($order)) {
            $order_data['payment_method'] = $payment_method;
        }

        // Set payment method title
        if ($payment_terms = $this->get_payment_terms($order)) {
            $order_data['payment_terms'] = $payment_terms;
        }

        // Set invoice due date
        if (isset($order_data['date_paid']) && Options::get('bexio_override_invoice_due_date')) {
            $order_data['invoice_due_date'] = $order_data['date_paid'];
        }

        // Set customer type based on billing address
        $address = $order_data['customer']['billing'];
        $order_data['customer']['type'] = isset($address['company']) && !empty($address['company']) ? 'company' : 'private';

        // Set customer birthdate
        if ($birthdate = $this->get_birthdate($order_id, $order)) {
            $order_data['customer']['birthdate'] = $birthdate;
        }

        // Set order items
        if ($order_items = $this->get_order_items($order)) {
            $order_data['order_items'] = $order_items;
        }

        // Add fees and shipping
        $order_data['fees'] = $this->get_fees($order);
        if ($shipping = $this->get_shipping($order)) {
            $order_data['fees'][] = $shipping;
        }

        // Add discount
        if ($discount = $this->get_discount($order)) {
            $discount_type = 'discounts';
            if (isset($discount['account_id']) && !empty($discount['account_id'])) {
                $discount_type = 'fees';
            }
            if (!isset($order_data[$discount_type])) {
                $order_data[$discount_type] = array();
            }
            $order_data[$discount_type][] = $discount;
        }

        // Add order notes as comments
        if ($comments = $this->get_order_comments($order)) {
            if (Options::get('bexio_comments_sync_mode') == 'default') {
                $order_data['comments'] = $comments;
            } else {
                $order_data['texts'] = array();
                foreach ($comments as $comment) {
                    $order_data['texts'][] = sprintf('%s: %s', __('Order notes', 'woocommerce'), $comment['text']);
                }
            }
        }

        return apply_filters('raptus_wc_bexio_connector_order_data', $order_data, $order);
    }

    /**
     * Get order items for order.
     *
     * @param  WC_Order $order The order object.
     * @return array The order item data.
     */
    protected function get_order_items($order) {
        $order_items = array();
        $incl_tax = get_option('woocommerce_tax_display_cart') == 'incl';
        foreach ($order->get_items() as $item) {
            if ($incl_tax) {
                $price = ($item['line_subtotal'] + $item['line_subtotal_tax']) / $item['qty'];
            } else {
                $price = $item['line_subtotal'] / $item['qty']; // WC3+ ($item->get_subtotal()/$item->get_quantity()),
            }

            $product = $this->get_product_from_item($item);
            $data = array(
                'sku' => $product ? $product->get_sku() : '',
                'name' => $item['name'], // WC3+ $item->get_name(),
                'quantity' => $item['qty'], // WC3+ $item->get_quantity(),
                'price' => $this->round_price($price),
                'tax_id' =>  Options::get('bexio_tax_' . ($item['tax_class'] ?: 'standard')), // WC3+ $item->get_tax_class();
                'account_id' => Options::get('bexio_account_products'),
            );

            if (Options::get('bexio_display_product_excerpt')) {
                // get the excerpt and convert line breaks for Bexio
                $excerpt = str_replace(["\r\n", "\r", "\n"], '<br>', get_the_excerpt($item['product_id']));
                if ($excerpt) {
                    $data['text'] = $excerpt;
                }
            }

            $data = apply_filters('raptus_wc_bexio_connector_order_item_data', $data, $item, $product, $order);
            array_push($order_items, $data);
        }
        return $order_items;
    }

    /**
     * Get the order discount if available.
     *
     * @param  WC_Order $order The current order object.
     * @return array The discount data or `null`.
     */
    protected function get_discount($order) {
        $discount = $order->get_total_discount(get_option('woocommerce_tax_display_cart') != 'incl');
        if ($discount > 0) {
            $data = array(
                'name' => apply_filters('raptus_wc_bexio_connector_discount_title', __('Coupon','raptus-wc-bexio-connector'), $order),
            );
            if ($account = Options::get('bexio_account_discount')) {
                $data = array_merge($data, array(
                    'price' => $this->round_price($discount) * -1,
                    'tax_id' =>  Options::get('bexio_tax_standard'),
                    'account_id' => $account,
                ));
            } else {
                $data = array_merge($data, array(
                    'value' => $this->round_price($discount),
                    'is_percentual' => 'false',
               ));
           }
           return apply_filters('raptus_wc_bexio_connector_discount_data', $data, $order);
        }
    }

    /**
     * Get the order fees if available.
     *
     * @param  WC_Order $order The current order object.
     * @return array The fees data or an empty array.
     */
    protected function get_fees($order) {
        $fees = [];
        $incl_tax = get_option('woocommerce_tax_display_cart') == 'incl';
        foreach ($order->get_fees() ?: array() as $fee) {
            $amount = $fee['line_total'] + ($incl_tax ? $fee['line_tax'] : 0);
            $fees[] = apply_filters('raptus_wc_bexio_connector_fee_data', array(
                'name' => $fee['name'],
                'price' => $this->round_price($amount),
                'tax_id' => Options::get('bexio_tax_' . ($fee['tax_class'] ?: 'standard')),
                'account_id' => Options::get('bexio_account_products') // TODO do we need a separate tax configuration for this?
           ), $order);
        }
        return $fees;
    }

    /**
     * Get the order shipping if available.
     *
     * @param  WC_Order $order The current order object.
     * @return array The shipping data or `null`.
     */
    protected function get_shipping($order) {
        $incl_tax = get_option('woocommerce_tax_display_cart') == 'incl';
        $shipping = $order->calculate_shipping();
        if ($shipping > 0 || Options::get('bexio_sync_free_shipping')) {
            if ($incl_tax) {
                $shipping += $order->get_shipping_tax();
            }
            return apply_filters('raptus_wc_bexio_connector_shipping_data', array(
                'name' => apply_filters('raptus_wc_bexio_connector_shipping_title', $order->get_shipping_method(), $order),
                'price' => $this->round_price($shipping),
                'tax_id' => Options::get('bexio_tax_shipping'),
                'account_id' => Options::get('bexio_account_shipping')
           ), $order);
        }
    }

    /**
     * Get the order customer note as order comment if available.
     *
     * @param  WC_Order $order The current order object.
     * @return array The customer note data `null`.
     */
    protected function get_order_comments($order) {
        $comments = array();
        $customer_note = null;
        if (Utils::woocommerce_version_check('3.0')) {
            $customer_note = $order->get_customer_note();
        } else {
            $customer_note = $order->customer_note;
        }
        if ($customer_note) {
            $comments[] = array(
                'text' => $customer_note
            );
        }
        return $comments;
    }

    /**
     * Get item attribute meta data. Compatible with WC2+ and WC3+.
     *
     * @param  array $item The WooCommerce order item.
     * @return string The item attribute meta data.
     */
    protected function get_item_attribute_meta($item) {
        if (function_exists('wc_display_item_meta')) {
            return wc_display_item_meta($item, array('echo' => false));
        } else {
            $item_meta = new WC_Order_Item_Meta($item, $this->get_product_from_item($item));
            return $item_meta->display();
        }
    }

    /**
     * Get product from order item.
     *
     * @param  array $item The WooCommerce order item.
     * @return WC_Product|boolean The found product instance or
     *         `false` if the product cannot be loaded.
     */
    protected function get_product_from_item($item) {
        if (is_callable(array($item, 'get_product'))) {
            $product = $item->get_product();
            if ($product) {
                return $product;
            }
        }
        $id = $item['product_id'];
        if ($item['variation_id']) {
            $id = $item['variation_id'];
        }
        $factory = new WC_Product_Factory();
        return $factory->get_product(is_array($id) ? $id[0] : $id);
    }

    /**
     * Get customer address from the given order object.
     * With function is compatible with WC2+ and WC3+.
     *
     * @param  WC_Order $order The order object.
     * @param  string $type The address type, must be either `shipping` or `billing`.
     * @return array The assembled address data.
     */
    protected function get_customer_address($order, $type = 'billing') {
        $countries = WC()->countries->countries;
        $customer = $order->get_address($type); // WC 2+ compatible
        $customer['country_code'] = $customer['country'];
        $customer['country_name'] = isset($countries[$customer['country']]) ? $countries[$customer['country']] : '';
        $customer['type'] = isset($customer['company']) && !empty($customer['company']) ? 'company' : 'private';

        if (!Options::get('bexio_append_state_to_city', true)) {
            unset($customer['state']);
        }

        if (Options::get('bexio_auto_optimize_address_data')) {
            $customer = $this->optimize_adress($customer);
        }

        if ($type == 'shipping' && Options::get('bexio_shipping_address_mode') == 'force_additional_address') {
            $customer['force_additional_address'] = true;
        }

        return $customer;
    }

    /**
     * Optimize the given address.
     *
     * @param  array $address The address data to optimize.
     * @return array The optimized address data.
     */
    protected function optimize_adress($address) {
        $optimizations = [
            'strtolower' => ['email'],
            'strtoupper' => ['country', 'country_code'],
            'strtolower,ucfirst' => ['city'],
            'strtolower,ucwords' => ['first_name', 'last_name', 'address_1', 'address_2', 'country_name'],
        ];
        foreach($address as $key => &$data) {
            foreach ($optimizations as $functions => $filter) {
                if (in_array($key, $filter)) {
                    foreach (explode(',', $functions) as $func) {
                        $data = call_user_func($func, $data);
                    }
                }
            }
        }
        return $address;
    }

    /**
     * Gets the order title based on the configured title template.
     * Replaces all template strings with actual values from the current order.
     *
     * @param  int $order_id The id of the order.
     * @param  WC_Order $order_id The order object.
     * @return string The order title.
     */
    protected function get_order_title($order_id, $order) {
        $pattern = Options::get('bexio_order_title_pattern', 'WC Order {order_id}');
        $replacements = $this->get_template_replacements($order_id, $order);
        $title = str_replace($replacements['search'], $replacements['replace'], $pattern);

        // Replace meta placeholders
        $title = $this->replace_meta_placeholders($title, $order_id);

        // Allow filtering of the order title for a more dynamic/complex behavior.
        $title = apply_filters('raptus_wc_bexio_connector_order_title', $title, $order);

        // Ensure max Bexio order title length is not exceeded (Bexio limitation)
        return substr($title, 0, self::BEXIO_MAX_TITLE_LENGTH);
    }

    /**
     * Gets a list of all supported template placeholders and their replacements values.
     * More formally, it returns an array with all template placeholders and another array
     * with all replacement values. The indices of these two array match an can therefore be
     * use with PHPs native `str_replace` function for easy replacement.
     *
     * @param  int $order_id The id of the order.
     * @param  \WC_Order $order The order object.
     * @return array The array with `search` and `replace` arrays set.
     */
    protected function get_template_replacements($order_id, $order) {
        $replace = Utils::flatten(apply_filters('raptus_wc_bexio_connector_order_title_template_replacements', array(
            'order_id'             => $order_id,
            'order_no'             => $order->get_order_number(),
            'shipping_method'      => $order->get_shipping_method(),
            'shipping'             => $this->get_customer_address($order, 'shipping'),
            'billing'              => $this->get_customer_address($order, 'billing'),
            'currency'             => $this->get_currency($order),
            'payment_method'       => $this->get_payment_method($order),
            'payment_method_title' => $this->get_payment_method_title($order),
            'top_product'          => $this->get_top_product($order),
        ), $order_id, $order));

        $search = array_map(function($placeholder) {
            return "{{$placeholder}}";
        }, array_keys($replace));

        return array('search' => $search, 'replace' => $replace);
    }

    /**
     * Replace all meta key placeholders with their respective meta value.
     *
     * @param string $order_title The order title to repalce.
     * @param string $order_id The id of the affected order.
     * @return string The modified order title.
     */
    protected function replace_meta_placeholders($order_title, $order_id) {
        return preg_replace_callback('/{meta:(.*?)}/', function($matches) use ($order_id) {
            return get_post_meta($order_id, $matches[1], true);
        }, $order_title);
    }

    /**
     * Round the price based on WooCommerce settings.
     *
     * @param  float $price The price to round.
     * @return float The rounded price.
     */
    protected function round_price($price) {
        if (Options::get('bexio_enable_price_rounding')) {
            $price = wc_format_decimal($price, wc_get_price_decimals());
        }
        return $price;
    }

    /**
    * Get the currency of the order.
    * Compatible with WC2+ and WC3+.
    *
    * @param  \WC_Order $order The order object.
    * @return string The currency of the order.
    */
    public function get_currency($order) {
        if (empty($order)) {
            return '';
        }
        if (method_exists($order, 'get_currency')) {
            return $order->get_currency();
        }
        $order_id = method_exists($order, 'get_id') ? $order->get_id() : $order->id;
        return $order->get_order_currency() ?: get_post_meta($order_id, '_order_currency', true);
    }

    /**
     * Get the payment method of the order.
     * Compatible with WC2+ and WC3+.
     *
     * @param  \WC_Order $order The order object.
     * @return string The payment method of the order.
     */
    public function get_payment_method($order) {
        if (empty($order)) {
            return '';
        }
        if (method_exists($order, 'get_payment_method')) {
            return $order->get_payment_method();
        }
        $order_id = method_exists($order, 'get_id') ? $order->get_id() : $order->id;
        return $order->payment_method ?: get_post_meta($order_id, '_payment_method', true);
    }

    /**
     * Get the payment method title of the order.
     * Compatible with WC2+ and WC3+.
     *
     * @param  \WC_Order $order The order object.
     * @return string The payment method title of the order.
     */
    public function get_payment_method_title($order) {
        if (empty($order)) {
            return '';
        }
        if (method_exists($order, 'get_payment_method_title')) {
            return $order->get_payment_method_title();
        }
        $order_id = method_exists($order, 'get_id') ? $order->get_id() : $order->id;
        return $order->payment_method_title ?: get_post_meta($order_id, '_payment_method_title', true);
    }

    /**
     * Get the top product of the order.
     * That is the most expensive product, disregarding quantity.
     *
     * @param  \WC_Order $order The order object.
     * @return string The top product's name of the order.
     */
    public function get_top_product($order) {
        $top_product = null;
        foreach ($this->get_order_items($order) as $item) {
            if (empty($top_product) || $item['price'] > $top_product['price']) {
                $top_product = $item;
            }
        }
        return $top_product && isset($top_product['name']) ? $top_product['name'] : '';
    }

    /**
     * Get the correct bank account id for the given order, based on the order payment method.
     *
     * @param  \WC_Order $order The order object.
     * @return string The matching bank account id for the order.
     */
    protected function get_bank_account_id($order) {
        return $this->get_payment_method_option($order, 'bank_account_id');
    }

    /**
     * Get the correct payment type id for the given order, based on the order payment method.
     *
     * @param  \WC_Order $order The order object.
     * @return string The matching payment type id for the order.
     */
    protected function get_payment_type_id($order) {
        return $this->get_payment_method_option($order, 'payment_type_id');
    }

    /**
     * Get the  payment terms for the given order, based on the order payment method.
     *
     * @param  \WC_Order $order The order object.
     * @return string The matching payment terms for the order.
     */
    protected function get_payment_terms($order) {
        return $this->get_payment_method_option($order, 'payment_terms');
    }

    /**
     * Get a payment method dependent option value for the given order, based on the order payment method.
     *
     * @param  \WC_Order $order The order object.
     * @return string The matching option value for the order.
     */
    protected function get_payment_method_option($order, $option) {
        $order_id = method_exists($order, 'get_id') ? $order->get_id() : $order->id;
        $payment_method = get_post_meta($order_id, '_payment_method', true);
        $option_id = 'bexio_payment_method_' . $payment_method . '_' . $option;
        return Options::get($option_id, false);
    }

    /**
     * Get date at which the given order was created.
     *
     * @param  int $order_id The order id.
     * @param  \WC_Order $order The order object.
     * @param  string $format The date format. Optional. Defaults to 'Y-m-d'.
     * @return string The found and formatted date or `null`.
     */
    protected function get_order_date($order_id, $order, $format = 'Y-m-d') {
        $date = null;
        if (method_exists($order, 'get_date_created')) {
            $date = $order->get_date_created();
        }
        if (empty($date) && method_exists($order, 'get_date_completed')) {
            $date = $order->get_date_completed();
        }
        if (empty($date)) {
            $date = $order->order_date;
        }
        if (!is_string($date)) {
            try {
                $date = $date->format($format);
            } catch (\Exception $ex) {
                $message = 'WCBA - Failed converting order date';
                error_log($message);
                Api::instance()->log_to_remote($message, 'error', array('error_message' => $ex->getMessage()));
            }
        }
        return $date;
    }

    /**
     * Get date at which the given order was paid.
     *
     * @param  int $order_id The order id.
     * @param  \WC_Order $order The order object.
     * @param  string $format The date format. Optional. Defaults to 'Y-m-d'.
     * @return string The found and formatted date or `null`.
     */
    protected function get_date_paid($order_id, $order, $format = 'Y-m-d') {
        $date_paid = null;
        if ($order->is_paid()) {
            if (method_exists($order, 'get_date_paid') && ($date = $order->get_date_paid())) {
                $date_paid = $date->format($format);
            } else if ($date = get_post_meta($order_id, '_paid_date', true)) {
                $date_paid = date($format, strtotime($date));
            }
        }
        return $date_paid;
    }

    /**
     * Get date at which the given order was paid.
     *
     * @param  int $order_id The order id.
     * @param  \WC_Order $order The order object.
     * @param  string $format The date format. Optional. Defaults to 'Y-m-d'.
     * @return string The found and formatted date or `null`.
     */
    protected function get_birthdate($order_id, $order, $format = 'Y-m-d') {
        $birthdate = null;

        // 1. get from billing address
        $address = $order->get_address('billing');
        if (isset($address['birthdate']) && !empty($address['birthdate'])) {
            $birthdate = $address['birthdate'];
        }

        // 2. get from order meta
        if ($dob = get_post_meta($order_id, '_billing_birthdate', true)) {
            $birthdate = $dob;
        }

        // 3. get from user meta
        $user_id = $order->get_user_id();
        if ($dob = get_user_meta($user_id, 'user_birthdate', true)) {
            $birthdate = $dob;
        }

        // reformat the date if one was found
        if ($birthdate && is_string($birthdate)) {
            $birthdate = date($format, strtotime($birthdate));
        }

        return $birthdate;
    }

    /**
     * Test if the given order is already synchronized.
     *
     * @param  int $order_id The order id to test.
     * @return boolean True if it is already synchronized, false otherwise.
     */
    private function is_synchronized($order_id) {
        // WC3+ if ($order->meta_exists('_bexio_order_nr') && !empty($order->get_meta('_bexio_order_nr')) ) {
        return metadata_exists('post', $order_id, '_bexio_order_nr') && !empty(get_post_meta($order_id, '_bexio_order_nr', true));
    }

    /**
     * Test if any synchronization activity is ongoing for the given order.
     *
     * @param  int $order_id The order id to test.
     * @return boolean True if synchronization is ongoing, false otherwise.
     */
    private function is_synchronizing($order_id) {
        return metadata_exists('post', $order_id, '_bexio_sync_processing') && get_post_meta($order_id, '_bexio_sync_processing', true) == 1;
    }

    /**
     * Start the synchronization progress, lock the order.
     *
     * @param  int $order_id The ID of the order to lock.
     */
    private function start_sync_progress($order_id) {
        update_post_meta($order_id, '_bexio_sync_processing', 1);
    }

    /**
     * Stop the synchronization progress, unlock the order.
     *
     * @param  int $order_id The ID of the order to lock.
     */
    private function stop_sync_progress($order_id) {
        delete_post_meta($order_id, '_bexio_sync_processing');
    }
}

?>