<?php

// Namespace declaration
namespace Raptus\WC\BexioConnector;

// Exit if accessed directly
defined('ABSPATH') or die();

// Dependencies
use WC_Tax;
use WC_Countries;

/**
 * The admin settings module of this plugin.
 */
class Settings {

    /**
     * The ID/slug of the plugin settings page.
     * @var string
     */
    const SETTINGS_PAGE_ID = 'raptus_wp_bexio_connector_options_admin';

    const CONFIG_SAVE_OPTION_KEY = 'raptus_wp_bexio_connector_sent_config';

    /**
     * Use Singleton trait to disallow multiple instances of this class.
     */
    use Singleton;

    /**
     * The Bexio users.
     * @var array
     */
    private $status;

    /**
     * The WooCommerce Bexio Automator settings URL.
     * @var string
     */
    private $settings_url;

    /**
     * The plugin settings page hook.
     * @var string
     */
    private $page_hook;

    /**
     * Constructs a new instance of the Admin class.
     * Registers all required filter and action hooks.
     */
    protected function __construct() {
        // Register the admin menu
        add_action('admin_menu', array($this, 'register_menu'));

        // Register the admin settings page and fields
        add_action('admin_init', array($this, 'register_settings'));

        // Allow Shop Managers to save Bexio settings
        add_filter('option_page_capability_raptus_wp_bexio_connector_option_group', array($this, 'settings_page_capability'));

        // Display status messages
        add_action('admin_notices', array($this, 'check_status'));

        // Make the options page a two column layout
        add_filter('screen_layout_columns', array($this, 'add_screen_layout_columns'), 10, 2);

        add_action('updated_option', array($this, 'log_configuration_change'), 10, 3);

        if(!get_option(self::CONFIG_SAVE_OPTION_KEY)) {
            add_action('init', [$this, 'send_current_config_to_remote'], 1000);
        }
    }

    /**
     * Send the current configuration to plugin service.
     * The storing of the Bexio Automator configurations aids the technical support
     * in helping customers in case of problems.
     * 
     * Initial configuration is only sent if it has not yet been sent and either:
     * - is a newer version with wizard and authorization is set + wizard was completed, or
     * - is older version (without wizard) and only authorization is set
     * */
    public function send_current_config_to_remote() {
        $authorization = Options::get('authorization');
        $wizard = Options::get('wizard');
        if(!(($authorization && $wizard && $wizard == 'done') || ($authorization && !$wizard))) {
            return;
        }

        $options = get_option(Options::ADMIN_OPTION_KEY, array());
        $logged = Api::instance()->log_to_remote('Plugin configuration initial save', 'info', array(
            'configuration' => $options
        ));
        if(isset($logged['stored_config']) && $logged['stored_config']) {
            update_option(self::CONFIG_SAVE_OPTION_KEY, 1);
        }
    }

    /**
     * Log changes to the plugin configuration remotely to the cloud.
     * Logging through the plugin can be disabled with the filter 'raptus_wc_bexio_connector_remote_log'.
     * 
     * To completely disable cloud logging, contact the support and request deactivation
     * of cloud logging on the service side (also includes logging through the plugin).
     * */
    public function log_configuration_change($option, $old_value, $value) {
        // Only send config change if an initial config has been sent and authorization is set
        if(!Options::get('authorization') || !get_option(self::CONFIG_SAVE_OPTION_KEY)) {
            return;
        }
        if($option == Options::ADMIN_OPTION_KEY) {
            Api::instance()->log_to_remote('Plugin configuration changed', 'info', array(
                'configuration' => $value
            ));
        }
    }

    /**
     * Get the settings page URL.
     *
     * @param  array The additional query args to append the URL. Optional.
     * @return string The settings page URL.
     */
    public function get_settings_page_url($args = array()) {
        if (!$this->settings_url) {
            $this->settings_url = menu_page_url(self::SETTINGS_PAGE_ID, false);
        }
        return add_query_arg($args, $this->settings_url);
    }

    /**
     * Make the options page a two column layout.
     */
    public function add_screen_layout_columns($columns, $screen) {
        if (!is_array($columns)) {
            return $columns;
        }
        if ($screen == $this->page_hook) {
            $columns[$this->page_hook] = 2;
        }
        return $columns;
    }

    /**
     * Returns the string `manage_woocommerce` statically.
     * This can be used as a callback to set custom capabilites.
     *
     * @return string `manage_woocommerce` statically.
     */
    public function settings_page_capability($cap) {
        return 'manage_woocommerce';
    }

    /**
     * Add menu page to WP settings
     */
    public function register_menu() {
        $this->page_hook = add_submenu_page('woocommerce',
            'Settings Admin',
            __('Bexio Automator', 'raptus-wc-bexio-connector'),
            'manage_woocommerce',
            self::SETTINGS_PAGE_ID,
            array($this, 'render_options_page')
        );
        add_action('load-' . $this->page_hook, array($this, 'add_meta_boxes'));
    }

    /**
     * Register settings meta boxes.
     */
    public function add_meta_boxes() {
        wp_enqueue_script('common');
        wp_enqueue_script('wp-lists');
        wp_enqueue_script('postbox');

        add_meta_box('raptus_wp_bexio_connector_registration',
            Options::get('isauthorized') ? $this->get_settings_title_html(__('Account information', 'raptus-wc-bexio-connector'), __('Manage your account and subscription', 'raptus-wc-bexio-connector')) : __('Create user account', 'raptus-wc-bexio-connector'),
            [$this, 'do_settings_sections_registration'],
            $this->page_hook,
            'normal',
            'core'
        );

        add_meta_box('raptus_wp_bexio_connector_actions',
            __('Actions'),
            [$this, 'print_action_buttons'],
            $this->page_hook,
            'side',
            'high'
        );

        add_meta_box('raptus_wp_bexio_connector_system',
            __('System information', 'raptus-wc-bexio-connector'),
            [$this, 'print_system_information'],
            $this->page_hook,
            'side',
            'low'
        );

        if (Options::get('isauthorized') || Options::get('wizard') == 'done') {
            add_meta_box('raptus_wp_bexio_connector_configuration',
                $this->get_settings_title_html(
                    __('Bexio options', 'raptus-wc-bexio-connector'),
                    __('Configure users, tax rates and bank accounts', 'raptus-wc-bexio-connector')
                ),
                [$this, 'do_settings_sections_configuration'],
                $this->page_hook,
                'normal',
                'core'
            );

            add_meta_box('raptus_wp_bexio_connector_advanced',
                $this->get_settings_title_html(
                    __('Synchronization data', 'raptus-wc-bexio-connector'),
                    __('Set order title, contacts and product data', 'raptus-wc-bexio-connector')
                ),
                [$this, 'do_settings_sections_advanced'],
                $this->page_hook,
                'normal',
                'core'
            );

            add_meta_box('raptus_wp_bexio_connector_sync_rules',
                $this->get_settings_title_html(
                    __('Synchronization rules', 'raptus-wc-bexio-connector'),
                    __('Create orders, invoices, delieveries and payments', 'raptus-wc-bexio-connector')
                ),
                [$this, 'do_settings_sections_sync_rules'],
                $this->page_hook,
                'normal',
                'core'
            );

            add_meta_box('raptus_wp_bexio_connector_payment_methods',
                $this->get_settings_title_html(
                    __('Payment methods', 'raptus-wc-bexio-connector'),
                    __('Configure payment conditions for your payment methods', 'raptus-wc-bexio-connector')
                ),
                [$this, 'do_settings_sections_payment_methods'],
                $this->page_hook,
                'normal',
                'core'
            );

            add_meta_box('raptus_wp_bexio_connector_meta_client_status',
                __('Account status', 'raptus-wc-bexio-connector'),
                [$this, 'print_client_status_callback'],
                $this->page_hook,
                'side',
                'core'
            );

            add_meta_box('raptus_wp_bexio_connector_meta_active_modules',
                __('Active modules', 'raptus-wc-bexio-connector'),
                [$this, 'print_active_modules'],
                $this->page_hook,
                'side',
                'core'
            );

            add_meta_box('raptus_wp_bexio_connector_meta_company_info',
                __('Connected company', 'raptus-wc-bexio-connector'),
                [$this, 'print_company_info'],
                $this->page_hook,
                'side',
                'core'
            );
        }

        do_action('raptus_wc_bexio_connector_meta_boxes', $this->page_hook);
    }

    /**
     * Render action buttons.
     */
    public function print_action_buttons() {
        echo '<p>';
        submit_button(null, 'primary', 'submit', false);
        echo '</p>';

        if (Options::get('authorization')) {
            ?><p><a class="button button-primary button-destructive" href="<?php echo $this->get_settings_page_url(array('reset' => 1)); ?>">
                <?php _e('Reset Data', 'raptus-wc-bexio-connector') ?>
            </a></p><?php
        }

        ?>
        <p>
            <a class="button button-link" href="<?php echo SUPPORT_URL ?>" target="_blank">
            <?php _e('Request support', 'raptus-wc-bexio-connector') ?>
            </a>
        </p>
        <?php
    }

    /**
     * Render system information.
     */
    public function print_system_information() {

        $update_available = Utils::is_update_available();
        if ($update_available) {
            echo '<p>';
            echo '<strong>' . __('An update is available!', 'raptus-wc-bexio-connector') . '</strong><br/>';
            echo '<a class="button button-primary" href="' . admin_url('update-core.php') . '">' . __('Update now', 'raptus-wc-bexio-connector') . '</a>';
            echo '</p>';
        }
        $style = $update_available ? 'style="color:red"' : '';
        echo '<ul>';
        echo '<li>' . __('Plugin version', 'raptus-wc-bexio-connector') . ': <span ' . $style . '>' . PLUGIN_VERSION . '</span></li>';

        global $wp_version;
        echo '<li>' . __('WordPress version', 'raptus-wc-bexio-connector') . ': ' . $wp_version . '</li>';

        $woocommerce_version = Utils::get_woocommerce_version();
        if (!$woocommerce_version) {
            $this->show_message('notice-warning', __('WooCommerce is currently not available, please install or activate the plugin.', 'raptus-wc-bexio-connector'));
            $woocommerce_version = __('N/A', 'raptus-wc-bexio-connector');
        }
        echo '<li>' . __('WooCommerce version', 'raptus-wc-bexio-connector') . ': ' . $woocommerce_version . '</li>';
        echo '<li>' . __('Domain', 'raptus-wc-bexio-connector') . ': ' . Utils::get_domain() . '</li>';
        echo '</ul>';
        echo '<em>' . __('Please provide these information to the support team if requested.', 'raptus-wc-bexio-connector') . '</em>';
    }

    /**
     * Render registration settings sections.
     */
    public function do_settings_sections_registration() {
        settings_fields('raptus_wp_bexio_connector_option_group');
        do_settings_sections(self::SETTINGS_PAGE_ID . '_registration');
    }

    /**
     * Render configuration settings sections.
     */
    public function do_settings_sections_configuration() {
        do_settings_sections(self::SETTINGS_PAGE_ID . '_configuration');
    }

    /**
     * Render advanced settings sections.
     */
    public function do_settings_sections_advanced() {
        do_settings_sections(self::SETTINGS_PAGE_ID . '_advanced');
    }

    /**
     * Render sync rules sections.
     */
    public function do_settings_sections_sync_rules() {
        do_settings_sections(self::SETTINGS_PAGE_ID . '_sync_rules');
    }

    /**
     * Render payment methods settings sections.
     */
    public function do_settings_sections_payment_methods() {
        do_settings_sections(self::SETTINGS_PAGE_ID . '_payment_methods');
    }

    /**
     * Process options page request and render correct template.
     */
    public function render_options_page() {
        do_action('raptus_wc_bexio_connector_render_options_page');

        // Handle response from Bexio
        if (!empty($_REQUEST['status'])) {
            $this->client_acccess();
        }

        // Check for client update
        if (!empty($_REQUEST['settings-updated'])) {
            $this->update_client_settings();
            $this->show_message('notice-success', __('Settings updated successfully.', 'raptus-wc-bexio-connector'));
        }

        // Request resend of verification mail and new salt (if no pending verification present)
        if (!empty($_REQUEST['resend_verification'])) {
            $response = Api::instance()->resend_verification_mail();
            $authorization = base64_encode($response['user'] . ':' . $response['salt']);
            Options::set('authorization', $authorization);
        }

        // Authorize client if already registered but not yet authorized
        if (Options::get('authorization') && !Options::get('isauthorized')) {
            $redirect_after = Options::get('authorize_uri') ? false : true;
            $authorize_result = $this->client_authorize();
            if ($redirect_after && $authorize_result !== false)
                $this->reload();
        }

        // Get Bexio settings information when authorized
        if (Options::get('isauthorized')) {
            // Start over if authorization is no longer valid
            if (!empty($_REQUEST['is-authorized'])) {
                $this->show_message('notice-success',
                    __('The autorization was successful! The next step is to configure WooCommerce Bexio Automator using the settings below.', 'raptus-wc-bexio-connector'),
                    sprintf(__('Having trouble with the configuration? Purchase the initial setup package to get help. Check <a href="%s" target="%s">here</a> for more information about this package.', 'raptus-wc-bexio-connector'), $this->get_settings_page_url(array('manage-account' => 1)), '_self')
                );
            }
        } else {
            if (!empty($_REQUEST['error_description']) && !empty($_REQUEST['state'])) {
                if ($_REQUEST['error_description'] == 'The user denied access to your application') {
                    $this->show_message('notice-error',
                        __('You denied the access to your Bexio account; no communication with Bexio is possible.', 'raptus-wc-bexio-connector') . ' ' .
                        __('Please try again.', 'raptus-wc-bexio-connector')
                    );
                } else {
                    $this->show_message('notice-error',
                        __('The required permission could not granted.', 'raptus-wc-bexio-connector') . ' ' .
                        __('Please try again.', 'raptus-wc-bexio-connector')
                    );
                }
            }
        }

        // Check for status request and update client status
        if (!empty($_REQUEST['check_status'])) {
            $this->verify_status();
        }

        // Check for errors
        if (!empty(Api::instance()->get_error())) {
            $this->show_message('error', Api::instance()->get_error_message());
        }

        // Handle upgrade requests
        if (isset($_REQUEST['manage-account'])) {
            if (isset($_REQUEST['upgrade-account-requested'])) {
                $this->send_upgrade_account_email();
            }
            if (isset($_REQUEST['cancel-account'])) {
                if (isset($_REQUEST['cancel-account-confirmed'])) {
                    if ($_POST['cancel-reason'] == "") {
                        // Show notice for missing cancellation reason
                        $this->reload(array('manage-account' => 1, 'cancel-account' => 1, 'missing-cancel-reason' => 1));
                    } else {
                        $this->send_cancel_account_email();
                    }
                }
            }
        }

        // Handle reset requests
        if (isset($_REQUEST['reset'])) {
            if (isset($_REQUEST['reset-requested'])) {
                if (!isset($_REQUEST['reset-confirmed']) || !$_REQUEST['reset-confirmed']) {
                    $this->show_message('notice-warning', __('Please confirm resetting all WooCommerce Bexio Automator data.', 'raptus-wc-bexio-connector'));
                } else {
                    $this->reset_bexio_connector_data();
                }
            }
            if (isset($_REQUEST['reset-authorization-requested'])) {
                if (!isset($_REQUEST['reset-authorization-confirmed']) || !$_REQUEST['reset-authorization-confirmed']) {
                    $this->show_message('notice-warning', __('Please confirm resetting the WooCommerce Bexio Automator authorization.', 'raptus-wc-bexio-connector'));
                } else {
                    $this->reset_bexio_connector_authorization();
                }
            }
        }

        // Show upgrade complete message
        if (isset($_REQUEST['upgrade-account-completed'])) {
            $this->show_message('notice-success', __('Your upgrade request has been sent and will be processed within one weekeday.', 'raptus-wc-bexio-connector'));
            $_SERVER['REQUEST_URI'] = remove_query_arg('upgrade-account-completed', $_SERVER['REQUEST_URI']);
        } else if (isset($_REQUEST['upgrade-account-failed'])) {
            $this->show_message('notice-error', sprintf(__('Your upgrade request could not be sent, please <a href="%s" target="_blank">contact us</a> or try again later.', 'raptus-wc-bexio-connector'), SUPPORT_URL));
            $_SERVER['REQUEST_URI'] = remove_query_arg('upgrade-account-failed', $_SERVER['REQUEST_URI']);
        }

        // Show cancel account message
        if (isset($_REQUEST['cancel-account-completed'])) {
            $this->show_message('notice-success', __('Your cancellation request has been sent and will be processed within one weekeday.', 'raptus-wc-bexio-connector'));
            $_SERVER['REQUEST_URI'] = remove_query_arg('cancel-account-completed', $_SERVER['REQUEST_URI']);
        } else if (isset($_REQUEST['cancel-account-failed'])) {
            $this->show_message('notice-error', sprintf(__('Your cancellation request could not be sent, please <a href="%s" target="_blank">contact us</a> or try again later.', 'raptus-wc-bexio-connector'), SUPPORT_URL));
            $_SERVER['REQUEST_URI'] = remove_query_arg('cancel-account-failed', $_SERVER['REQUEST_URI']);
        }

        // Show reset complete message
        if (isset($_REQUEST['reset-completed'])) {
            $this->show_message('notice-warning', __('All WooCommerce Bexio Automator data has been erased.', 'raptus-wc-bexio-connector'));
            $_SERVER['REQUEST_URI'] = remove_query_arg('reset-completed', $_SERVER['REQUEST_URI']);
        }

        // Show reset authorization complete message
        if (isset($_REQUEST['reset-authorization-completed'])) {
            $this->show_message('notice-warning', __('The WooCommerce Bexio Automator authorization has been erased, please re-authenticate.', 'raptus-wc-bexio-connector'));
            $_SERVER['REQUEST_URI'] = remove_query_arg('reset-authorization-completed', $_SERVER['REQUEST_URI']);
        }

        // Show modules saved message
        if (isset($_REQUEST['modules-updated'])) {
            $modules = isset($_REQUEST['modules']) ? array_keys($_REQUEST['modules']) : [];
            Options::set('modules', $modules);
            Modules::load();
            Api::instance()->log_to_remote('Modules updated', 'info', array('active_modules' => $modules));
            $this->reload(['modules-activated' => 1]);
        }

        if (isset($_REQUEST['modules-activated'])) {
            $this->show_message('notice-success', __('The selected modules have been successfully activated.', 'raptus-wc-bexio-connector'));
        }

        $this->show_template();
    }

    /**
     * Show the appropriate settings template, dependent of the client status.
     */
    protected function show_template() {
        // Preload variables used in authorized templates
        if (Options::get('isauthorized')) {
            $this->verify_status();
            $status = Data::get_client_status();
            $subscription_states = Data::get_subscription_states();
            $subscription_types = Data::get_subscription_types();
            if ($status['type'] != 'trial') {
                unset($subscription_types['trial']);
            }
        }
        $template = $this->get_template();
        $page_hook = $this->page_hook;
        $settings_page_id = self::SETTINGS_PAGE_ID;
        $settings_url = $this->get_settings_page_url();
        $authorization_url = Options::get('authorize_uri');
        $authorized = Options::get('isauthorized');


        if (isset($_GET['manage-modules']) && $_GET['manage-modules']) {
            $modules = Modules::get();
        }

        $base_template = apply_filters('raptus_wc_bexio_connector_settings_base_template', TEMPLATES_DIR . 'index.php');
        if ($args = apply_filters('raptus_wc_bexio_connector_settings_template_args', [])) {
            extract($args);
        }
        if (file_exists($base_template)) {
            include($base_template);
        }
    }

    /**
     * Get the currently active template.
     */
    private function get_template() {
        $template = 'register';
        if (isset($_GET['manage-account']) && $_GET['manage-account']) {
            $template = 'manage-account';
        } else if (isset($_GET['manage-modules']) && $_GET['manage-modules']) {
            $template = 'manage-modules';
        } else if (isset($_GET['reset']) && $_GET['reset']) {
            $template = 'reset';
        } else if (Options::get('authorize_uri') && !Options::get('isauthorized')) {
            $template = 'authorize';
        } else if (Options::get('isblocked')) {
            $template = 'blocked';
        }
        return apply_filters('raptus_wc_bexio_connector_settings_template', "$template.php");
    }

    /**
     * Verify the current client status.
     */
    protected function verify_status() {
        if (Options::get('authorization')) {
            $status = Data::get_client_status();
            if ($status && Options::get('isblocked') && !in_array(Options::get('isblocked'), array('unauthorized_access', 'invalid_salt'))) {
                if (($status['status'] != 'blocked' && $status['status'] != 'cancelled') || ($status['status'] == 'cancelled' && isset($status['is_expired']) && !$status['is_expired'])) {
                    Options::remove('isblocked');
                    $this->reload();
                }
            } elseif (!empty($_REQUEST['check_status']) && $status && Options::get('isblocked') && Options::get('isblocked') == 'invalid_salt') {
                // Clear the client status cache to request the client status anew
                Data::clear_cache('client_status');
                $this->reload();
            }
        } else {
            $this->show_message('error', __('Cannot check the status without proper authorization! Please reset instead', 'raptus-wc-bexio-connector'));
        }
    }

    /**
     * Reload the settings page.
     *
     * @param array The additional query args to append the URL. Optional.
     */
    public function reload($args = array()) {
        if (wp_redirect($this->get_settings_page_url($args))) {
            exit;
        }
    }

    /**
     * Checks the current Bexio connection status.
     * Prints and error message if the status is blocked.
     */
    public function check_status() {
        $status = Data::get_client_status();
        if(!Options::get('isblocked') && ($expires = Options::get('expires')) && strtotime($expires) < time() && (isset($status['type']) && $status['type'] == 'trial') && !isset($_GET['manage-account'])) {
            ?>
            <div class="notice-warning settings-error notice is-dismissible">
                <p><strong>WooCommerce Bexio Automator:</strong> <?php printf(__('Your trial period has expired, please upgrade your account if you wish to continue using this plugin.', 'raptus-wc-bexio-connector'), SUPPORT_URL) ?></p>
                <p><a href="<?php echo $this->get_settings_page_url(array('manage-account' => 1)); ?>" class="button button-primary" type="submit">
                    <?php _e('Manage subscription', 'raptus-wc-bexio-connector'); ?>
                </a></p>
                <button type="button" class="notice-dismiss"><span class="screen-reader-text"><?php _e('Dismiss this notice.', 'woocommerce') ?>.</span></button>
            </div>
            <?php
        }
        // do not display blocked error message if on WooCommerce Bexio Automator options page (this is handled by the blocked template)
        if (isset($_REQUEST['page']) && $_REQUEST['page'] == 'raptus_wp_bexio_connector_options_admin') {
            return;
        }
        if (Options::get('isblocked')) {
            ?>
            <div class="error settings-error notice is-dismissible">
                <p><strong><?php _e('Communication with Bexio is blocked.', 'raptus-wc-bexio-connector'); ?></strong></p>
                <p><?php printf(__('Check your <a href="%s">settings</a> for more information about the issue.', 'raptus-wc-bexio-connector'), add_query_arg('page', 'raptus_wp_bexio_connector_options_admin', admin_url('admin.php'))) ?></p>
                <button type="button" class="notice-dismiss"><span class="screen-reader-text"><?php _e('Dismiss this notice.', 'woocommerce') ?>.</span></button>
            </div>
            <?php
        }
    }

    /**
     * Show admin messages
     */
    protected function show_message($type, $message, $info = false) {
        if ($type == 'error' || $type == 'notice-error') {
            $info = Utils::support_message(true);
        }
        add_settings_error(
            self::SETTINGS_PAGE_ID,
            esc_attr('settings_updated'),
            $message . ($info ? '<br><span style="font-weight: normal">' . $info . '</span>' : ''),
            $type
        );
    }

    /**
     * Process an update client settings request correctly.
     *
     * If a client is already authorized, we perform an update
     * If a client is not yet authorized, we first perform a remote registration,
     * and as a second step store the received authorization data. This authorization
     * is further injected to the underlying remote API to authorize requests.
     */
    protected function update_client_settings() {
        if (Options::get('authorization')) {
            $this->client_update();
        } else {
            if ($this->client_register()) {
                Api::instance()->set_authorization(Options::get('authorization'));
            }
        }
    }

    /**
     * Register a new client
     */
    private function client_register() {
        $data = Options::get();
        $data['domain'] = Utils::get_domain();
        $response = Api::instance()->client_register($data);
        $error = Api::instance()->get_error();
        if (!empty($error)) {
            // Try to update already registered client
            if ($error['status'] == 400 &&
                $error['type'] == 'controller_error' &&
                preg_match('/^Client registration failed: the domain "[^"]+" is already registered, please perform an update instead\.$/', $error['message'])) {
                $response = Api::instance()->client_update($data);
                $error = Api::instance()->get_error();
                if ($response['status'] == 'update_verification') {
                    $this->show_message('notice-warning', sprintf(__('Since there is already a registration for the given domain available, the changes need to be verified by the original registrar. An e-mail has been sent to %s to confirm the registration', 'raptus-wc-bexio-connector'), $response['email']));
                }
            }
            if (!empty($error)) {
                return false;
            }
        }
        $authorization = base64_encode($response['user'] . ':' . $response['salt']);
        Options::set('authorization', $authorization);
        return $response;
    }

    /**
     * Update client information
     */
    private function client_update() {
        $data = Options::get();
        $data['domain'] = Utils::get_domain();
        $response = Api::instance()->client_update($data);
        if (isset($response['status']) && $response['status'] == 'update_verification') {
            // Update the salt in the authorization when a new one is received from the service
            $authorization = base64_encode($response['user'] . ':' . $response['salt']);
            Options::set('authorization', $authorization);
        }
        if (!empty(Api::instance()->get_error())) {
            $this->show_message('error', Api::instance()->get_error_message());
            return false;
        }
        return $response;
    }

    /**
     * Authorize client with OAuth
     */
    private function client_authorize() {
        $redirect_uri = menu_page_url(self::SETTINGS_PAGE_ID, false);

        $response = Api::instance()->client_authenticate(array(
            'redirect_uri' => $redirect_uri
        ));
        // Check for errors
        if (!empty(Api::instance()->get_error())) {
            $this->show_message('error', Api::instance()->get_error_message());
            return false;
        } else {
            Options::set('authorize_uri', $response['url']);
        }
        return $response;
    }

    /**
     * Request client access
     */
    private function client_acccess() {
        $status =  $_REQUEST['status'];

        // Check for error
        if (!empty(Api::instance()->get_error())) {
            $this->show_message('error', Api::instance()->get_error_message());
            return false;
        } else {
            Options::remove('authorize_uri');
            Options::set('isauthorized', 1);
            if (get_option('raptus_wp_bexio_connector_needs_reauthorization'))
                delete_option('raptus_wp_bexio_connector_needs_reauthorization');
            $this->reload(array('is-authorized' => $status ? 1 : 0));
        }
    }

    /**
     * Send the upgrade account email.
     */
    private function send_upgrade_account_email() {
        $upgrade_data = array(
            __('Type', 'raptus-wc-bexio-connector') => $_POST['subscription'],
            __('Comment', 'raptus-wc-bexio-connector') => $_POST['comment'],
        );

        if (Email::instance()->send_upgrade_account_email($upgrade_data)) {
            $this->reload(array('upgrade-account-completed' => 1));
        } else {
            $this->reload(array('upgrade-account-failed' => 1));
        }
    }

    /**
     * Sent the account cancellation email.
     */
    private function send_cancel_account_email() {
        $cancellation_reason = array(
            _x('Reason', 'Cancel account email', 'raptus-wc-bexio-connector') => $this->get_cancellation_reasons($_POST['cancel-reason']),
            _x('Comments', 'Cancel account email', 'raptus-wc-bexio-connector') => $_POST['cancellation-comments']
        );

        if (Email::instance()->send_cancel_account_email($cancellation_reason)) {
            $this->reload(array('cancel-account-completed' => 1));
        } else {
            $this->reload(array('cancel-account-failed' => 1));
        }
    }

    /**
     * Get a list of possible reasons for canceling the account.
     * 
     * @param string $key The optional key for the label to return.
     * @return array The list of cancellation reasons.
     * */
    public function get_cancellation_reasons($key=null) {
        $cancellation_reasons = array(
            'no_longer_needed' => _x('Plugin is no longer needed', 'Account cancellation reasons', 'raptus-wc-bexio-connector'),
            'switching_solution' => _x('Switching to another solution', 'Account cancellation reasons', 'raptus-wc-bexio-connector'),
            'sync_errors' => _x('Plugin had synchronization errors', 'Account cancellation reasons', 'raptus-wc-bexio-connector'),
            'too_expensive' => _x('Plugin is too expensive', 'Account cancellation reasons', 'raptus-wc-bexio-connector'),
            'unusable' => _x('Plugin is unusable', 'Account cancellation reasons', 'raptus-wc-bexio-connector'),
            'other' => _x('Other', 'Account cancellation reasons', 'raptus-wc-bexio-connector')
        );
        if ($key) {
            return $cancellation_reasons[$key];
        }
        return $cancellation_reasons;
    }

    /**
     * Resets all plugin data and redirects.
     */
    private function reset_bexio_connector_data() {
        // Delete wcba-data contents
        Data::delete_wcba_data_contents();
        // Delete options
        Options::reset();
        // Delete stock cache table and stock file
        @unlink($upload_dir['basedir'] . '/stock.json');
        Stock::instance()->drop_stock_table();
        Api::instance()->log_to_remote('Data reset - plugin data was reset by user', 'warning');
        $this->reload(array('reset-completed' => 1));
    }

    /**
     * Resets the authorization data and redirects.
     */
    private function reset_bexio_connector_authorization() {
        Options::remove('isauthorized');
        Api::instance()->log_to_remote('Data reset - bexio authorization was reset by user', 'warning');
        $this->reload(array('reset-authorization-completed' => 1));
    }

    /**
     * Register and add settings
     */
    public function register_settings() {
        $settings_page_id = self::SETTINGS_PAGE_ID;

        register_setting(
            'raptus_wp_bexio_connector_option_group',
            'raptus_wp_bexio_connector_options',
            array($this, 'sanitize')
        );

        $settings_page_id = self::SETTINGS_PAGE_ID . '_registration';
        add_settings_section(
            'raptus_wp_bexio_connector_options_settings',
            '',
            array($this, 'print_section_info'),
            $settings_page_id
        );

        $this->add_settings_field(
            'name',
            __('Name', 'raptus-wc-bexio-connector'),
            array($this, 'name_callback'),
            $settings_page_id,
            'raptus_wp_bexio_connector_options_settings',
            array('class' => 'required-setting')
        );

        $this->add_settings_field(
            'company',
            __('Company', 'raptus-wc-bexio-connector'),
            array($this, 'company_callback'),
            $settings_page_id,
            'raptus_wp_bexio_connector_options_settings',
            array('class' => 'required-setting')
        );

        $this->add_settings_field(
            'email',
            __('E-Mail', 'raptus-wc-bexio-connector'),
            array($this, 'email_callback'),
            $settings_page_id,
            'raptus_wp_bexio_connector_options_settings',
            array('class' => 'required-setting')
        );

        $this->add_settings_field(
            'address',
            __('Address', 'raptus-wc-bexio-connector'),
            array($this, 'address_callback'),
            $settings_page_id,
            'raptus_wp_bexio_connector_options_settings',
            array('class' => 'required-setting')
        );

        $this->add_settings_field(
            'zip',
            __('Zip', 'raptus-wc-bexio-connector'),
            array($this, 'zip_callback'),
            $settings_page_id,
            'raptus_wp_bexio_connector_options_settings',
            array('class' => 'required-setting')
        );

        $this->add_settings_field(
            'city',
            __('City', 'raptus-wc-bexio-connector'),
            array($this, 'city_callback'),
            $settings_page_id,
            'raptus_wp_bexio_connector_options_settings',
            array('class' => 'required-setting')
        );

        $this->add_settings_field(
            'country',
            __('Country', 'raptus-wc-bexio-connector'),
            array($this, 'country_callback'),
            $settings_page_id,
            'raptus_wp_bexio_connector_options_settings',
            array('class' => 'required-setting')
        );

        $this->add_settings_field(
            'domain',
            __('Domain', 'raptus-wc-bexio-connector'),
            array($this, 'domain_callback'),
            $settings_page_id,
            'raptus_wp_bexio_connector_options_settings'
        );

        /**
         * Check if client is authorized
         * and Bexio options can be selected.
         */
        if (Options::get('isauthorized')) {
            $settings_page_id = self::SETTINGS_PAGE_ID . '_configuration';

            add_settings_section(
                'raptus_wp_bexio_connector_options_settings_bexio',
                '',
                array($this, 'print_section_info_bexio'),
                $settings_page_id
            );

            $this->add_settings_field(
                'bexio_user_creator',
                __('User for creating orders', 'raptus-wc-bexio-connector'),
                array($this, 'bexio_user_creator_callback'),
                $settings_page_id,
                'raptus_wp_bexio_connector_options_settings_bexio'
            );

            $this->add_settings_field(
                'bexio_user_owner',
                __('User that owns created orders', 'raptus-wc-bexio-connector'),
                array($this, 'bexio_user_owner_callback'),
                $settings_page_id,
                'raptus_wp_bexio_connector_options_settings_bexio'
            );

            $this->add_settings_field(
               'bexio_document_template',
               __('Document template', 'raptus-wc-bexio-connector'),
               array($this, 'bexio_document_template'),
               $settings_page_id,
               'raptus_wp_bexio_connector_options_settings_bexio'
            );

            $this->add_settings_field(
                'bexio_account_products',
                __('Account for booking product income', 'raptus-wc-bexio-connector'),
                array($this, 'bexio_account_products_callback'),
                $settings_page_id,
                'raptus_wp_bexio_connector_options_settings_bexio',
                array('class' => 'required-setting')
            );

            $this->add_settings_field(
                'bexio_account_shipping',
                __('Account for booking shipping income', 'raptus-wc-bexio-connector'),
                array($this, 'bexio_account_shipping_callback'),
                $settings_page_id,
                'raptus_wp_bexio_connector_options_settings_bexio',
                array('class' => 'required-setting')
            );

            $this->add_settings_field(
                'bexio_account_discount',
                __('Account for booking discount', 'raptus-wc-bexio-connector'),
                array($this, 'bexio_account_discount_callback'),
                $settings_page_id,
                'raptus_wp_bexio_connector_options_settings_bexio'
            );

            $this->add_settings_field(
                'bexio_disable_taxes',
                __('Disable taxes', 'raptus-wc-bexio-connector'),
                array($this, 'bexio_disable_taxes'),
                $settings_page_id,
                'raptus_wp_bexio_connector_options_settings_bexio',
                array('class' => 'required-setting')
            );

           $this->add_settings_field(
               'bexio_tax_rates_table_header',
               $this->get_settings_title_html(__('WooCommerce tax rate', 'raptus-wc-bexio-connector'), __('Tax class', 'raptus-wc-bexio-connector')),
               array($this, 'bexio_tax_rates_table_header'),
               $settings_page_id,
               'raptus_wp_bexio_connector_options_settings_bexio',
               array('tax_key' => 'shipping', 'class' => 'wcba-tax-rate-setting')
           );

           $this->add_settings_field(
               'bexio_tax_shipping',
               __('Shipping costs', 'raptus-wc-bexio-connector'),
               array($this, 'bexio_tax_rates'),
               $settings_page_id,
               'raptus_wp_bexio_connector_options_settings_bexio',
               array('tax_key' => 'shipping', 'class' => 'wcba-table-header wcba-tax-rate-setting required-setting')
           );

            foreach ($this->get_tax_classes() as $key => $tax_class) {
                $key = sanitize_title($tax_class);
                $this->add_settings_field(
                    'bexio_tax_' . $key,
                    $tax_class,
                    //sprintf(__('Tax rate for %s', 'raptus-wc-bexio-connector'), $tax_class),
                    array($this, 'bexio_tax_rates'),
                    $settings_page_id,
                    'raptus_wp_bexio_connector_options_settings_bexio',
                    array('tax_key' => $key, 'class' => 'wcba-table-header wcba-tax-rate-setting required-setting')
               );
            }

            $this->add_settings_field(
                'bexio_tax_rates_table_footer',
                '',
                array($this, 'bexio_tax_rates_table_footer'),
                $settings_page_id,
                'raptus_wp_bexio_connector_options_settings_bexio',
                array('class' => 'wcba-tax-rate-setting required-setting')
            );

            // Optional Bexio Settings
            $settings_page_id = self::SETTINGS_PAGE_ID . '_advanced';
            add_settings_section(
                'raptus_wp_bexio_connector_options_settings_bexio_advanced',
                '',
                array($this, 'print_section_info_bexio_advanced'),
                $settings_page_id
            );

            $this->add_settings_field(
                'bexio_order_title_pattern',
                __('Title for Bexio orders', 'raptus-wc-bexio-connector'),
                array($this, 'bexio_order_title_pattern_callback'),
                $settings_page_id,
                'raptus_wp_bexio_connector_options_settings_bexio_advanced'
            );

            $this->add_settings_field(
                'bexio_failed_order_sync_notification_email',
                __('Error nofication email', 'raptus-wc-bexio-connector'),
                array($this, 'bexio_failed_order_sync_notification_email'),
                $settings_page_id,
                'raptus_wp_bexio_connector_options_settings_bexio_advanced',
                array('class' => 'required-setting')
            );

            $this->add_settings_field(
                'bexio_contact_group',
                __('Bexio contact group', 'raptus-wc-bexio-connector'),
                array($this, 'bexio_contact_group_callback'),
                $settings_page_id,
                'raptus_wp_bexio_connector_options_settings_bexio_advanced'
            );

            $this->add_settings_field(
                'bexio_contact_creation_mode',
                __('Bexio contact creation mode', 'raptus-wc-bexio-connector'),
                array($this, 'bexio_contact_creation_mode'),
                $settings_page_id,
                'raptus_wp_bexio_connector_options_settings_bexio_advanced'
            );

            $this->add_settings_field(
                'bexio_shipping_address_mode',
                __('Bexio shipping address mode', 'raptus-wc-bexio-connector'),
                array($this, 'bexio_shipping_address_mode'),
                $settings_page_id,
                'raptus_wp_bexio_connector_options_settings_bexio_advanced'
            );

            $this->add_settings_field(
                'bexio_display_product_excerpt',
                __('Display product excerpt', 'raptus-wc-bexio-connector'),
                array($this, 'bexio_display_product_excerpt'),
                $settings_page_id,
                'raptus_wp_bexio_connector_options_settings_bexio_advanced'
            );

            $this->add_settings_field(
                'bexio_comments_sync_mode',
                _x('How to sync order comments', 'Comment synchronization mode label', 'raptus-wc-bexio-connector'),
                array($this, 'bexio_comments_sync_mode'),
                $settings_page_id,
                'raptus_wp_bexio_connector_options_settings_bexio_advanced'
            );

            $this->add_settings_field(
                'bexio_append_state_to_city',
                __('Append state to city', 'raptus-wc-bexio-connector'),
                array($this, 'bexio_append_state_to_city'),
                $settings_page_id,
                'raptus_wp_bexio_connector_options_settings_bexio_advanced'
            );

           $this->add_settings_field(
                'bexio_disable_stock_synchronization',
                __('Disable stock synchronization', 'raptus-wc-bexio-connector'),
                array($this, 'bexio_disable_stock_synchronization'),
                $settings_page_id,
                'raptus_wp_bexio_connector_options_settings_bexio_advanced',
                array('class' => 'required-setting')
            );

            $this->add_settings_field(
                'bexio_sync_order_states',
                __('Synchronize order states', 'raptus-wc-bexio-connector'),
                array($this, 'bexio_sync_order_states'),
                $settings_page_id,
                'raptus_wp_bexio_connector_options_settings_bexio_advanced',
                array('class' => 'required-setting')
            );

            $this->add_settings_field(
                'bexio_async_order_creation',
                __('Order synchronization mode', 'raptus-wc-bexio-connector'),
                array($this, 'bexio_async_order_creation'),
                $settings_page_id,
                'raptus_wp_bexio_connector_options_settings_bexio_advanced'
            );

            $this->add_settings_field(
                'bexio_sync_free_shipping',
                __('Synchronize free shipping', 'raptus-wc-bexio-connector'),
                array($this, 'bexio_sync_free_shipping'),
                $settings_page_id,
                'raptus_wp_bexio_connector_options_settings_bexio_advanced'
            );

            $this->add_settings_field(
                'bexio_override_invoice_due_date',
                __('Override invoice due date for online payments', 'raptus-wc-bexio-connector'),
                array($this, 'bexio_override_invoice_due_date'),
                $settings_page_id,
                'raptus_wp_bexio_connector_options_settings_bexio_advanced'
            );

            $this->add_settings_field(
                'bexio_auto_optimize_address_data',
                __('Auto optimize address data', 'raptus-wc-bexio-connector'),
                array($this, 'bexio_auto_optimize_address_data'),
                $settings_page_id,
                'raptus_wp_bexio_connector_options_settings_bexio_advanced'
            );

            $this->add_settings_field(
                'bexio_enable_price_rounding',
                __('Round prices', 'raptus-wc-bexio-connector'),
                array($this, 'bexio_enable_price_rounding'),
                $settings_page_id,
                'raptus_wp_bexio_connector_options_settings_bexio_advanced'
            );

            // Synchronization rules
            $settings_page_id = self::SETTINGS_PAGE_ID . '_sync_rules';
            add_settings_section(
                'raptus_wp_bexio_connector_options_settings_bexio_sync_rules',
                '',
                array($this, 'print_section_info_bexio_sync_rules'),
                $settings_page_id
            );

            $this->add_settings_field(
                'bexio_sync_rules_table_header',
                $this->get_settings_title_html(__('Payment method', 'raptus-wc-bexio-connector'), __('Product type', 'raptus-wc-bexio-connector')),
                array($this, 'bexio_sync_rules_table_header'),
                $settings_page_id,
                'raptus_wp_bexio_connector_options_settings_bexio_sync_rules'
            );

            foreach (Data::get_payment_gateways() as $gateway_id => $gateway) {
                $this->add_settings_field(
                    'bexio_sync_rules_table_' . $gateway_id,
                    $this->get_payment_method_title_html($gateway),
                    $this->get_bexio_sync_rules_table_callback($gateway_id),
                    $settings_page_id,
                    'raptus_wp_bexio_connector_options_settings_bexio_sync_rules',
                    array('class' => 'wcba-table-header')
                );
            }

            $this->add_settings_field(
                'bexio_sync_rules_table_virtual',
                $this->get_settings_title_html(__('Virtual orders', 'raptus-wc-bexio-connector'), __('Orders with only virtual products', 'raptus-wc-bexio-connector')),
                $this->get_bexio_sync_rules_table_callback('virtual', true),
                $settings_page_id,
                'raptus_wp_bexio_connector_options_settings_bexio_sync_rules',
                array('class' => 'wcba-table-header')
            );

            $this->add_settings_field(
                'bexio_sync_rules_table_footer',
                '',
                array($this, 'bexio_sync_rules_table_footer'),
                $settings_page_id,
                'raptus_wp_bexio_connector_options_settings_bexio_sync_rules'
            );

            $this->add_settings_field(
                'bexio_issue_immediate_delivery',
                __('Issue immediate delivery', 'raptus-wc-bexio-connector'),
                array($this, 'bexio_issue_immediate_delivery'),
                $settings_page_id,
                'raptus_wp_bexio_connector_options_settings_bexio_sync_rules'
            );

            $this->add_settings_field(
                'bexio_issue_immediate_invoice',
                __('Issue immediate invoice', 'raptus-wc-bexio-connector'),
                array($this, 'bexio_issue_immediate_invoice'),
                $settings_page_id,
                'raptus_wp_bexio_connector_options_settings_bexio_sync_rules'
            );

            $this->add_settings_field(
                'bexio_sync_free_orders',
                __('Synchronize free orders', 'raptus-wc-bexio-connector'),
                array($this, 'bexio_sync_free_orders'),
                $settings_page_id,
                'raptus_wp_bexio_connector_options_settings_bexio_sync_rules'
            );

            // Payment Methods Settings
            $settings_page_id = self::SETTINGS_PAGE_ID . '_payment_methods';
            add_settings_section(
                'raptus_wp_bexio_connector_options_settings_payment_methods',
                '',
                array($this, 'print_section_info_payment_methods'),
                $settings_page_id
            );

            $this->add_settings_field(
                'bexio_payment_methods_table_header',
                __('Payment method', 'raptus-wc-bexio-connector'),
                array($this, 'bexio_payment_methods_table_header'),
                $settings_page_id,
                'raptus_wp_bexio_connector_options_settings_payment_methods'
            );

            foreach (Data::get_payment_gateways() as $gateway_id => $gateway) {
                $this->add_settings_field(
                    'bexio_payment_methods_table_' . $gateway_id,
                    $this->get_payment_method_title_html($gateway),
                    $this->get_bexio_payment_methods_table_callback($gateway_id),
                    $settings_page_id,
                    'raptus_wp_bexio_connector_options_settings_payment_methods',
                    array('class' => 'wcba-table-header required-setting')
                );
            }
        }
    }

    /**
     * Add a new settings field.
     * Allows prevention of registration of settings fields via filter.
     * Allows adding additional settings after a specific setting via action.
     * */
    private function add_settings_field($id, $title, $callback, $page, $section = 'default', $args = array()) {
        if(apply_filters('raptus_wp_bexio_connector_register_setting/'.$id, true, $page, $section)) {
            add_settings_field(
                $id,
                $title,
                $callback,
                $page,
                $section,
                $args
            );
        }

        do_action('raptus_wp_bexio_connector_after_settings/'.$id, $page, $section);
    }

    private function get_payment_method_title_html($gateway) {
        $method_title = $gateway->get_method_title();
        if ($gateway->get_title() != $method_title) {
            $method_title = $this->get_settings_title_html($method_title, $gateway->get_title());
        }
        return $method_title ?: $gateway->get_title();
    }

    private function get_settings_title_html($title, $subtitle) {
        return $title . '<br><small style="font-weight:300">' . $subtitle . '</small>';
    }

    /**
     * Gets all WooCommerce tax classes.
     *
     * @return array The WooCommerce tax classes.
     */
    private function get_tax_classes() {
        $tax_classes = WC_Tax::get_tax_classes();
        array_unshift($tax_classes, __('Standard', 'woocommerce'));

        $taxes = array();
        foreach ($tax_classes as $tax) {
            $key = sanitize_title($tax);
            $taxes[$key] = $tax;
        }

        return $taxes;
    }

    /**
     * Sanitize each setting field as needed.
     *
     * @param  array $input Contains all settings fields as array keys.
     * @return array The sanitized settings fields.
     */
    public function sanitize($input) {
        $new_input = array('version' => PLUGIN_VERSION);

        if (isset($input['name'])) {
            $new_input['name'] = sanitize_text_field($input['name']);
        }

        if (isset($input['company'])) {
            $new_input['company'] = sanitize_text_field($input['company']);
        }

        if (isset($input['email'])) {
            $new_input['email'] = sanitize_text_field($input['email']);
        }

        if (isset($input['address'])) {
            $new_input['address'] = sanitize_text_field($input['address']);
        }

        if (isset($input['zip'])) {
            $new_input['zip'] = sanitize_text_field($input['zip']);
        }

        if (isset($input['city'])) {
            $new_input['city'] = sanitize_text_field($input['city']);
        }

        if (isset($input['country'])) {
            $new_input['country'] = sanitize_text_field($input['country']);
        }

        if (Options::get('isauthorized') && !isset($_REQUEST['code']) && !isset($_REQUEST['state']) || Options::get('wizard') == 'done') {
            if (isset($input['bexio_order_title_pattern'])) {
                $new_input['bexio_order_title_pattern'] = sanitize_text_field($input['bexio_order_title_pattern']);
            }

            if (isset($input['bexio_user_creator'])) {
                $new_input['bexio_user_creator'] = $input['bexio_user_creator'];
            }

            if (isset($input['bexio_user_owner'])) {
                $new_input['bexio_user_owner'] = $input['bexio_user_owner'];
            }

            if (isset($input['bexio_document_template'])) {
                $new_input['bexio_document_template'] = $input['bexio_document_template'];
            }

            if (isset($input['bexio_account_products'])) {
                $new_input['bexio_account_products'] = $input['bexio_account_products'];
            }

            if (isset($input['bexio_account_shipping'])) {
                $new_input['bexio_account_shipping'] = $input['bexio_account_shipping'];
            }

            if (isset($input['bexio_account_discount'])) {
                $new_input['bexio_account_discount'] = $input['bexio_account_discount'];
            }

            if (isset($input['bexio_tax_shipping'])) {
                $new_input['bexio_tax_shipping'] = $input['bexio_tax_shipping'];
            }

            foreach ($this->get_tax_classes() as $key => $tax_class ) {
                if (isset($input['bexio_tax_' . $key])) {
                    $new_input['bexio_tax_' . $key] = $input['bexio_tax_' . $key];
                }
            }

            if (isset($input['bexio_disable_taxes'])) {
                $new_input['bexio_disable_taxes'] = $input['bexio_disable_taxes'] ? 1 : 0;
            } else {
                $new_input['bexio_disable_taxes'] = 0;
            }

            if (isset($input['bexio_failed_order_sync_notification_email'])) {
                $new_input['bexio_failed_order_sync_notification_email'] = trim($input['bexio_failed_order_sync_notification_email']);
            }

            if (isset($input['bexio_contact_group'])) {
                $new_input['bexio_contact_group'] = $input['bexio_contact_group'];
            }

            if (isset($input['bexio_contact_creation_mode']) && !empty($input['bexio_contact_creation_mode'])) {
                $new_input['bexio_contact_creation_mode'] = $input['bexio_contact_creation_mode'];
            }

            if (isset($input['bexio_shipping_address_mode']) && !empty($input['bexio_shipping_address_mode'])) {
                $new_input['bexio_shipping_address_mode'] = $input['bexio_shipping_address_mode'];
            }

            if (isset($input['bexio_display_product_excerpt'])) {
                $new_input['bexio_display_product_excerpt'] = $input['bexio_display_product_excerpt'] ? 1 : 0;
            } else {
                $new_input['bexio_display_product_excerpt'] = 0;
            }

            if (isset($input['bexio_comments_sync_mode']) && !empty($input['bexio_comments_sync_mode'])) {
                $new_input['bexio_comments_sync_mode'] = $input['bexio_comments_sync_mode'];
            }

            foreach (Data::get_payment_gateways() as $gateway_id => $gateway) {
                foreach (Data::get_sync_types() as $type_id => $type_name) {
                    $settings_id = 'bexio_sync_rules_' . $gateway_id . '_' . $type_id;
                    if (isset($input[$settings_id]) && !empty($input[$settings_id])) {
                        $new_input[$settings_id] = $input[$settings_id];
                    }
                }
            }

            foreach (Data::get_sync_types() as $type_id => $type_name) {
                $settings_id = 'bexio_sync_rules_virtual_' . $type_id;
                if (isset($input[$settings_id]) && !empty($input[$settings_id])) {
                    $new_input[$settings_id] = $input[$settings_id];
                }
            }

            if (isset($input['bexio_issue_immediate_delivery'])) {
                $new_input['bexio_issue_immediate_delivery'] = $input['bexio_issue_immediate_delivery'] ? 1 : 0;
            } else {
                $new_input['bexio_issue_immediate_delivery'] = 0;
            }

            foreach (Data::get_sync_types() as $type_id => $type_name) {
                $settings_id = 'bexio_sync_free_' . $type_id;
                if ($type_id == 'payment')
                    continue;
                if ($type_id == 'order' && (!isset($input[$settings_id]) || empty($input[$settings_id]))) {
                    $new_input[$settings_id] = 0;
                    $new_input['bexio_sync_free_delivery'] = 0;
                    $new_input['bexio_sync_free_invoice'] = 0;
                    break;

                }
                if (isset($input[$settings_id])) {
                    $new_input[$settings_id] = $input[$settings_id] ? 1 : 0;
                } else {
                    $new_input[$settings_id] = 0;
                }
            }

            if (isset($input['bexio_issue_immediate_invoice'])) {
                $new_input['bexio_issue_immediate_invoice'] = $input['bexio_issue_immediate_invoice'] ? 1 : 0;
            } else {
                $new_input['bexio_issue_immediate_invoice'] = 0;
            }

            if (isset($input['bexio_disable_stock_synchronization'])) {
                $new_input['bexio_disable_stock_synchronization'] = $input['bexio_disable_stock_synchronization'] ? 1 : 0;
            } else {
                $new_input['bexio_disable_stock_synchronization'] = 0;
            }

            if (isset($input['bexio_sync_order_states'])) {
                $new_input['bexio_sync_order_states'] = $input['bexio_sync_order_states'] ? 1 : 0;
            } else {
                $new_input['bexio_sync_order_states'] = 0;
            }

            if (isset($input['bexio_async_order_creation']) && !empty($input['bexio_async_order_creation'])) {
                $new_input['bexio_async_order_creation'] = $input['bexio_async_order_creation'];
            }

            if (isset($input['bexio_append_state_to_city'])) {
                $new_input['bexio_append_state_to_city'] = $input['bexio_append_state_to_city'] ? 1 : 0;
            } else {
                $new_input['bexio_append_state_to_city'] = 0;
            }

            if (isset($input['bexio_sync_free_shipping'])) {
                $new_input['bexio_sync_free_shipping'] = $input['bexio_sync_free_shipping'] ? 1 : 0;
            } else {
                $new_input['bexio_sync_free_shipping'] = 0;
            }

            if (isset($input['bexio_override_invoice_due_date'])) {
                $new_input['bexio_override_invoice_due_date'] = $input['bexio_override_invoice_due_date'] ? 1 : 0;
            } else {
                $new_input['bexio_override_invoice_due_date'] = 0;
            }

            if (isset($input['bexio_auto_optimize_address_data'])) {
                $new_input['bexio_auto_optimize_address_data'] = $input['bexio_auto_optimize_address_data'] ? 1 : 0;
            } else {
                $new_input['bexio_auto_optimize_address_data'] = 0;
            }

            if (isset($input['bexio_enable_price_rounding'])) {
                $new_input['bexio_enable_price_rounding'] = $input['bexio_enable_price_rounding'] ? 1 : 0;
            } else {
                $new_input['bexio_enable_price_rounding'] = 0;
            }

            foreach (Data::get_payment_gateways() as $gateway_id => $gateway) {
                foreach (array('bank_account_id', 'payment_type_id', 'payment_terms') as $type) {
                    $settings_id = 'bexio_payment_method_' . $gateway_id . '_' . $type;
                    if (isset($input[$settings_id]) && !empty($input[$settings_id])) {
                        $new_input[$settings_id] = $input[$settings_id];
                    }
                }
            }
        }

        return apply_filters('raptus_wc_bexio_connector_sanitize_settings', $new_input, $input);
    }

    /**
     * Print the registration section text.
     */
    public function print_section_info() {
        echo '<p><em>';
        _e('Enter your settings below:', 'raptus-wc-bexio-connector');
        echo '</em></p>';
    }

    /**
     * Metabox callback to print the client status information.
     */
    public function print_client_status_callback($arg1 = null, $arg2 = null) {
        $this->print_client_status(false);
    }

    /**
     * Print the client status information.
     */
    public function print_client_status($hide_upgrade_button = false) {
        $subscription_types = Data::get_subscription_types();
        $subscription_states = Data::get_subscription_states();
        $status = Data::get_client_status();
        include(TEMPLATES_DIR . 'status.php');

        if (!$hide_upgrade_button) {
            ?>
            <p><a href="<?php echo $this->get_settings_page_url(array('manage-account' => 1)); ?>" class="button button-secondary" type="submit">
                <?php _e('Manage subscription', 'raptus-wc-bexio-connector'); ?>
            </a></p>
            <?php
        }
    }

    /**
     * Print the active modules.
     */
    public function print_active_modules() {

        if (Options::get('authorization')) {
        ?>
            <table class="status-table">

                <?php foreach (Modules::get() as $module) : ?>

                <tr>
                    <th<?= $module->available() ? '' : ' class="disabled"'; ?>><?= $module->get_name(); ?>
                    <?php if (!$module->available()) : ?>
                        <span class="badge"><?= __('Pro', 'raptus-wc-bexio-connector') ?></span>
                    <?php endif; ?></th>
                    <td><div class="form-row switch-pill<?= $module->available() ? '' : ' disabled'; ?>">
                        <input type="checkbox" id="<?= $module->get_id() ?>" name="modules[<?= $module->get_id() ?>]" value="1" <?php disabled(true) ?> <?php checked($module->active()) ?>>
                        <label for="<?= $module->get_id() ?>"></label>
                    </div></td>
                </tr>

                <?php endforeach; ?>

            </table>

            <p>
                <a class="button button-secondary" href="<?php echo $this->get_settings_page_url(array('manage-modules' => 1)); ?>">
                <?php _e('Manage modules', 'raptus-wc-bexio-connector') ?>
                </a>
            </p>
            <?php
        }
    }

    /**
     * Print the connmected company information.
     */
    public function print_company_info() {
        $company = Data::get_client_company();
        echo '<p>';
        if (isset($company['logo_base64']) && !empty($company['logo_base64'])) {
            ?><img src="data:image/png;base64, <?php echo $company['logo_base64']; ?>" alt="Logo" style="max-width:25%; float: left; margin-right: 10px;"/><?php
        }
        echo $company['name'] . '<br/>';
        echo $company['address'] . ' ' . $company['address_nr'] . '<br/>';
        echo $company['postcode'] . ' ' . $company['city'] . '<br/>';
        echo $company['country_name'];
        echo '</p>';
    }

    /**
     * Print the bexio Section text
     */
    public function print_section_info_bexio() {
        echo '<p><em>';
        _e('Select which user, account and taxes should be used:', 'raptus-wc-bexio-connector');
        echo '</em></p>';
    }

    public function name_callback() {
        printf(
            '<input class="regular-text" type="text" id="name" name="raptus_wp_bexio_connector_options[name]" value="%s" required/>',
            esc_attr(Options::get('name'))
       );
    }

    public function company_callback() {
        printf(
            '<input class="regular-text" type="text" id="company" name="raptus_wp_bexio_connector_options[company]" value="%s" required/>',
            esc_attr(Options::get('company'))
       );
    }

    public function email_callback() {
        printf(
            '<input class="regular-text" type="text" id="email" name="raptus_wp_bexio_connector_options[email]" value="%s" required/>',
            esc_attr(Options::get('email'))
       );
    }

    public function address_callback() {
        printf(
            '<input class="regular-text" type="text" id="address" name="raptus_wp_bexio_connector_options[address]" value="%s" required/>',
            esc_attr(Options::get('address'))
       );
    }

    public function zip_callback() {
        printf(
            '<input class="regular-text" type="text" id="zip" name="raptus_wp_bexio_connector_options[zip]" value="%s" required/>',
            esc_attr(Options::get('zip'))
       );
    }

    public function city_callback() {
        printf(
            '<input class="regular-text" type="text" id="city" name="raptus_wp_bexio_connector_options[city]" value="%s" required/>',
            esc_attr(Options::get('city'))
       );
    }

    public function country_callback() {
        $countries = new WC_Countries();
        $selected = Options::get('country');
        if (!$selected) {
            $selected = wc_get_base_location()['country'];
        }
        echo '<select id="country" name="raptus_wp_bexio_connector_options[country]" class="regular-text" class="regular-text">';
        foreach ($countries->get_countries() as $iso => $country) {
            printf('<option value="%s" %s>%s</option>',
                $iso, selected($selected, $iso, false), $country);
        }
        echo "</select>";

    }

    public function domain_callback() {
        printf(
            '<input class="regular-text code" type="text" readonly="readonly" id="domain" name="raptus_wp_bexio_connector_options[domain]" value="%s" required/>',
            esc_attr(Utils::get_domain())
       );
   }

    public function bexio_order_title_pattern_callback() {
        echo '<input class="large-text" type="text" value="' . Options::get('bexio_order_title_pattern')
           . '" id="bexio_order_title_pattern" name="raptus_wp_bexio_connector_options[bexio_order_title_pattern]">';
        echo '<p class="description">';
        echo __('Beware that the maximum length of the title is limited to 80 characters.', 'raptus-wc-bexio-connector') . ' ';
        echo __('You can also make use of template strings within the title in the following format: <code>{key}</code>.', 'raptus-wc-bexio-connector') . ' ';
        echo __('The template strings will be replaced with the according values later on.', 'raptus-wc-bexio-connector') . ' ';
        echo __('The following template strings are currently supported:', 'raptus-wc-bexio-connector') . '<br/>';
        echo '</p>';

        echo '<label for="toggle-placeholder-list">' . __('Toggle placeholder list', 'raptus-wc-bexio-connector') . '</label><input id="toggle-placeholder-list" type="checkbox" class="toggle" target="placeholder-list" value="Show placeholder list">';
        echo '<div class="placeholder-list">';
        echo '<div class="placeholder-row">' . implode('</div><div class="placeholder-row">', $this->get_string_template_placeholders()) . '</div>';
        echo '</div>';
    }

    public function bexio_failed_order_sync_notification_email() {
        echo '<input class="regular-text" type="email" value="' . Options::get('bexio_failed_order_sync_notification_email')
           . '" id="bexio_failed_order_sync_notification_email" name="raptus_wp_bexio_connector_options[bexio_failed_order_sync_notification_email]">';
        echo '<p class="description">';
        echo __('Enter an email address if you wish to receive error notifications via email regarding failed order synchronizations.', 'raptus-wc-bexio-connector') . ' ';
        echo __('Emails will be sent immediately as soon as the error occurs.', 'raptus-wc-bexio-connector') . ' ';
        echo __('Leave this field empty if you do not wish to receive any notifications via email.', 'raptus-wc-bexio-connector');
        echo '</p>';
    }

    /**
     * Get a list of all supported string template placeholders.
     */
    public function get_string_template_placeholders() {
        $placeholders = array(
            'order_id'         => __('The WooCommerce order id', 'raptus-wc-bexio-connector'),
            'order_no'         => __('The WooCommerce order number', 'raptus-wc-bexio-connector'),
            'currency'         => __('The currency of this order', 'raptus-wc-bexio-connector'),
            'top_product'      => __('The name of the most expensive product of the order', 'raptus-wc-bexio-connector'),
            'payment_method'   => __('The payment method of this order as selected by the customer', 'raptus-wc-bexio-connector'),
            'payment_method_title' => __('The payment method title of this order as selected by the customer', 'raptus-wc-bexio-connector'),
            'shipping_method'  => __('The shipping method of this order as selected by the customer', 'raptus-wc-bexio-connector'),

            'shipping' => array(
                'first_name'   => __('The first name of the shipping address', 'raptus-wc-bexio-connector'),
                'last_name'    => __('The last name of the shipping address', 'raptus-wc-bexio-connector'),
                'company'      => __('The company of the shipping address', 'raptus-wc-bexio-connector'),
                'address_1'    => __('The first address line of the shipping address', 'raptus-wc-bexio-connector'),
                'address_2'    => __('The second address line of the shipping address', 'raptus-wc-bexio-connector'),
                'city'         => __('The city of the shipping address', 'raptus-wc-bexio-connector'),
                'state'        => __('The state of the shipping address', 'raptus-wc-bexio-connector'),
                'postcode'     => __('The postcode of the shipping address', 'raptus-wc-bexio-connector'),
                'country_code' => __('The ISO-3 country code of the shipping address', 'raptus-wc-bexio-connector'),
                'country_name' => __('The full country name of the shipping address', 'raptus-wc-bexio-connector'),
                'email'        => __('The email address of the shipping address', 'raptus-wc-bexio-connector'),
                'phone'        => __('The phone number of the shipping address', 'raptus-wc-bexio-connector'),
            ),

            'billing' => array(
                'first_name'   => __('The first name of the billing address', 'raptus-wc-bexio-connector'),
                'last_name'    => __('The last name of the billing address', 'raptus-wc-bexio-connector'),
                'company'      => __('The company of the billing address', 'raptus-wc-bexio-connector'),
                'address_1'    => __('The first address line of the billing address', 'raptus-wc-bexio-connector'),
                'address_2'    => __('The second address line of the billing address', 'raptus-wc-bexio-connector'),
                'city'         => __('The city of the billing address', 'raptus-wc-bexio-connector'),
                'state'        => __('The state of the billing address', 'raptus-wc-bexio-connector'),
                'postcode'     => __('The postcode of the billing address', 'raptus-wc-bexio-connector'),
                'country_code' => __('The ISO-3 country code of the billing address', 'raptus-wc-bexio-connector'),
                'country_name' => __('The full country name of the billing address', 'raptus-wc-bexio-connector'),
                'email'        => __('The email address of the billing address', 'raptus-wc-bexio-connector'),
                'phone'        => __('The phone number of the billing address', 'raptus-wc-bexio-connector'),
            ),

            'meta:meta_key'    => __('Any order meta data; replace "meta_key" with a meta key from the database', 'raptus-wc-bexio-connector'),
        );
        $placeholders = apply_filters('raptus_wc_bexio_connector_order_title_template_placeholders', $placeholders);
        $placeholders = Utils::flatten($placeholders);

        return array_map(function($placeholder) use ($placeholders) {
            return "<code class=\"placeholder\">{{$placeholder}}</code><span class=\"help\">{$placeholders[$placeholder]}</span>";
        }, array_keys($placeholders));
    }

    public function bexio_user_creator_callback() {
        echo '<select id="bexio_user_creator" name="raptus_wp_bexio_connector_options[bexio_user_creator]" class="regular-text">';
        foreach (Data::get_users() as $item) {
            printf('<option value="%s" %s>%s</option>',
                $item['id'], selected(Options::get('bexio_user_creator'), $item['id'], false), $item['firstname'].' '.$item['lastname'].' ('.$item['email'].')');
        }
        echo '</select>';
    }

    public function bexio_user_owner_callback() {
        echo '<select id="bexio_user_owner" name="raptus_wp_bexio_connector_options[bexio_user_owner]" class="regular-text">';
        foreach (Data::get_users() as $item) {
            printf('<option value="%s" %s>%s</option>',
                $item['id'], selected(Options::get('bexio_user_owner'), $item['id'], false), $item['firstname'].' '.$item['lastname'].' ('.$item['email'].')');
        }
        echo '</select>';
    }

    public function bexio_document_template() {
        echo '<select id="bexio_document_template" name="raptus_wp_bexio_connector_options[bexio_document_template]" class="regular-text">';
        printf('<option value="%s" %s>%s</option>',
            'default', selected(Options::get('bexio_document_template'), 'default', false), __('Bexio default', 'raptus-wc-bexio-connector'));
        foreach (Data::get_document_templates() as $item) {
            printf('<option value="%s" %s>%s</option>',
                $item['template_slug'], selected(Options::get('bexio_document_template'), $item['template_slug'], false), $item['name']);
        }
        echo '</select>';
        echo '<p class="description">';
        echo __('Template used to create new Bexio orders, invoices and deliveries', 'raptus-wc-bexio-connector');
        echo '</p>';
    }

    public function bexio_account_products_callback() {
        echo '<select id="bexio_account_products" name="raptus_wp_bexio_connector_options[bexio_account_products]" class="regular-text">';
        foreach (Data::get_accounts() as $item) {
            printf('<option value="%s" %s>%s: %s</option>',
                $item['id'], selected(Options::get('bexio_account_products'), $item['id'], false), $item['account_no'], $item['name']);
        }
        echo '</select>';
        echo '<p class="description">';
        echo __('This account will be used for both, product positions and default positions.', 'raptus-wc-bexio-connector');
        echo '</p>';
    }

    public function bexio_account_shipping_callback() {
        echo '<select id="bexio_account_shipping" name="raptus_wp_bexio_connector_options[bexio_account_shipping]" class="regular-text">';
        foreach (Data::get_accounts() as $item) {
            printf('<option value="%s" %s>%s: %s</option>',
                $item['id'], selected(Options::get('bexio_account_shipping'), $item['id'], false), $item['account_no'], $item['name']);
        }
        echo '</select>';
    }

    public function bexio_account_discount_callback() {
        echo '<select id="bexio_account_discount" name="raptus_wp_bexio_connector_options[bexio_account_discount]" class="regular-text">';
        echo '<option value="0" ' . selected(Options::get('bexio_account_discount'), 0, false) . '>' . __('No booking account') .'</option>';
        foreach (Data::get_accounts() as $item) {
            printf('<option value="%s" %s>%s: %s</option>',
                $item['id'], selected(Options::get('bexio_account_discount'), $item['id'], false), $item['account_no'], $item['name']);
        }
        echo '</select>';
        echo '<p class="description">';
        echo __('If you select a booking account for discounts, standard positions with a negative value will be created in Bexio. Leave it empty to use normal bexio discount positions without booking account.', 'raptus-wc-bexio-connector');
        echo '</p>';
        echo '<small>';
        echo __('ATTENTION: the usage of a default position with negative amount when using mixed (different) VAT rates could lead to incorrect accounting of VAT in your bookkeeping (since the amount can thus be deducted only via 1 tax rate).', 'raptus-wc-bexio-connector');
        echo '</small>';
    }

    public function bexio_tax_rates_table_header() {
        echo $this->get_settings_title_html('<strong>' . __('Bexio tax rate', 'raptus-wc-bexio-connector') . '</strong>', __('VAT code', 'raptus-wc-bexio-connector'));
    }

    public function bexio_tax_rates($args) {
        echo '<select id="bexio_tax_' . $args['tax_key'] . '" name="raptus_wp_bexio_connector_options[bexio_tax_' . $args['tax_key'] . ']" class="regular-text">';
        $types = Data::get_tax_types();
        foreach ($this->get_tax_groups() as $type => $group) {
            echo '<optgroup label="' . $types[$type] . '">';
            foreach ($group as $item) {
                if (isset($item['display_name']) && isset($item['code'])) {
                    printf('<option value="%s" data-code="%s" %s>%s</option>',
                        $item['id'], $item['code'], selected(Options::get('bexio_tax_' . $args['tax_key']), $item['id'], false), $item['display_name']);
                } else {
                    // Backwards compatibility for older data potentially from cache
                    printf('<option value="%s" %s>%s%% (ID: %s)</option>',
                        $item['id'], selected(Options::get('bexio_tax_' . $args['tax_key']), $item['id'], false), $item['value'], $item['id']);
                }
            }
            echo '</optgroup>';
        }
        echo '</select>';
    }

    public function bexio_tax_rates_table_footer() {
        $this->print_tax_notice();
    }

    private function get_tax_groups() {
        $taxes = array();
        foreach (Data::get_taxes() as $item) {
            if (!isset($taxes[$item['type']])) {
                $taxes[$item['type']] = array();
            }
            $taxes[$item['type']][] = $item;
        }
        return $taxes;
    }

    public function bexio_disable_taxes() {
        echo '<label>';
        echo '<input type="checkbox" id="bexio_disable_taxes" name="raptus_wp_bexio_connector_options[bexio_disable_taxes]" '
           . checked(Options::get('bexio_disable_taxes'), true, false) . '>';
        echo  __('Disable taxes', 'raptus-wc-bexio-connector') . '</label>';
        echo '<p class="description">';
        echo __('Disable taxes altogether if you are exempt from taxes.', 'raptus-wc-bexio-connector');
        echo '<br/>';
        echo __('The previously defined tax rates take no effect if you disable taxes.', 'raptus-wc-bexio-connector');
        echo '</p>';
        echo '<br/><hr class="wcba-tax-rate-setting"/>';
        ?>
        <script type="text/javascript">
            jQuery(document).ready( function($) {
                var disable_taxes_cb = function() {
                    if ($('#bexio_disable_taxes').is(':checked')) {
                        $('.wcba-tax-rate-setting').fadeOut();
                    } else {
                        $('.wcba-tax-rate-setting').fadeIn();
                    }
                };
                $('#bexio_disable_taxes').click(disable_taxes_cb);
                disable_taxes_cb();
            });
        </script>
        <?php
    }

    /**
     * Print a notice about the ambiguity of Bexio taxes and how to identify them correctly.
     */
    protected function print_tax_notice() {
        echo '<p class="description">';
        printf(
            __('Bexio tax rates are only identifiable by their unique ID. Please visit your <a href="%s" target="_blank">Bexio tax rates list</a>, click your desired tax rate and retrieve your unique tax ID from the browsers address bar.', 'raptus-wc-bexio-connector'),
            BEXIO_BASE_URL . 'accountingSettings/taxList'
        );
        echo '</p>';
    }

    public function bexio_contact_group_callback() {
        echo '<select id="bexio_contact_group" name="raptus_wp_bexio_connector_options[bexio_contact_group]" class="regular-text">';
        echo '<option value="">' . __('Please select a contact group...', 'raptus-wc-bexio-connector') . '</option>';
        foreach (Data::get_contact_groups() as $item) {
            printf('<option value="%s" %s>%s</option>',
                $item['id'], selected(Options::get('bexio_contact_group'), $item['id'], false), $item['name']
            );
        }
        echo '</select>';
        echo '<p class="description">';
        printf(
            __('Select the contact group to assign to Bexio contacts that are associated with synchronized orders. Please create the according contact group in Bexio first; visit your <a href="%s" target="_blank">Bexio contact groups</a> to do so. Leave the selection empty to disable this functionality.', 'raptus-wc-bexio-connector'),
            BEXIO_BASE_URL . 'kontakt_kategorie/list'
        );
        echo '</p>';
    }

    public function bexio_contact_creation_mode() {
        echo '<select id="bexio_contact_creation_mode" name="raptus_wp_bexio_connector_options[bexio_contact_creation_mode]" class="regular-text">';
        foreach (Data::get_contact_creation_modes() as $item) {
            printf('<option value="%s" %s>%s</option>',
                $item['id'], selected(Options::get('bexio_contact_creation_mode'), $item['id'], false), $item['name']
            );
        }
        echo '</select>';
        echo '<p class="description">';
        echo __('Select the contact creation mode.
                 Flat means that the contact person of company contacts are included in the same contact (default).
                 Nested means that contact persons of company contacts are created separately and associated
                 with the company via a contact relation (recommended).', 'raptus-wc-bexio-connector');
        echo '</p>';
    }

    public function bexio_shipping_address_mode() {
        echo '<select id="bexio_shipping_address_mode" name="raptus_wp_bexio_connector_options[bexio_shipping_address_mode]" class="regular-text">';
        foreach (Data::get_shipping_address_modes() as $item) {
            printf('<option value="%s" %s>%s</option>',
                $item['id'], selected(Options::get('bexio_shipping_address_mode'), $item['id'], false), $item['name']
            );
        }
        echo '</select>';
        echo '<p class="description">';
        echo sprintf(__('Select synchronization mode for the shipping address.
                 Regular means that no explicit additional address will be used, if the checkout address matches the address of the Bexio contact.
                 Force additional address means that for the shipping address, an additional address will always be set, even if it is the same as the address of the Bexio contact itself. Use this mode if you have marked additional addresses as "address for invoice" in Bexio, to prevent problems. For more information about this subject, see the <a href="%s" target="_blank">FAQ</a>.', 'raptus-wc-bexio-connector'), 'https://helpcenter.bexio-shop.ch/hc/de/articles/15139761562129');
        echo '</p>';
    }

    /**
     * Print the advanced Bexio section text.
     */
    public function print_section_info_bexio_advanced() {
        echo '<p><em>';
        _e('Section for advanced Bexio settings. Please change these settings only if your familiar with how Bexio handles orders.', 'raptus-wc-bexio-connector');
        echo ' ';
        _e('Use this section to configure which data you wish to synchronize with bexio.', 'raptus-wc-bexio-connector');
        echo '</em></p>';
    }

    public function bexio_display_product_excerpt() {
        echo '<label>';
        echo '<input type="checkbox" id="bexio_display_product_excerpt" name="raptus_wp_bexio_connector_options[bexio_display_product_excerpt]" '
           . checked(Options::get('bexio_display_product_excerpt'), true, false) . '>';
        echo __('Display product excerpt', 'raptus-wc-bexio-connector') . '</label>';
        echo '<p class="description">';
        echo __('Display or hide the product excerpt for order item positions within Bexio orders.', 'raptus-wc-bexio-connector');
        echo '</p>';
    }

    public function bexio_comments_sync_mode() {
        echo '<select id="bexio_comments_sync_mode" name="raptus_wp_bexio_connector_options[bexio_comments_sync_mode]" class="regular-text">';
        foreach (Data::get_comment_sync_modes() as $item) {
            printf('<option value="%s" %s>%s</option>',
                $item['id'], selected(Options::get('bexio_comments_sync_mode'), $item['id'], false), $item['name']
            );
        }
        echo '</select>';
        echo '<p class="description">';
        echo _x('Select the order comments synchronization mode.
                 Default means that comments get synchronized as comments on the bexio order.
                 As position means that comments get synced as text positions on the bexio order.', 'Comment synchronization mode description', 'raptus-wc-bexio-connector');
        echo '</p>';
    }

    public function bexio_product_includetax() {
        echo '<p class="description">';
        echo sprintf(__('This option has been removed. Please use the WooCommerce option "%s" to control whether prices are displayed inclusive tax or not.', 'raptus-wc-bexio-connector'),
                     '<a href="' . admin_url('admin.php') . '?page=wc-settings&tab=tax' . '">' . __('Display prices during cart and checkout', 'woocommerce') . '</a>');
        echo '</p>';
    }

    /**
     * Print the Bexio sync rules Bexio section text.
     */
    public function print_section_info_bexio_sync_rules() {
        echo '<p><em>';
        echo __('Settings to control the synchronization of Bexio orders, invoices, deliveries and payments.', 'raptus-wc-bexio-connector');
        echo ' ' . __('Use this section to configure when to execute the synchronization. Please only change these settings if you know what you are doing.', 'raptus-wc-bexio-connector');
        echo ' ' . sprintf(__('If you have any questions, please consult our <a href="%s" target="_blank">automation guide</a> to correctly configure the synchronization rules.', 'raptus-wc-bexio-connector'), 'https://helpcenter.bexio-shop.ch/hc/de/articles/9606928850961');
        echo '</em></p>';
    }

    public function bexio_sync_rules_table_footer() {
        echo '<hr/>';
    }

    public function bexio_sync_rules_table_header() {
        $sync_types = Data::get_sync_types();
        echo '<table class="wcba-settings-table">';
        echo '<thead>';
        echo '<tr>';
        foreach ($sync_types as $type_id => $type_name) {
            echo '<th>';
            echo $this->get_settings_title_html($type_name, sprintf(__('When to create a Bexio %s', 'raptus-wc-bexio-connector'), $type_name));
            echo '</th>';
        }
        echo '</tr>';
        echo '</thead>';
        echo '</table>';
    }

    private function get_bexio_sync_rules_table_callback($gateway_id, $default = false) {
        return function() use ($gateway_id, $default) {
            echo '<table class="wcba-settings-table">';
            echo '<tbody>';
            echo '<tr>';
            foreach (Data::get_sync_types() as $type_id => $type_name) {
                echo '<td>';
                $id = 'bexio_sync_rules_' . $gateway_id . '_' . $type_id;
                echo '<select id="' . $id . '" name="raptus_wp_bexio_connector_options[' . $id . ']" class="all-options">';
                if ($default) {
                    printf('<option value="%s" %s>%s</option>',
                        'default', selected(Options::get($id), 'default', false), __('Default', 'raptus-wc-bexio-connector'));
                }
                printf('<option value="%s" %s>%s</option>',
                    'never', selected(Options::get($id), 'never', false), __('Never', 'raptus-wc-bexio-connector'));
                foreach (Data::get_order_states() as $status_id => $status_name) {
                    printf('<option value="%s" %s>%s</option>',
                        $status_id, selected(Options::get($id), $status_id, false), $status_name);
                }
                echo '</select>';
                echo '</td>';
            }
            echo '</tr>';
            echo '</tbody>';
            echo '</table>';
        };
    }

    /*public function bexio_create_immediate_delivery() {
        echo '<input type="checkbox" id="bexio_create_immediate_delivery" name="raptus_wp_bexio_connector_options[bexio_create_immediate_delivery]" '
           . checked(Options::get('bexio_create_immediate_delivery'), true, false) . '>';
        echo '<p class="description">';
        echo __('Whether or not to create Bexio deliveries immediately after the order creation.', 'raptus-wc-bexio-connector');
        echo '</p>';
    }*/

    public function bexio_issue_immediate_delivery() {
        echo '<label>';
        echo '<input type="checkbox" id="bexio_issue_immediate_delivery" name="raptus_wp_bexio_connector_options[bexio_issue_immediate_delivery]" '
           . checked(Options::get('bexio_issue_immediate_delivery'), true, false) . '>';
        echo __('Issue immediate delivery', 'raptus-wc-bexio-connector') . '</label>';
        echo '<p class="description">';
        echo __('Whether or not to issue Bexio deliveries immediately after their creation (mark it as completed).
                 Note that the stock within Bexio does not get visually updated, until a delivery has been explicitly issued.
                 Prior to the delivery issuing, the stock is just reserved for the created order.
                 However, this has no influence to the WooCommerce stock, since always the available stock form Bexio is used.
                 The available stock is the full stock minus the reserved stock.', 'raptus-wc-bexio-connector');
        echo '</p>';
    }

    /*public function bexio_create_immediate_invoice() {
        echo '<input type="checkbox" id="bexio_create_immediate_invoice" name="raptus_wp_bexio_connector_options[bexio_create_immediate_invoice]" '
           . checked(Options::get('bexio_create_immediate_invoice'), true, false) . '>';
        echo '<p class="description">';
        echo __('Whether or not to create Bexio invoices immediately after the order creation.', 'raptus-wc-bexio-connector');
        echo '</p>';
    }*/

    public function bexio_issue_immediate_invoice() {
        echo '<label>';
        echo '<input type="checkbox" id="bexio_issue_immediate_invoice" name="raptus_wp_bexio_connector_options[bexio_issue_immediate_invoice]" '
           . checked(Options::get('bexio_issue_immediate_invoice'), true, false) . '>';
        echo __('Issue immediate invoice', 'raptus-wc-bexio-connector') . '</label>';
        echo '<p class="description">';
        echo __('Whether or not to issue Bexio invoices immediately after their creation (mark it as open).', 'raptus-wc-bexio-connector');
        echo '</p>';
    }

    /*public function bexio_pay_immediate_invoice() {
        echo '<input type="checkbox" id="bexio_pay_immediate_invoice" name="raptus_wp_bexio_connector_options[bexio_pay_immediate_invoice]" '
           . checked(Options::get('bexio_pay_immediate_invoice'), true, false) . '>';
        echo '<p class="description">';
        echo __('Whether or not to pay Bexio invoices immediately after their creation.
                 Payments will only be made if the order already has been paid for via online payment.
                 Furthermore it is only possible to make a payment if the invoice has already been issued.', 'raptus-wc-bexio-connector');
        echo '</p>';
    }*/

    public function bexio_sync_free_orders() {
        foreach (Data::get_sync_types() as $type_id => $type_name) {
            if ($type_id == 'payment') continue;
            echo '<label class="wcba-label">';
            echo '<input type="checkbox" id="bexio_sync_free_' . $type_id . '" name="raptus_wp_bexio_connector_options[bexio_sync_free_' . $type_id . ']" '
            . checked(Options::get('bexio_sync_free_' . $type_id), true, false) . '>';
            echo sprintf(__('Synchronize free %s', 'raptus-wc-bexio-connector'), $type_name) . '</label>';
        }
        echo '<p class="description">';
        echo __('Configure which data to synchronize for free orders (order total is 0.00). By default invoices are not synchronized, since a 0-payment is unnecessary.', 'raptus-wc-bexio-connector');
        echo '</p>';
        ?>
        <script type="text/javascript">
            jQuery(document).ready( function($) {
                var free_order_cb = function() {
                    $('#bexio_sync_free_delivery, #bexio_sync_free_invoice').prop('disabled', !$('#bexio_sync_free_order').is(':checked'));
                };
                $('#bexio_sync_free_order').click(free_order_cb);
                free_order_cb();
            });
        </script>
        <?php
    }

    public function bexio_disable_stock_synchronization() {
        echo '<label>';
        echo '<input type="checkbox" id="bexio_disable_stock_synchronization" name="raptus_wp_bexio_connector_options[bexio_disable_stock_synchronization]" '
           . checked(Options::get('bexio_disable_stock_synchronization'), true, false) . '>';
        echo __('Disable stock synchronization', 'raptus-wc-bexio-connector') . '</label>';
        echo '<p class="description">';
        echo __('Disable the stock synchronization if you do not wish to pull stock information from Bexio.', 'raptus-wc-bexio-connector');
        echo '<br/>';
        echo __('Note that stock synchronization only works for products that have a matching product in Bexio (matching SKU number).', 'raptus-wc-bexio-connector');
        echo '</p>';
    }

    public function bexio_sync_order_states() {
        echo '<label>';
        echo '<input type="checkbox" id="bexio_sync_order_states" name="raptus_wp_bexio_connector_options[bexio_sync_order_states]" '
           . checked(Options::get('bexio_sync_order_states'), true, false) . '>';
        echo __('Synchronize order states', 'raptus-wc-bexio-connector') . '</label>';
        echo '<p class="description">';
        echo __('Enable the order status synchronization if you wish to pull order status information from Bexio.
                 If enabled, the order status from Bexio will be periodically fetched.
                 Any status transitions of Bexio orders that have been created by WooCommerce Bexio Automator,
                 will be applied to the corresponding WooCommerce order.', 'raptus-wc-bexio-connector');
        echo '</p>';
    }

    public function bexio_async_order_creation() {
        echo '<select id="bexio_async_order_creation" name="raptus_wp_bexio_connector_options[bexio_async_order_creation]" class="regular-text">';
        foreach (Data::get_order_creation_modes() as $item) {
            printf('<option value="%s" %s>%s</option>',
                $item['id'], selected(Options::get('bexio_async_order_creation'), $item['id'], false), $item['name']
            );
        }
        echo '</select>';
        echo '<p class="description">';
        echo __('Select the order synchronization mode.
                 Regular means that the order is synchronized to bexio on creation (default).
                 Asynchronous means that orders are synchronized to bexio once per hour.', 'raptus-wc-bexio-connector');
        echo '</p>';
    }

    public function bexio_append_state_to_city() {
        echo '<label>';
        echo '<input type="checkbox" id="bexio_append_state_to_city" name="raptus_wp_bexio_connector_options[bexio_append_state_to_city]" '
           . checked(Options::get('bexio_append_state_to_city'), true, false) . '>';
        echo __('Append state to city', 'raptus-wc-bexio-connector') . '</label>';
        echo '<p class="description">';
        echo __('There is no dedicated field within Bexio to store the state, therefore you have the option to append it after the city instead, separated by a comma.', 'raptus-wc-bexio-connector');
        echo '</p>';
    }

    public function bexio_sync_free_shipping() {
        echo '<label>';
        echo '<input type="checkbox" id="bexio_sync_free_shipping" name="raptus_wp_bexio_connector_options[bexio_sync_free_shipping]" '
           . checked(Options::get('bexio_sync_free_shipping'), true, false) . '>';
        echo __('Synchronize free shipping', 'raptus-wc-bexio-connector') . '</label>';
        echo '<p class="description">';
        echo __('Whether free shipping should be synchronized or not. If enabled a default position will always be created in the Bexio order for the shipping costs, with the shpping method title and the price (which might be 0.00).', 'raptus-wc-bexio-connector');
        echo '</p>';
    }

    public function bexio_override_invoice_due_date() {
        echo '<label>';
        echo '<input type="checkbox" id="bexio_override_invoice_due_date" name="raptus_wp_bexio_connector_options[bexio_override_invoice_due_date]" '
           . checked(Options::get('bexio_override_invoice_due_date'), true, false) . '>';
        echo __('Override invoice due date for online payments', 'raptus-wc-bexio-connector') . '</label>';
        echo '<p class="description">';
        echo __('If this option is activated, the invoice date "payable by" is overwritten with the date of creation, provided the invoice has already been paid when it was created in Bexio.', 'raptus-wc-bexio-connector');
        echo '</p>';
    }

    public function bexio_auto_optimize_address_data() {
        echo '<label>';
        echo '<input type="checkbox" id="bexio_auto_optimize_address_data" name="raptus_wp_bexio_connector_options[bexio_auto_optimize_address_data]" '
           . checked(Options::get('bexio_auto_optimize_address_data'), true, false) . '>';
        echo __('Auto optimize address data', 'raptus-wc-bexio-connector') . '</label>';
        echo '<p class="description">';
        echo __('If this option is activated, the address data is automatically optizmed to a consistent format. For example the the email is converted to lowercase, names to titlecase and country codes to uppercase. Existing Bexio addresses will not be modified.', 'raptus-wc-bexio-connector');
        echo '</p>';
    }

    public function bexio_enable_price_rounding(){
        echo '<label>';
        echo '<input type="checkbox" id="bexio_enable_price_rounding" name="raptus_wp_bexio_connector_options[bexio_enable_price_rounding]" '
           . checked(Options::get('bexio_enable_price_rounding'), true, false) . '>';
        echo __('Round prices', 'raptus-wc-bexio-connector') . '</label>';
        echo '<p class="description">';
        echo __('Enable price rounding for all order positions (regular prices, discount prices, shipping prices) sent to Bexio. Rounding precision is based on the setting in WooCommerce.', 'raptus-wc-bexio-connector');
        echo '</p>';
    }

    /**
     * Print the payment methods matrix section text.
     */
    public function print_section_info_payment_methods() {
        echo '<p><em>';
        _e('Use this configuration section to link your WooCommerce payment methods to Bexio accounts and payment types.', 'raptus-wc-bexio-connector');
        echo '</em></p>';
    }

    public function bexio_payment_methods_table_header() {
        echo '<table class="wcba-settings-table">';
        echo '  <thead>';
        echo '    <tr>';
        echo '      <th>' . $this->get_settings_title_html(__('Bank account', 'raptus-wc-bexio-connector'), __('Bexio bank account for orders, invoices & deliveries', 'raptus-wc-bexio-connector')) . '</th>';
        echo '      <th>' . $this->get_settings_title_html(__('Payment type', 'raptus-wc-bexio-connector'), __('Bexio payment type for orders, invoices & deliveries', 'raptus-wc-bexio-connector')) . '</th>';
        echo '      <th>' . $this->get_settings_title_html(__('Payment terms', 'raptus-wc-bexio-connector'), __('Override Bexio payment terms for orders & invoices', 'raptus-wc-bexio-connector')) . '</th>';
        echo '    </tr>';
        echo '  </thead>';
        echo '</table>';
    }

    private function get_bexio_payment_methods_table_callback($gateway_id, $default = false) {
        return function() use ($gateway_id, $default) {
            echo '<table class="wcba-settings-table">';
            echo '  <tbody>';
            echo '    <tr class="required-setting">';
            echo '      <td class="required-setting">';
            $this->bexio_payment_method_bank_accounts($gateway_id);
            echo '      </td>';
            echo '      <td>';
            $this->bexio_payment_method_payment_types($gateway_id);
            echo '      </td>';
            echo '      <td>';
            $this->bexio_payment_method_payment_terms($gateway_id);
            echo '      </td>';
            echo '    </tr>';
            echo '  </tbody>';
            echo '</table>';
        };
    }

    private function bexio_payment_method_bank_accounts($gateway_id) {
        $id = 'bexio_payment_method_' . $gateway_id . '_bank_account_id';
        echo '<select id="' . $id . '" name="raptus_wp_bexio_connector_options[' . $id . ']" class="regular-text">';
        foreach (Data::get_bank_accounts() as $item) {
            printf('<option value="%s" %s>%s</option>',
                $item['id'], selected(Options::get($id), $item['id'], false), $item['name']);
        }
        echo '</select>';
    }

    private function bexio_payment_method_payment_types($gateway_id) {
        $id = 'bexio_payment_method_' . $gateway_id . '_payment_type_id';
        echo '<select id="' . $id . '" name="raptus_wp_bexio_connector_options[' . $id . ']" class="regular-text">';
        foreach (Data::get_payment_types() as $item) {
            printf('<option value="%s" %s>%s</option>',
                $item['id'], selected(Options::get($id), $item['id'], false), $item['name']);
        }
        echo '</select>';
    }

    private function bexio_payment_method_payment_terms($gateway_id) {
        $id = 'bexio_payment_method_' . $gateway_id . '_payment_terms';
        echo '<textarea id="' . $id . '" name="raptus_wp_bexio_connector_options[' . $id . ']" class="regular-text">';
        echo Options::get($id);
        echo '</textarea>';
    }
}
