<?php

// Namespace declaration
namespace Raptus\WC\BexioConnector;

// Exit if accessed directly
defined('ABSPATH') or die();

/**
 * The stock module of this plugin.
 */
class Stock {

    /**
     * The default number of seconds to cache stock information on products
     * (6 hours).
     * @var integer
     */
    const DEFAULT_STOCK_EXPIRATION_TIME = 60 * 60 * 6;

    const STOCK_CACHE_TABLE_EXISTS = 'rwcbc_stock_cache_table_exists';
    const CACHE_TABLE_NAME = 'rwcbc_stock';

    /**
     * Whether the cache has already been updated once.
     * @var boolean
     */
    private $cache_updated = false;

    /**
     * Cache update threshold
     * @var integer
     */
    private $cache_update_threshold;

    /**
     * The requested stock information.
     * @var array
     */
    private $stocks;

    /**
     * The stock cache file.
     * @var string
     */
    private $stock_cache_file;

    /**
     * The time the stock expires.
     * @var integer
     * */
    private $stock_expiration_time;

    /**
     * Use Singleton trait to disallow multiple instances of this class.
     */
    use Singleton;

    /**
     * Constructs a new instance of the this class.
     * Registers all required filter and action hooks.
     */
    protected function __construct() {
        if ($this->is_stock_sync_enabled()) {
            // Check quantity before creating a new order.
            add_action('woocommerce_after_checkout_validation', array($this, 'validate_checkout'), 10);

            // Remove cache time upon saving of a product to enforce a stock update
            add_action('save_post', array($this, 'unset_cache_time'));

            // Ensure compatiblilty with MDM
            add_filter('mdmtl_product_get_stock', array($this, 'get_stock_quantity'), 10, 2);

            if (Utils::woocommerce_version_check('3.0')) {
                add_filter('woocommerce_product_get_stock_quantity', array($this, 'get_stock_quantity'), 10, 2);
                // Ensure that the stock gets updated, even if the stock quantity is not explicitly requested
                add_filter('woocommerce_get_stock_html', array($this, 'get_stock_html'), 10, 2);
            } else {
                add_filter('woocommerce_get_stock_quantity', array($this, 'get_stock_quantity'), 10, 2);
            }
        }

        $upload_dir = wp_upload_dir();
        $this->stock_cache_file = $upload_dir['basedir'] . '/stock.json';

        if (!get_option(self::STOCK_CACHE_TABLE_EXISTS)) {
            add_action('init', [$this, 'migrate_stock'], 1000);
        }

        $this->stock_expiration_time = defined('BEXIO_CONNECTOR_STOCK_EXPIRATION_TIME') ?
            intval(BEXIO_CONNECTOR_STOCK_EXPIRATION_TIME) : self::DEFAULT_STOCK_EXPIRATION_TIME;
    }

    /**
     * Check quantity before creating a new order.
     * Raise a WooCommerce error notice if an item is out of stock.
     */
    public function validate_checkout($data = null, $errors = null) {
        $items = array();
        foreach (WC()->cart->get_cart() as $key => $values) {
            $product = $values['data'];
            $stock_managing_product = $this->get_stock_managing_product($product);
            if ($stock_managing_product) {
                if(empty($stock_managing_product->get_sku())) {
                    // If the SKU of the product with stock management enabled is empty, skip syncing stock with Bexio
                    continue;
                }
                $items[$stock_managing_product->get_sku()] = array(
                    'id' => $stock_managing_product->get_id(),
                    'sku' => $stock_managing_product->get_sku(),
                    'name' => $product->get_title(),
                    'quantity' => $values['quantity']
                );
            }
        }

        // Add error when product is out of stock
        if (!empty($items)) {
            $outofstock = $this->update_cart_stock($items);
            if (count($outofstock) > 0) {
                wc_add_notice(__('The following products are out of stock: ', 'raptus-wc-bexio-connector') . implode($outofstock, ', '), 'error');
            }

            /*
            Because after checkout stock cache value corresponds to value fetched during
            checkout, after a completed checkout, that value is 'old'. Therefore, we set
            the cachetime to an expired timestamp so that when next shown / synced, the current
            stock gets fetched from Bexio. Like that, it does not matter that after a completed
            checkout, the stock cache value is higher than actual stock in Bexio / WooCommerce.
            */
            $skus_in_cart = array_keys($items);
            foreach ($skus_in_cart as $sku) {
                $this->update_cache_time($sku, time() - $this->stock_expiration_time);
            }
        }
    }

    /**
     * Check stock with Bexio API.
     *
     * @param  string $html The current stock HTML.
     * @param  array \WC_Product $product The product instance.
     * @return string The unmodified stock HTML.
     */
    public function get_stock_html($html, $product) {
        $amount = $product->get_stock_quantity();
        $this->get_stock_quantity($amount, $product);
        return $html;
    }

    /**
     * Check stock with Bexio API.
     *
     * @param  array int $amount The current stock amount.
     * @param  array \WC_Product $product The product instance.
     * @return string The new stock amount from bexio (if available).
     */
    public function get_stock_quantity($amount, $product) {
        // Check if SKU is set
        if (empty($product) || empty($product->get_sku()) || !$product->get_manage_stock()) {
            return $amount;
        }

        // Check if the cache is still valid or if an update was already made
        if (!$this->is_cache_expired($product->get_sku()) || !$this->needs_stock_update()) {
            return $amount;
        }

        $product_ids = [$product->get_id()];

        global $wp_query;

        // Also check if there are other products whose stock needs to be refreshed (to reduce requests)
        if (!empty($wp_query->posts)) {
            foreach ($wp_query->posts as $post) {
                if ($post->post_type != 'product' || $post->ID == $product->get_id()) {
                    continue;
                }

                /*
                Skip checking cachetime (will be done in `iter_stock()` and via `maybe_update_stock_cache()`
                in `query_skus_and_cache_time()`).

                Empty SKUs will be checked for in `get_non_empty_skus_by_id()` and `iter_stock()`.

                Stock management will be checked for in `update_stock()`.
                */
                $product_ids[] = $post->ID;
            }
        }
        $skus = $this->get_non_empty_skus_by_id($product_ids);

        foreach ($this->iter_stock($skus) as $sku => $row) {
            $prod_id = array_search($sku, $skus);
            if (!$prod_id) {
                // If this SKU does not exist in the iterated array, skip it
                continue;
            }

            $prod = wc_get_product($prod_id);
            if ($prod) {
                $this->update_stock($row->stock, $prod);

                if ($prod_id == $product->get_id()) {
                    $amount = $row->stock;
                }
            }

            unset($skus[$prod_id]);
        }

        return $amount;
    }

    /**
     * Update the stock of the given product.
     *
     * @param  array int $amount The new stock amount.
     * @param  array \WC_Product $product The product instance.
     */
    public function update_stock($amount, $product) {
        // If the product has stock management disabled, or an empty SKU, do nothing
        if (!$product->managing_stock() || empty($product->get_sku())) {
            return;
        }
        update_post_meta($product->get_id(), '_stock_info', $this->get_stock_info($amount, $product));
        $stock_status = apply_filters('raptus_wc_bexio_connector_update_stock',
            $amount > 0 ? 'instock' : ($product->backorders_allowed() ? 'onbackorder' : 'outofstock'),
            $product,
            $amount
        );

        if (!Utils::woocommerce_version_check('3.0')) {
            remove_filter('woocommerce_get_stock_quantity', array($this, 'get_stock_quantity'), 10);
        } else {
            remove_filter('woocommerce_product_get_stock_quantity', array($this, 'get_stock_quantity'), 10);
        }

        if ($stock_status != $product->get_stock_status()) {
            wc_update_product_stock_status($product->get_id(), $stock_status);
        }

        wc_update_product_stock($product->get_id(), $amount);

        if (!Utils::woocommerce_version_check('3.0')) {
            add_filter('woocommerce_get_stock_quantity', array($this, 'get_stock_quantity'), 10, 2);
        } else {
            add_filter('woocommerce_product_get_stock_quantity', array($this, 'get_stock_quantity'), 10, 2);
        }
    }

    /**
     * Get stock information for the given product based on the specified stock amount.
     *
     * @param  array int $stock The stock amount.
     * @param  array \WC_Product $product The product instance.
     * @return string The stock information of the given product.
     */
    private function get_stock_info($stock, $product) {
        static $notify_no_stock_amount, $notify_low_stock_amount;
        isset($notify_no_stock_amount) || $notify_no_stock_amount = get_option('woocommerce_notify_no_stock_amount');
        isset($notify_low_stock_amount) || $notify_low_stock_amount = get_option('woocommerce_notify_low_stock_amount');

        $stock_info = 'normal';
        if ($stock <= $notify_no_stock_amount) {
            if ($product->backorders_allowed()) {
                $stock_info = 'orderable';
            } else {
                $stock_info = 'unavailable';
            }
        } else if ($stock <= $notify_low_stock_amount) {
            $stock_info = 'low';
        }
        return $stock_info;
    }

    /**
     * Get a list off all SKUs which are not empty from the product meta lookup table.
     *
     * @param array $product_ids The array of product IDs to get the SKUs for.
     * @return array The list of SKUs (excluding empty ones).
     * */
    private function get_non_empty_skus_by_id($product_ids = []) {
        global $wpdb;
        $skus = $wpdb->get_results("
            SELECT product_id, sku FROM {$wpdb->wc_product_meta_lookup}
            WHERE product_id IN (".join(',', $product_ids).")
            AND sku <> ''
        ", OBJECT_K);
        $skus = array_column($skus, 'sku', 'product_id');
        return $skus;
    }

    /**
     * Get a list of all SKUs to get stock for. Will exclude SKUs that are in our cache table
     * but still have a valid cachetime, unless force-requesting the SKUs. If no SKUs are
     * specified, gets all non-empty SKUs by default.
     *
     * @param array $skus The SKUs to get.
     * @return array The list of SKUs to get stock for.
     * */
    private function query_skus_and_cache_time($skus = [], $validate_cache = true) {
        global $wpdb;
        $params = [];
        $query = "SELECT m.sku FROM {$wpdb->wc_product_meta_lookup} AS m
        LEFT JOIN {$wpdb->prefix}".self::CACHE_TABLE_NAME." AS s ON m.sku = s.sku";
        if (!empty($skus)) {
            $query .= " WHERE m.sku IN (".implode(', ', array_fill(0, count($skus), '%s')).")";
            $params += $skus;
        } else {
            $query .= " WHERE m.sku <> ''";
        }
        if ($validate_cache) {
            $query .= " AND (
                (s.stock_cachetime + %d) < UNIX_TIMESTAMP()
                OR s.stock_cachetime IS NULL
            )";
            $params[] = $this->stock_expiration_time;
        }

        return $wpdb->get_col($wpdb->prepare($query, ...$params));
    }

    /**
     * Maybe update stock cache.
     *
     * After getting current stocks, remove entries with empty SKU / expired cache.
     * 
     * @param  boolean $force_read Wether to forecully request the stock remotely. Default `false`.
     * @param  string|array $sku The SKUs
     * @return boolean|array $response False if no force-read is set or no stock update needed, the SKUs whose stock was updated otherwise
     */
    public function maybe_update_stock_cache($force_read = false, $sku = null) {
        if (!$force_read && !$this->needs_stock_update()) {
            return false;
        }

        if (!is_null($sku) && !is_array($sku)) {
            $sku = [$sku];
        }
        if (empty($sku) || !$force_read) {
            $sku = $this->query_skus_and_cache_time($sku, !$force_read);
        }
        $chunks = array_chunk($sku, 1000);

        $response = array();
        $result = true;
        while ($chunks) {
            $stocks = $this->fetch_bexio_stock(array_shift($chunks));
            $response = array_merge($response, array_column($stocks, 'sku'));
            $result = $this->write_stock($stocks) && $result;
        }

        if ($result) {
            $this->cache_updated = true;
        }

        /*
        Remove entries which have an empty SKU or whose cache time has expired
        (including a 6h grace period)
        */
        global $wpdb;
        $wpdb->query($wpdb->prepare("DELETE FROM {$wpdb->prefix}".self::CACHE_TABLE_NAME."
            WHERE sku = ''
            OR stock_cachetime < %d", time() - $this->stock_expiration_time - 21600));

        return $response;
    }

    private function fetch_bexio_stock($skus) {
        $mode = apply_filters('raptus_wc_bexio_connector_stock_fetch_mode', null);
        $sku_modes = apply_filters('raptus_wc_bexio_connector_stock_products', array(), array_filter($skus));
        return Api::instance()->get_stock($skus, $mode, $sku_modes);
    }

    /**
     * Update the stock of all cart items and return a list of all items that are out of stock.
     *
     * @param  array A list of cart items to update.
     * @return array A list of cart items that are out of stock.
     */
    public function update_cart_stock($items) {
        // Get stock directly from bexio to ensure it's up-to-date
        $stocks = $this->fetch_bexio_stock(array_keys($items));
        $outofstock = array();
        foreach ($stocks as $stock) {
            $item = $items[$stock['sku']];
            $product = wc_get_product($item['id']);
            if (!$product) { // prevent fatal error in combination with woocommerce-free-gifts plugin
                continue;
            }
            // Ensure current stock gets stored in the cache table
            $this->write_stock([
                [
                    'sku' => $stock['sku'],
                    'stock' => $stock['stock']
                ]
            ]);
            // Update stock on the WC product
            $this->update_stock($stock['stock'], $product);
            if ($stock['stock'] < $item['quantity'] && !$product->backorders_allowed()) {
                $outofstock[$item['sku']] = $item['name'];
            }
        }
        return $outofstock;
    }

    /**
     * Gets the stock managing product of the given product.
     *
     * There is 3 possible return values:
     * - The product itself, if it is self-managing
     * - The parent product, if it is a variation with parent stock management
     * - The value `false` if stock management is disabled
     *
     * @param  \WC_Product $product The product to check
     * @return \WC_Product|boolean The stock managing product or `false` if stock management is disabled.
     */
    public function get_stock_managing_product($product) {
        $manage_stock = $product->managing_stock();
        if ($manage_stock) {
            if ($manage_stock === 'parent') {
                if (Utils::woocommerce_version_check('3.0')) {
                    return wc_get_product($product->get_parent_id());
                } else {
                    return wc_get_product($product->get_parent());
                }
            }
            return $product;
        }
        return false;
    }

    /**
     * Check if the cachetime for the given SKU is expired.
     *
     * @param string $sku The SKU to check the cachetime for.
     * @return boolean Returns `true` when expired or `false` when still valid.
     * */
    public function is_cache_expired($sku) {
        $cachetime = $this->get_cache_time($sku);
        return empty($cachetime) || ($cachetime + $this->stock_expiration_time) < time();
    }

    /**
     * Get a cache time for the given post-id.
     *
     * @param int $post_id The SKU of the entry to get.
     * @return boolean `true` if there is a cache time, `false` otherwise.
     * */
    public function get_cache_time_by_id($post_id) {
        global $wpdb;
        $result = $wpdb->get_results($wpdb->prepare("SELECT rs.stock_cachetime FROM {$wpdb->wc_product_meta_lookup} AS ml
            LEFT JOIN {$wpdb->prefix}".self::CACHE_TABLE_NAME." AS rs ON rs.sku = ml.sku
            WHERE ml.product_id = %d", $post_id));
        if (!empty($result[0])) {
            $cachetime = $result[0]->stock_cachetime;
            return intval($cachetime);
        }
        return false;
    }

    /**
     * Get a cache time for the given SKU.
     *
     * @param string $sku The SKU of the entry to get.
     * @return int / boolean Return the timestamp if there is a cache time, `false` otherwise.
     * */
    public function get_cache_time($sku) {
        global $wpdb;
        $result = $wpdb->get_var($wpdb->prepare("SELECT stock_cachetime FROM {$wpdb->prefix}".self::CACHE_TABLE_NAME." WHERE sku = %s", $sku));
        if (!empty($result)) {
            return intval($result);
        }
        return false;
    }

    /**
     * Update the cache time for the entry with the matching post-id.
     *
     * @param int $sku The SKU of the entry to update.
     * @param int $cache_time The timestamp to set.
     * @return boolean `true` if the cachetime could be updated, `false` otherwise.
     * */
    public function update_cache_time_by_id($post_id, $cache_time) {
        global $wpdb;
        $result = $wpdb->query($wpdb->prepare("UPDATE {$wpdb->prefix}".self::CACHE_TABLE_NAME." AS sc
            LEFT JOIN {$wpdb->wc_product_meta_lookup} AS ml ON ml.sku = sc.sku
            SET sc.stock_cachetime = %d
            WHERE ml.product_id = %d", $cache_time, $post_id));
        return !empty($result);
    }

    /**
     * Update the cache time for the entry with the matching post-id.
     *
     * @param int $sku The SKU of the entry to update.
     * @param int $cache_time The timestamp to set.
     * @return boolean `true` if the cachetime could be updated, `false` otherwise.
     * */
    public function update_cache_time($sku, $cache_time) {
        global $wpdb;
        $result = $wpdb->query($wpdb->prepare("UPDATE {$wpdb->prefix}".self::CACHE_TABLE_NAME." SET stock_cachetime = ".$cache_time." WHERE sku = %s", $sku));
        return !empty($result);
    }

    /**
     * Set cache time to zero if it is set, to force a new stock synchronization.
     *
     * @param  array $post_id The ID of the post to unset.
     */
    public function unset_cache_time($post_id) {
        $post = get_post($post_id);
        if ($post->post_type == 'product') {
            $cachetime = $this->get_cache_time_by_id($post_id);
            if ($cachetime !== false) {
                $this->update_cache_time_by_id($post_id, 0);
            }
        }
    }

    /**
     * Write stock data to database
     *
     * @param  array $data The stock data to write.
     * @return boolean `true` on success or `false` on failure.
     */
    private function write_stock($data) {
        global $wpdb;

        $rows = [];
        $result = true;
        while ($data) {
            $row = array_shift($data);
            $rows[] = $wpdb->prepare('(%s, %d, %d)', $row['sku'], $row['stock'], time());
        }

        if ($rows) {
            $write_stock_query = "REPLACE INTO {$wpdb->prefix}" . self::CACHE_TABLE_NAME . " (sku, stock, stock_cachetime) VALUES " . implode(', ', $rows);
            $result = $wpdb->query($write_stock_query) && $result;
        }

        return !!$result;
    }

    /**
     * Remove stock from the cache table
     *
     * @param array $skus The SKUs of stocks to remove.
     * @return boolean `true` on success or `false` on failure.
     * */
    private function remove_stock($skus) {
        global $wpdb;
        $result = $wpdb->query($wpdb->prepare("DELETE FROM {$wpdb->prefix}" . self::CACHE_TABLE_NAME . " WHERE sku IN (%s)", implode(', ', $skus)));
        return !!$result;
    }

    /**
     * Filters out empty SKUs.
     *
     * @param array $skus The SKUs to filter.
     * @return array The array of filtered SKUs, or empty array if param was an empty string.
     * */
    private function filter_empty_skus($skus) {
        if (is_array($skus)) {
            $skus = array_filter($skus);
        } else {
            if (empty($skus)) {
                $skus = [];
            }
        }
        return $skus;
    }

    /**
     * Read stock data for one or more SKUs
     *
     * @param  string|array $sku The SKU(s) of the product(s) to read the stock of.
     * @return integer|array The stock data that was read.
     */
    private function read_stock($sku) {
        $result = [];

        foreach ($this->iter_stock($sku) as $s => $row) {
            $result[$s] = $row->stock;
        }

        if ($sku && !is_array($sku)) {
            return isset($result[$sku]) ? $result[$sku] : null;
        }

        return $result;
    }

    /**
     * Iterator over cached stock data
     * 
     * @return array All cached over all cached stock data
     */
    public function iter_stock($sku = false) {
        $sku = $this->filter_empty_skus($sku);

        $this->maybe_update_stock_cache(false, $sku);

        global $wpdb;
        $where = '';
        if ($sku) {
            if (is_array($sku)) {
                $where = " WHERE `sku` IN (" . implode(',', array_map(function($s) use ($wpdb) {
                    return $wpdb->prepare('%s', $s);
                }, $sku)) . ")";
            } else {
                $where = $wpdb->prepare(" WHERE `sku` = %s", $sku);
            }
        } else {
            // Exclude any empty SKUs potentially still left in the cache table
            $where = " WHERE `sku` <> ''";
        }
        $where .= $wpdb->prepare(" AND stock_cachetime > %d", time() - $this->stock_expiration_time);

        $query = "SELECT `sku`, `stock` FROM {$wpdb->prefix}" . self::CACHE_TABLE_NAME . $where;

        // As of WP v6.4, check dbh for value 'mysqli', do not use property 'use_mysqli' as fallback for older WP versions
        $result = null;
        if (!empty($wpdb->dbh) && get_class($wpdb->dbh) == 'mysqli') {
            $result = mysqli_query($wpdb->dbh, $query);
        }
        if (!empty($result)) {
            while ($row = mysqli_fetch_object($result)) {
                yield $row->sku => $row;
            }
        } else {
            $result = $wpdb->get_results($query, OBJECT_K);
            foreach ($result as $key => $value) {
                yield $key => $value;
            }
        }
    }

    /**
     * Check if an update of the cached stock is necessary.
     *
     * @return boolean `true` if the stock update is due, `false` otherwise.
     */
    private function needs_stock_update() {
        return !$this->cache_updated;
    }

    /**
     * Tests if the stock synchronization is enabled.
     * The synchronization is enabled if the woocommerce option to manage the stock is activated
     * and the WooCommerce Bexio Automator stock synchronization is not disabled.
     *
     * @return boolean `true` if the stock synchronization is enabled, `false` otherwise.
     */
    public function is_stock_sync_enabled() {
        return apply_filters('raptus_wc_bexio_connector_enable_stock_synchronization',
            get_option('woocommerce_manage_stock') === 'yes' && !Options::get('bexio_disable_stock_synchronization', false)
        );
    }

    /**
     * Synchronize the stock of all products with the given IDs.
     * The cachetime is ignored, and the products are synchronized anyway.
     *
     * @param  array $product_ids The products ids to update the stock of.
     * @param  boolean $force_read Wether to forecully request the stock remotely,
     *         or rely on the underlying stock caching mechanism. Default `false`.
     * @return array The updated products.
     */
    public function sync_all($product_ids = array(), $force_read = false) {
        $skus = array();
        if($product_ids) {
            $skus = $this->get_non_empty_skus_by_id($product_ids);

            // If trying to sync specific product(s) and there are no SKUs to get stock for, abort (to avoid unnecessary requests and potential problems)
            if (empty($skus)) {
                return [
                    'updated' => [],
                    'not_updated' => [],
                    'skipped' => $product_ids
                ];
            }
        }
        $updated_skus = $this->maybe_update_stock_cache($force_read, $skus ?? null);

        if (!empty($skus)) {
            /**
             * Compare the requested and the actually updated stocks.
             * If a requested SKU did not get a new stock (because no Bexio product exists), remove it from our cache table.
             * */
            $nonexistent_stocks = array_diff($skus, $updated_skus);
            if (!empty($nonexistent_stocks)) {
                $this->remove_stock($nonexistent_stocks);
            }
        }

        // Update stock
        $updated = array();
        $not_updated = array();
        $skipped = array();
        foreach ($product_ids as $product_id) {
            $main_product = wc_get_product($product_id);
            $products = array_merge(array($main_product), $this->get_product_children($main_product));
            foreach($products as $product) {
                // Skip product if stock management is disabled or SKU is empty (as empty SKU can be ambiguous and leads to problems when updating stock)
                if ($product->managing_stock() && !empty($product->get_sku())) {
                    $stock = $this->read_stock($product->get_sku());
                    if (!is_null($stock)) {
                        $this->update_stock($stock, $product);
                        $updated[] = $product;
                    } else {
                        $not_updated[] = $product;
                    }
                } else {
                    $skipped[] = $product;
                }
            }
        }

        return compact('updated', 'not_updated', 'skipped');
    }

    /**
     * Get the child products of the given parent products, if applicable.
     *
     * @param  \WC_Product $product The parent product.
     * @return array The child products or an empty array.
     */
    private function get_product_children($product) {
        $children = array();
        if ($product->has_child()) {
            foreach($product->get_children() as $child_id) {
                $child = wc_get_product($child_id);
                $children[] = $child;
            }
        }
        return $children;
    }

    /**
     * Create stock cache table and
     * Migrate stock data from file system cache to database
     *
     *  STOCK_CACHE_TABLE_CREATED
     *  false: Table not yet created
     *  1 : Table created
     * 
     */
    public function migrate_stock() {
        global $wpdb;
        $charset_collate = $wpdb->get_charset_collate();
        $table_name = $wpdb->prefix . self::CACHE_TABLE_NAME;
        $varchar_size = 255; //default
        $postmeta_varchar = $wpdb->get_col_length( $wpdb->postmeta, 'meta_key');
        if ($postmeta_varchar['type'] == 'char' && $postmeta_varchar['length'] < $varchar_size) {
            $varchar_size = $postmeta_varchar['length'];
        }
        $varchar_size = apply_filters('raptus_wc_bexio_connector_cache_table_varchar_size', $varchar_size);

        $sql = $wpdb->prepare("CREATE TABLE IF NOT EXISTS {$table_name} (
            `sku` VARCHAR(%d) NOT NULL,
            `stock` INT NOT NULL,
            `stock_cachetime` INT DEFAULT 0,
            PRIMARY KEY (sku)
            ) {$charset_collate};", $varchar_size);

        $result = $wpdb->query($sql);

        if ($result !== false) {   //https://developer.wordpress.org/reference/classes/wpdb/#running-general-querie

            update_option(self::STOCK_CACHE_TABLE_EXISTS, 1);

            if (file_exists($this->stock_cache_file)) {
                $data = json_decode(file_get_contents($this->stock_cache_file), true);
                if (!empty($data)) {
                    $this->write_stock($data);
                }
            }
        }
    }

    /**
     *  Drop stock table if reset link is applied
     */
    public function drop_stock_table() {
        global $wpdb;
        $table_name = $wpdb->prefix . self::CACHE_TABLE_NAME;
        $wpdb->query( "DROP TABLE IF EXISTS {$table_name}" );
        delete_option(self::STOCK_CACHE_TABLE_EXISTS);
        delete_option('rwcbc_stock_migrated');
    }
}
