<?php

// Namespace declaration
namespace Raptus\WC\BexioConnector;

// Exit if accessed directly
defined('ABSPATH') or die();

/**
 * Test if a synchronization can take place, based on the following factors:
 * - Entity to be synchronized (order, invoice, delivery, payment)
 * - Payment gateway of the order
 * - Order status
 * - (Produt types of the order)
 */
class SyncRules {

    /**
     * The order state priorities.
     * @var array
     */
    private static $order_states;

    /**
     * Test if an order can be synchronized for the given WooCommerce order
     * The test is based on the user defined synchronization rules.
     *
     * @param  \WC_Order $order The order to test.
     * @return boolean True if a synchronization is permitted, false otherwise.
     */
    public static function can_sync_order($order) {
        return self::can_sync('order', $order);
    }

    /**
     * Test if an invoice can be synchronized for the given WooCommerce order.
     * The test is based on the user defined synchronization rules.
     *
     * @param  \WC_Order $order The order to test.
     * @return boolean True if a synchronization is permitted, false otherwise.
     */
    public static function can_sync_invoice($order) {
        return self::can_sync('invoice', $order);
    }

    /**
     * Test if an delivery can be synchronized for the given WooCommerce order.
     * The test is based on the user defined synchronization rules.
     *
     * @param  \WC_Order $order The order to test.
     * @return boolean True if a synchronization is permitted, false otherwise.
     */
    public static function can_sync_delivery($order) {
        return self::can_sync('delivery', $order);
    }

    /**
     * Test if an payment can be synchronized for the given WooCommerce order.
     * The test is based on the user defined synchronization rules.
     *
     * @param  \WC_Order $order The order to test.
     * @return boolean True if a synchronization is permitted, false otherwise.
     */
    public static function can_sync_payment($order) {
        return self::can_sync('payment', $order);
    }

    /**
     * Test if a synchronization of the given type for the specified order is permitted,
     * based on the user defined synchronization rules.
     *
     * @param  string $type The synchroniatzion type to test (order, invoice, delivery, payment).
     * @param  \WC_Order $order The order to test.
     * @return boolean True if a synchronization is permitted, false otherwise.
     */
    public static function can_sync($type, $order) {
        $gateway_id = apply_filters('raptus-wc-bexio-connector-sync-' . $type . '_gateway', Order::instance()->get_payment_method($order), $order);

        if (empty($gateway_id)) { // Prevent orders without a payment method from being synched, unless it was free
            $can_sync = $order->get_total() <= 0 && Options::get('bexio_sync_free_' . $type);
            return apply_filters('raptus-wc-bexio-connector-sync-' . $type, $can_sync, $order);
        }

        $rule = self::get_sync_rule($type, $gateway_id);

        if (self::has_only_virtual_products($order)) {
            $virtual_rule = self::get_sync_rule($type, 'virtual');
            if ($virtual_rule != 'default') {
                $rule = $virtual_rule;
            }
        }
        $can_sync = self::can_sync_status($order->get_status(), $rule) && self::can_sync_free($type, $order);
        return apply_filters('raptus-wc-bexio-connector-sync-' . $type, $can_sync, $order);
    }

    /**
     * Test if a synchronization of the given type for the specified order is permitted,
     * based on the user defined synchronization rules for free orders.
     *
     * @param  string $type The synchroniatzion type to test (order, invoice, delivery, payment).
     * @param  \WC_Order $order The order to test.
     * @return boolean True if a synchronization is permitted, false otherwise.
     */
    public static function can_sync_free($type, $order) {
        return Options::get('bexio_sync_free_' . $type) || $order->get_total() > 0;
    }

    /**
     * Get the synchronization rule for the given type and payment gateway.
     *
     * @param  string $type The synchroniatzion type to test (order, invoice, delivery, payment).
     * @param  string $gateway_id The gateway identifier.
     * @return string The synchronization rule if available.
     */
    public static function get_sync_rule($type, $gateway_id) {
        $settings_id = join('_', array_filter(['bexio_sync_rules', $gateway_id, $type]));
        return Options::get($settings_id);
    }

    /**
     * Test if the order contains only virtual products.
     *
     * @param  \WC_Order $order The order to test.
     * @return boolean True if only virtual products where found, false otherwise.
     */
    public static function has_only_virtual_products($order) {
        foreach ($order->get_items() as $item) {
            if (is_object($item) && $item->is_type('line_item')) {
                if ($product = $item->get_product()) {
                    if (!$product->is_virtual()) {
                        return false;
                    }
                }
            }
        }
        return true;
    }

    /**
     * Test if an order can be marked as delivered.
     *
     * @param  \WC_Order $order The order to test.
     * @return boolean True if a the order can be delivered, false otherwise.
     */
    public static function can_deliver_order($order) {
        return $order->get_status() == 'completed';
    }

    /**
     * Test if a synchronization is permitted based on two given order states.
     *
     * @param  string $order_status The order status.
     * @param  string $rule The expected status, the sync rule.
     * @return boolean True if a synchronization is permitted, false otherwise.
     */
    private static function can_sync_status($order_status, $rule) {
        if ($order_status == $rule) {
            return true;
        }

        if ($rule == 'never') {
            return false;
        }

        if ($order_status == 'failed') {
            return false;
        }

        if ($order_status == 'cancelled') {
            return false;
        }

        $states = self::get_order_state_priorities();
        if (isset($states[$order_status]) && isset($states[$rule])) {
            if ($states[$order_status] >= $states[$rule]) {
                return true;
            }
        }

        return false;
    }

    /**
     * Get the order state priorities.
     *
     * @return array The order state prorities.
     */
    private static function get_order_state_priorities() {
        if (!self::$order_states) {
            self::$order_states = apply_filters('raptus_wc_bexio_connector_order_state_priorities', array(
                'pending' => 10,
                'on-hold' => 20,
                'processing' => 30,
                'completed' => 40,
                'refunded' => 50,
                //'failed' => 60,
                //'cancelled' => 60,
            ));
        }
        return self::$order_states;
    }

    /*
     * Private constructor to avoid instances of this utlity class.
     */
    private function __construct() {}
}

?>