<?php

// Namespace declaration
namespace Raptus\WC\BexioConnector;

// Exit if accessed directly
defined('ABSPATH') or die();

// Dependencies
use stdClass;

/**
 * Handles auto-updating fuctionality.
 */
class Update {

    /**
     * Use Singleton trait to disallow multiple instances of this class.
     */
    use Singleton;

    /**
     * Constructs a new instance of this class.
     * Registers all required filter and action hooks.
     */
    protected function __construct() {
        // define the alternative API for update checking
        add_filter('pre_set_site_transient_update_plugins', array($this, 'check_update'));

        // Define the alternative response for information checks
        add_filter('plugins_api', array($this, 'check_info'), 10, 3);
    }

    /**
     * Add our self-hosted auto-update plugin to the update plugins transient.
     * Does nothing if the given transient not yet marked as checked.
     *
     * @param  object $transient The current update plugins transient
     * @return object The modified transient with the new plugin information.
     */
    public function check_update($transient) {
        if (empty($transient->checked)) {
            return $transient;
        }

        // If a newer version is available, add the update information to the transient
        $latest = Api::instance()->get_latest_plugin();
        if ($latest && version_compare(PLUGIN_VERSION, $latest['version'], '<')) {
            $obj = new stdClass();
            $obj->slug = PLUGIN_SLUG;
            $obj->plugin = PLUGIN_SLUG . '/' . PLUGIN_SLUG . '.php';
            $obj->plugin_name = $latest['name'];
            $obj->new_version = $latest['version'];
            $obj->url = $latest['homepage'];
            $obj->package = $latest['url'];
            $obj->date = $latest['date'];
            $obj->author = $latest['author'];
            $obj->author_url = $latest['author_url'];
            $obj->sections = array(
                'description' => $latest['description'],
                'changelog' => $latest['changelog'],
            );
            $obj->banners = array(
                'low' => $latest['banner'],
                'high' => $latest['banner'],
            );
            $obj->icons = array(
                '1x' => $latest['logo'],
                '2x' => $latest['logo'],
                'default' => $latest['logo'],
            );
            $transient->response[PLUGIN_SLUG . '/' . PLUGIN_SLUG . '.php'] = $obj;
        }
        return $transient;
    }

    /**
     * Add our self-hosted plugin description to the update API.
     *
     * Notes from WordPress:
     * - Passing a non-false value will effectively short-circuit the WordPress.org API request.
     * - If the action is ‘query_plugins’ or ‘plugin_information’, an object MUST be passed.
     * - If $action is ‘hot_tags’ or ‘hot_categories’, an array should be passed.
     *
     * In our case we always pass an object.
     *
     * @param  boolean $false The value to enable/disable the WordPress.org API request.
     * @param  array $action The current action (‘query_plugins’, ‘plugin_information’, ‘hot_tags’ or ‘hot_categories’)
     * @param  object $args The update parameters, needed to check the current plugin via slug.
     * @return bool|object Plugin update information if currently dealing with this plugin, or `false` otherwise.
     */
    public function check_info($false, $action, $args) {
        if (property_exists($args, 'slug') && $args->slug === PLUGIN_SLUG) {
            $transient = get_site_transient('update_plugins');
            if (!$transient || !isset($transient->response[PLUGIN_SLUG . '/' . PLUGIN_SLUG . '.php']))
                return false;
            $plugin_info = $transient->response[PLUGIN_SLUG . '/' . PLUGIN_SLUG . '.php'];
            if ($plugin_info) {
                $data = new stdClass();
                $data->slug = PLUGIN_SLUG;
                $data->name = $plugin_info->plugin_name;
                $data->version = $plugin_info->new_version;
                $data->last_updated = date('r', $plugin_info->date);
                $data->download_link = $plugin_info->package;
                $data->author = $plugin_info->author;
                $data->homepage = $plugin_info->url;
                $data->sections = $plugin_info->sections;
                $data->banners = $plugin_info->banners;

                return $data;
            }
        }
        return false;
    }
}

?>