<?php

// Namespace declaration
namespace Raptus\WC\BexioConnector;

// Exit if accessed directly
defined('ABSPATH') or die();

/**
 * Static class for providing plugin utility functions.
 */
class Utils {

    /**
     * Test if WooCommerce is active.
     */
    public static function is_woocommerce_active() {
        $plugin = 'woocommerce/woocommerce.php';
        if (function_exists('is_woocommerce_active')) {
            return is_woocommerce_active();
        }
        if (function_exists('is_plugin_active')) {
            return is_plugin_active($plugin);
        }
        return in_array($plugin, (array) get_option('active_plugins', array()));
    }

    /**
     * Check the the current WooCommerce version against the given one.
     *
     * @param  string $version The version to check in the format `x.x`. Defaults to `3.0`.
     * @return boolean The result of the version check.
     *         Returns `true` if the WooCommerce version is greater or equal to the given version.
     *         Returns `false` if the WooCommerce version is less than the given version.
     * @see    WordPress native function `version_compare` for more information about version comparison.
     */
    public static function woocommerce_version_check($version = '3.0') {
        static $woocommerce_version;
        isset($woocommerce_version) || $woocommerce_version = self::get_woocommerce_version();
        if (version_compare($woocommerce_version, $version, ">=")) {
            return true;
        }
        return false;
    }

    /**
     * Gets the current WooCommerce version if available.
     *
     * @return string The current WooCommerce version or `false` if not available.
     */
    public static function get_woocommerce_version() {
        $version = false;
        if (function_exists('is_woocommerce_active') && is_woocommerce_active()) {
            global $woocommerce;
            $version = $woocommerce->version;
        }
        if (!$version) {
            $woocommerce_path = dirname(PLUGIN_DIR) . '/woocommerce/woocommerce.php';
            if (function_exists('is_plugin_active') && is_plugin_active('woocommerce/woocommerce.php')) {
                $plugin_data = get_plugin_data($woocommerce_path, false, false);
                $version = $plugin_data['Version'];
            }
            if (!$version && function_exists('get_file_data')) {
                $plugin_data = get_file_data($woocommerce_path, array('Version' => 'Version'));
                $version = $plugin_data['Version'];
            }
        }
        return $version;
    }

    /**
     * Convenient function to check if an update of the plugin is currently available.
     * Note that this function makes use of the `update_plugins` transient
     * and does not perform any real-time checks.
     *
     * @return boolean `true` if an update is available, `false` otherwise.
     */
    public static function is_update_available() {
        $transient = get_site_transient('update_plugins');
        $key = PLUGIN_SLUG . '/' . PLUGIN_SLUG . '.php';
        if (isset($transient->response[$key])) {
            $plugin_info = $transient->response[$key];
            if ($plugin_info && version_compare(PLUGIN_VERSION, $plugin_info->new_version, '<')) {
                return true;
            }
        }
        return false;
    }

    /**
     * Utility function which displays a general support message to be appended to error messages.
     *
     * @param boolean $return Whether to echo or return the support message. Optional. Default `false`.
     */
    public static function support_message($return = false) {
        $msg = '';
        $status = Data::get_client_status();
        if ($status && $status['type'] == 'trial') {
            $msg .= sprintf(__('Having trouble with the configuration? Purchase the initial setup package to get help. Check <a href="%s" target="%s">here</a> for more information about this package.', 'raptus-wc-bexio-connector'),
                        Settings::instance()->get_settings_page_url(array('manage-account' => 1)), '_self');
        } else {
            $msg .= sprintf(__('If the error persists, take a look at the %s for solutions to known issues.', 'raptus-wc-bexio-connector'),
                          '<a href="' . FAQ_URL . '">' . __('FAQ', 'raptus-wc-bexio-connector') . '</a>');
        }
        if ($return) {
            return $msg;
        }
        echo $msg;
    }

    /**
     * Flatten an associative array recursively.
     *
     * *Example:*
     *
     * The following call
     * `flatten([0 => 'a', 1 => ['b', 'c']], 'x_', '_')`
     *
     * will result in the following output
     * `['x_0' => 'a', 'x_1_0' => 'b', 'x_1_1' => 'c']`
     *
     * @param  array $array The array to flatten.
     * @param  string $prefix The prefix to add before to all array keys. Default empty string.
     * @param  string $glue The glue to add between array keys. Default underscore `_`.
     * @return array The flattened array.
     */
    public static function flatten($array, $prefix = '', $glue = '_') {
        $result = array();
        foreach ($array as $key => $value) {
            if (is_array($value)) {
                $result = $result + self::flatten($value, $prefix . $key . $glue, $glue);
            } else {
                $result[$prefix . $key] = $value;
            }
        }
        return $result;
    }

    /**
     * Print a (dismissible) admin notice of the given type with the specified message.
     *
     * @param  string $type The message type: success, warning, error.
     * @param  string $message The message to display.
     * @param  boolean $dismissable Whether this notrice is dismissible or not. Optional. Default `true`.
     * @param  string $wrap The container wrap, defaults to `p`
     */
    public static function print_notice($type, $message, $dismissable = true, $wrap = 'p', $skip_log = false) {
        if(in_array($type, array('error', 'warning', 'info')) && !$skip_log) {
            Api::instance()->log_to_remote($message, $type);
        }
        $lines = is_array($message) ? $message : preg_split('/<br\/?>/', $message);
        ?>
        <div class="notice notice-<?php echo $type; ?><?php echo $dismissable ? ' is-dismissible': ''; ?>">
            <<?= $wrap ?><?= $wrap != 'p' ? ' style="margin-top:10px;"' : '' ?>>
                <strong><?php echo array_shift($lines); ?></strong>
                <?php echo !empty($lines) ? '<br/>' . join('<br/>', $lines) . '<br/>' : ''; ?>
            </<?= $wrap ?>>
            <?php if ($dismissable) { ?>
            <button type="button" class="notice-dismiss">
                <span class="screen-reader-text"><?php _e('Dismiss this notice.', 'woocommerce') ?>.</span>
            </button>
            <?php } ?>
            <?php if ($type == 'error') { ?>
                <p><em><?php static::support_message() ?></em></p>
            <?php } ?>
        </div>
        <?php
    }

    /**
     * Get the default domain based on `WP_HOME`.
     * The domain is fetched without https or https.
     *
     * @return string The domain.
     */
    public static function get_domain() {
        $domain = get_option('home', '');
        $domain = implode('', array_slice(explode('//', $domain), 1));
        return $domain;
    }

    /**
     * Get the language of the given post ID.
     * Currently supported translation frameworks include:
     * - WPML
     * - Polylang
     * - qTranslate X
     * - Native WordPress language setting (fallback)
     *
     * @param int $post_id The post ID.
     * @param string $type The post type.
     * @return string|boolean The found language code (ISO-2) or `false`.
     */
    public static function get_language_code($post_id = null, $type = null) {
        if (function_exists('icl_object_id') && $post_id != null && $type != null) {
            $args = array('element_id' => $post_id, 'element_type' => $type);
            $lang = apply_filters('wpml_element_language_code', null, $args);
        } else if (has_filter('wpml_current_language')) {
            $lang = apply_filters('wpml_current_language', null);
        } else if (function_exists('pll_get_post_language') && $post_id != null) {
            $lang = pll_get_post_language($post_id, 'slug');
        } else if (function_exists('pll_current_language')) {
            $lang = pll_current_language();
        } else if (function_exists('qtranxf_getLanguage')) {
            $lang = qtranxf_getLanguage();
        }
        if (!isset($lang) || empty($lang)) {
            $lang = get_locale();
        }
        if ($lang) {
            $parts = explode('_', $lang);
            return array_shift($parts);
        }
        return false;
    }

    /**
     * Test if plugin settings are expired.
     * The settings are expired if they have not yet been saved since the last plugin update.
     *
     * @return boolean `true` if the setting are expired, `false` otherwise.
     */
    public static function settings_expired() {
        if ($settings_version = Options::get('version')) {
            return version_compare($settings_version, PLUGIN_VERSION, '<');
        }
        return true;
    }

    /**
     * Private constructor to avoid instances of this static utility class.
     */
    private function __construct() {}
}

?>