<?php

// Namespace declaration
namespace Raptus\WC\BexioConnector;

use Raptus\WC\BexioConnector\Settings;

// Exit if accessed directly
defined('ABSPATH') or die();

/**
 * The wizard module of this plugin.
 */
class Wizard {

    /**
     * Use Singleton trait to disallow multiple instances of this class.
     */
    use Singleton;

    /**
     * Constructs a new instance of the this class.
     * Registers all required filter and action hooks.
     */
    protected function __construct() {
        add_action('raptus_wc_bexio_connector_render_options_page', array($this, 'process'), 5);
        add_filter('raptus_wc_bexio_connector_sanitize_settings', array($this, 'sanitize'), 10, 2);

        if (!$this->isDone()) {
            if (is_admin() && isset($_GET['page']) && $_GET['page'] == Settings::SETTINGS_PAGE_ID) {
                add_action('admin_head', array($this, 'enqueue_wizard_css'));
                add_action('admin_enqueue_scripts', array($this, 'enqueue_wizard_js'));
            }

            add_filter('raptus_wc_bexio_connector_defaults', array($this, 'defaults'), 10, 1);
            add_filter('raptus_wc_bexio_connector_settings_base_template', array($this, 'template'), 10, 1);
            add_filter('raptus_wc_bexio_connector_settings_template_args', array($this, 'templateArgs'), 10, 1);
            add_action('raptus_wc_bexio_connector_meta_boxes', array($this, 'markMetaboxes'), 10, 1);
        }
    }

    /**
     * Enqueue custom CSS
     * */
    public function enqueue_wizard_css() {
        wp_enqueue_style('wc-bexio-automator-wizard-styles', \Raptus\WC\BexioConnector\ASSETS_URL . 'css/wizard.css');
    }

    /**
     * Enqueue custom JS
     * */
    public function enqueue_wizard_js() {
        wp_enqueue_script('wc-bexio-automator-wizard-script', \Raptus\WC\BexioConnector\ASSETS_URL . 'js/wizard.js', ['jquery'], \Raptus\WC\BexioConnector\PLUGIN_VERSION);
    }

    /**
     * Mark the meta boxes that are part of the wizard with according classes.
     */
    public function markMetaboxes($screen_id) {
        $steps = $this->steps();
        $current = $steps[$this->currentStep()];
        foreach ($this->steps() as $i => $step) {
            $box_id = 'raptus_wp_bexio_connector_' . $step;
            add_filter("postbox_classes_{$screen_id}_{$box_id}", function($classes) use ($step, $current) {
                $classes[] = 'wizard-section';
                if ($step == $current) {
                    $classes[] = 'active';
                }
                return $classes;
            }, 10, 1);
        }
    }

    /**
     * Process the current request.
     * Update the wizard state if necessary.
     */
    public function process() {
        if (isset($_GET['wizard'])) {
            Options::set('wizard', $_GET['wizard']);
            Settings::instance()->reload();
        }
    }

    /**
     * Sanitize the additional settings for the wizard.
     * Save the current wizard state in the options table.
     *
     * @param array $new_input The sanitized input values.
     * @param array $input The current input values.
     * @return array The modified sanitized input values.
     */
    public function sanitize($new_input, $input) {
        if (isset($_POST['wizard'])) {
            $new_input['wizard'] = $_POST['wizard'];
        } else if ($this->isDone()) {
            $new_input['wizard'] = 'done';
        }
        return $new_input;
    }

    /**
     * Inject the default values for the additional settings of the wizard.
     *
     * @param array $defaults The default options array.
     * @return array The modified default options array.
     */
    public function defaults($defaults) {
        $defaults['wizard'] = $this->firstStep();
        return $defaults;
    }

    /**
     * Change the settings template if the wizard is active.
     *
     * @param string $template The current base template.
     * @return string The modified base template.
     */
    public function template($template) {
        if (!$this->isDone()) {
            $template = TEMPLATES_DIR . 'wizard.php';
        }
        return $template;
    }

    /**
     * Provide additional tmplate parameters to the wizard template.
     *
     * @param array $args The template args array.
     * @return array The modified template args array.
     */
    public function templateArgs($args) {
        if (!$this->isDone()) {
            $args['wizard'] = $this;
        }
        return $args;
    }

    /**
     * Get the wizard steps.
     */
    public function steps() {
        $steps = [
            1 => 'registration',
            2 => 'authorization',
            3 => 'configuration',
            4 => 'advanced',
            5 => 'payment_methods',
        ];

        if (Options::get('isauthorized')) {
            unset($steps[2]);
        }

        $indexedSteps = array();
        $index = 1;
        foreach ($steps as $step) {
            $indexedSteps[$index] = $step;
            $index++;
        }

        return $indexedSteps;
    }

    /**
     * Test if the current wizard step is equal to the given one.
     * Both nueric and textual identifiers are supported.
     *
     * @return int|string $step The wizard step to test against.
     * @return boolean `true` if the current wizard is equal, `false` otherwise.
     */
    public function isStep($step) {
        $current = $this->currentStep();
        if (is_numeric($step) || $step == 'done') {
            return $current == $step;
        }
        $steps = $this->steps();
        $index = array_search($step, $steps);
        return $index == $current;
    }

    /**
     * Test if the given wizard step exists.
     * Both nueric and textual identifiers are supported.
     *
     * @return int|string $step The wizard step to test against.
     * @return boolean `true` if the wizard step exists, `false` otherwise.
     */
    public function hasStep($step) {
        $steps = $this->steps();
        if (is_numeric($step)) {
            return isset($steps[$step]);
        }
        $index = array_search($step, $steps);
        return $index !== false;
    }

    /**
     * Get the current wizard step from the settings.
     *
     * @return int The wizard step index or 'done'.
     */
    public function currentStep() {
        return Options::get('wizard') ?: $this->firstStep();
    }

    /**
     * Get the previous wizard step.
     *
     * @return int The wizard previous step index.
     */
    public function previousStep() {
        return $this->currentStep() - 1;
    }

    /**
     * Test if the wizard has a previous step.
     *
     * @return boolean `true` if the wizard has a previous step, `false` otherwise.
     */
    public function hasPreviousStep() {
        if ($this->hasStep('authorization')) {
            return $this->currentStep() > 3;
        }
        return !$this->isFirstStep();
    }

    /**
     * Get the previous wizard step label.
     *
     * @return string The previous step label.
     */
    public function previousStepLabel() {
        return __('Back', 'raptus-wc-bexio-connector');
    }
    /**
     * Get the next wizard step.
     *
     * @return int The wizard next step index or 'done'.
     */
    public function nextStep() {
        $next = $this->currentStep() + 1;
        return $next > $this->lastStep() ? 'done' : $next;
    }

    /**
     * Test if the wizard has a next step.
     *
     * @return boolean `true` if the wizard has a next step, `false` otherwise.
     */
    public function hasNextStep() {
        return $this->currentStep() <= $this->lastStep() && !$this->isStep('authorization');
    }

    /**
     * Get the next wizard step label.
     *
     * @return string The next step label.
     */
    public function nextStepLabel() {
        $label = __('Next', 'raptus-wc-bexio-connector');
        if ($this->isLastStep()) {
            $label = __('Finish', 'raptus-wc-bexio-connector');
        } else if ($this->isStep('authorization')) {
            $label = __('Activate Bexio API', 'raptus-wc-bexio-connector');
        } else if ($this->isStep('registration') && !empty(Options::get('name'))) {
            $label = __('Register', 'raptus-wc-bexio-connector');
        }
        return $label;
    }

    /**
     * Get the first wizard step.
     *
     * @return int The first wizard step index.
     */
    public function firstStep() {
        return 1;
    }

    /**
     * Test if the wizard is currently in the first step.
     *
     * @return boolean `true` if the wizard is in the first step, `false` otherwise.
     */
    public function isFirstStep() {
        return $this->currentStep() == $this->firstStep();
    }

    /**
     * Get the last wizard step.
     *
     * @return int The last wizard step index.
     */
    public function lastStep() {
        return count($this->steps());
    }

    /**
     * Test if the wizard is currently in the last step.
     *
     * @return boolean `true` if the wizard is in the last step, `false` otherwise.
     */
    public function isLastStep() {
        return $this->currentStep() >= $this->lastStep();
    }

    /**
     * Test if the wizard is already completed. This function will also ensure backwards compatibility.
     * The wizard considered "done" after a plugin update, if the configuration has been previously done.
     *
     * @return boolean `true` if the wizard is done, `false` otherwise.
     */
    public function isDone() {
        if ($this->currentStep() == 'done') {
            return true;
        }

        // Ensure backwards compatibility, consider the wizard done after an update
        if (Utils::settings_expired()) {
            $options = Options::get();
            if (empty($options)) {
                return false;
            }
            if (isset($options['authorization']) && !isset($options['version'])) {
                return true;
            }
        }

        return false;
    }
}
