<?php

namespace Raptus\WC\BexioConnector\Compatibility;

defined('ABSPATH') or die;

use Raptus\WC\BexioConnector\Singleton;
use Raptus\WC\BexioConnector\Admin;
use Raptus\WC\BexioConnector\Order;

/**
 * Provides compatibility with the following plugin:
 * - WooCommerce Subscriptions
 *
 * @see https://woocommerce.com/products/woocommerce-subscriptions/
 */
class WooCommerceSubscriptions {

    /**
     * Use Singleton trait to disallow multiple instances of this class.
     */
    use Singleton;

    /**
     * Constructs a new instance of this class.
     * Registers all required filter and action hooks.
     */
    protected function __construct() {
        add_action('woocommerce_admin_order_data_after_order_details', array($this, 'disable_admin_order_meta'), 5, 1);
        add_filter('wcs_new_order_created', [$this, 'sync_renewals'], 999, 1);
        add_filter('wc_subscriptions_renewal_order_data', array($this, 'do_not_copy_bexio_metadata'), 10, 3);
        add_filter('wc_subscriptions_subscription_data', array($this, 'do_not_copy_bexio_metadata'), 10, 3);
        add_filter('raptus_wc_bexio_connector_transition_order_state', array($this, 'transition_order_state'), 10, 2);
    }

    /**
     * Do not copy bexio meta data to subscription and renewal orders
     *
     * @param array $data The meta data to copy
     * @return array $data The meta data to copy
     */
    public function do_not_copy_bexio_metadata($data, $to_object, $from_object) {
        foreach ($data as $key => $value) {
            if(strpos($key, '_bexio_') !== false) {
                unset($data[$key]);
            }
        }
        return $data;
    }

    /**
     * Do not add Bexio information to subscriptions detail admin page.
     *
     * @param \WC_Order $order The current order.
     */
    public function disable_admin_order_meta($order) {
        if (get_class($order) == 'WC_Subscription') {
            remove_action('woocommerce_admin_order_data_after_order_details', array(Admin::instance(), 'admin_order_meta'), 10, 1);
        }
    }

    /**
     * Synchronize subscription renewals.
     *
     * @param \WC_Order $order The created order.
     */
    public function sync_renewals($order) {
        Order::instance()->sync_all($order->get_id());
        return $order;
    }

    /**
     * Do not transition order state if order is a subscription.
     *
     * @param boolean $transition_order Whether to transition the order state.
     * @param $order The current order.
     */
    public function transition_order_state($transition_order, $order) {
        if(get_class($order) == 'WC_Subscription') {
            return false;
        }
        return $transition_order;
    }
}

WooCommerceSubscriptions::instance();

?>