<?php

namespace Raptus\WC\BexioConnector\Modules;

defined('ABSPATH') or die;

use WC_Tax;
use Raptus\WC\BexioConnector\Options;
use Raptus\WC\BexioConnector\Settings;
use Raptus\WC\BexioConnector\Data;

/**
 * Provides functionality to set country-specific booking accounts and taxes.
 * - Order numbers
 * - Invoice numbers
 * - Delivery numbers
 *
 * This feature can only be used, if the automatic document numbering is disabled in Bexio.
 */
class CountrySettings extends Module {

    /**
    * The field types.
     * @var array
     */
    private $documents;

    /**
    * The document types.
     * @var array
     */
    private $types;

    /**
     * The accounts to create options for.
     * */
    private $accounts = [];

    /**
     * The countries to create options for.
     * */
    private $countries = [];

    /**
     * Constructs a new instance of this class.
     * Registers all required filter and action hooks.
     */
    public function __construct() {
        parent::__construct('countrysettings', _x('Country settings', 'Country settings module', 'raptus-wc-bexio-connector'));
    }

    /**
     * Load this module.
     * This function should only be invoked if this module is active.
     */
    protected function load() {
        add_action('admin_init', [$this, 'add_meta_boxes'], 20);
        add_filter('raptus_wc_bexio_connector_sanitize_settings', [$this, 'sanitize'], 10, 2);
        add_filter('raptus_wc_bexio_connector_defaults', [$this, 'defaults']);

        add_filter('raptus_wc_bexio_connector_order_data', [$this, 'add_country_order_data'], 10, 2);
        add_filter('raptus_wc_bexio_connector_order_item_data', [$this, 'add_country_order_item_data'], 11, 4);
        add_filter('raptus_wc_bexio_connector_fee_data', [$this, 'add_country_fee_data'], 12, 2);
        add_filter('raptus_wc_bexio_connector_shipping_data', [$this, 'add_country_shipping_data'], 13, 2);
        add_filter('raptus_wc_bexio_connector_discount_data', [$this, 'add_country_discount_data'], 13, 2);
    }

    /**
     * Get the description of this module.
     *
     * @return string the description of this module
     */
    public function get_description() {
        return _x('This additional module allows setting individual booking accounts and taxes for selected shipping countries.', 'Country settings module', 'raptus-wc-bexio-connector');
    }

    /**
     * Add new meta boxes with settings for country-specific taxes and accounts.
     */
    public function add_meta_boxes() {
        if(Options::get('isauthorized')) {
            add_meta_box('raptus_wp_bexio_connector_country_settings',
                _x('Country specific settings', 'Country settings module', 'raptus-wc-bexio-connector'),
                [$this, 'do_settings_sections_country_settings'],
                'woocommerce_page_raptus_wp_bexio_connector_options_admin',
                'normal',
                'core'
            );

            $settings_page_id = Settings::SETTINGS_PAGE_ID . '_country_settings';
            add_settings_section(
                'raptus_wp_bexio_connector_options_settings_bexio_country_settings',
                '',
                [$this, 'print_section_info_bexio_country_settings'],
                $settings_page_id
            );

            add_settings_field(
                'bexio_country_settings_table_header',
                _x('Country', 'Country settings module', 'raptus-wc-bexio-connector'),
                [$this, 'bexio_country_settings_table_header'],
                $settings_page_id,
                'raptus_wp_bexio_connector_options_settings_bexio_country_settings'
            );

            $countries = $this->get_country_list();
            foreach ($countries as $code => $title) {
                add_settings_field(
                    'bexio_order_country'.$code.'_settings',
                    $title . ' (' . $code . ')',
                    $this->country_settings($code),
                    $settings_page_id,
                    'raptus_wp_bexio_connector_options_settings_bexio_country_settings'
                );
            }
        }
    }

    /**
     * Get a list of all WooCommerce countries the shop is shipping to.
     */
    public function get_country_list() {
        if (!$this->countries || empty($this->countries)) {
            global $woocommerce;
            $countries_obj = new \WC_Countries();
            $this->countries = $countries_obj->get_shipping_countries();
        }
        return $this->countries;
    }

    /**
     * Render country settings sections.
     */
    public function do_settings_sections_country_settings() {
        do_settings_sections(Settings::SETTINGS_PAGE_ID . '_country_settings');
    }

    /**
     * Print the Bexio country settings section text.
     */
    public function print_section_info_bexio_country_settings() {
        echo '<p><em>';
        _ex('Settings to control the account and tax set on order items, based on the country of the shipping address.', 'Country settings module', 'raptus-wc-bexio-connector');
        echo '<br>';
        _ex('The default value uses the account set for product income or the corresponding tax setting.', 'Country settings module', 'raptus-wc-bexio-connector');
        if(Options::get('bexio_account_use_product_account')) {
            echo '<br><br>';
            echo '<b>';
            _ex('ATTENTION: the option to use the account of the bexio product is set. As such, the country specific account will not be applied and thus has no effect.', 'Country settings module', 'raptus-wc-bexio-connector');
            echo '</b>';
        }
        echo '</em></p>';
    }

    /**
     * Print the settings table header.
     * */
    public function bexio_country_settings_table_header() {
        echo '<table class="wcba-settings-table">';
            echo '<thead>';
                echo '<tr>';
                    echo '<th>'._x('Setting', 'Country settings module', 'raptus-wc-bexio-connector').'</th>';
                    echo '<th>'._x('Value', 'Country settings module', 'raptus-wc-bexio-connector').'</th>';
                echo '</tr>';
            echo '</thead>';
        echo '</table>';
    }

    /**
     * Print the settings table.
     * */
    public function country_settings($code) {
        return function() use ($code) {
            echo '<table class="wcba-settings-table">';
                echo '<thead>';
                    echo '<tr>';
                        echo '<td>';
                            echo '<h4>';
                                echo _x('Account', 'Country settings module', 'raptus-wc-bexio-connector');
                            echo '</h4>';
                        echo '</td>';
                        echo '<td>';
                        echo '</td>';
                    echo '</tr>';
                    echo '<tr class="country-settings-account">';
                        echo '<td><h5>';
                        echo __('Account for booking product income', 'raptus-wc-bexio-connector');
                        echo '</h5></td>';
                        echo '<td>';
                        $this->country_account_select($code);
                        echo '</td>';
                    echo '</tr>';
                    echo '<tr><td><hr></td><td><hr></td></tr>';
                    $this->country_tax_classes($code);
                echo '</thead>';
            echo '</table>';
        };
    }

    /**
     * Print the tax rate selection for the active tax classes.
     * */
    private function country_tax_classes($code) {
        echo '<tr>';
            echo "<td><h4>";
            echo _x('Taxes', 'Country settings module', 'raptus-wc-bexio-connector');
        echo '</h4></td>';

        $classes = Data::get_tax_classes();
        $classes = array_merge([
            'shipping' => __('Shipping costs', 'raptus-wc-bexio-connector')
        ], $classes);
        foreach ($classes as $class => $title) {
            echo '<tr>';
                echo "<td><h5>";
                echo $title;
                echo '</h5></td>';

                echo "<td>";
                echo $this->country_tax_select($code, $class);
                echo '</td>';
            echo '</tr>';
        }
    }

    /**
     * Add a select field to set the tax for a country.
     * 
     * @param string $code The language code.
     */
    private function country_tax_select($code, $class) {
        $tax_id = 'bexio_country_tax_'.$code.'_'.$class;
        echo '<select id="'.$tax_id.'" name="raptus_wp_bexio_connector_options['.$tax_id.']" class="regular-text">';
        $types = Data::get_tax_types();
        $selected = "";
        $tax_option = Options::get($tax_id);
        foreach ($this->get_tax_groups() as $type => $group) {
            if($type == 'default') {
                printf('<option value="%s" %s>%s</option>',
                    $group['id'],
                    selected($tax_option, 'default', false),
                    $group['value']
                );
            } else {
                echo '<optgroup label="'.$types[$type].'">';
                foreach ($group as $item) {
                    printf('<option value="%s" %s>%s%% (ID: %s)</option>',
                        $item['id'],
                        selected($tax_option, $item['id'], false),
                        $item['value'],
                        $item['id']
                    );
                }
                echo '</optgroup>';
            }
        }
        echo '</select>';
    }

    /**
     * Add a select field to set the account for a country.
     * Selection is disabled if 'bexio_account_use_product_account' is set to use bexio as source.
     * 
     * @param string $code The language code.
     */
    private function country_account_select($code) {
        $account_id = 'bexio_country_account_'.$code;
        printf('<select id="%s" name="%s" class="regular-text"%s>',
            $account_id,
            'raptus_wp_bexio_connector_options['.$account_id.']',
            Options::get('bexio_account_use_product_account') ? ' disabled' : ''
        );
        $account_option = Options::get($account_id);
        foreach ($this->get_accounts() as $item) {
            if($item['id'] == 'default') {
                printf('<option value="%s" %s>%s</option>',
                    $item['id'],
                    selected(Options::get($item['id']), 'default', false),
                    $item['name']
                );
            } else {
                printf('<option value="%s" %s>%s: %s</option>',
                    $item['id'],
                    selected($account_option, $item['id'], false),
                    $item['account_no'],
                    $item['name']
                );
            }
        }
        echo '</select>';
    }

    /**
     * Get a list of tax groups and taxes and add a default setting.
     */
    private function get_tax_groups() {
        $taxes = [];
        $taxes['default'] = [
            'id' => 'default',
            'value' => _x('Default Value', 'Country settings module', 'raptus-wc-bexio-connector')
        ];
        foreach (Data::get_taxes() as $item) {
            if (!isset($taxes[$item['type']])) {
                $taxes[$item['type']] = [];
            }
            $taxes[$item['type']][] = $item;
        }
        return $taxes;
    }

    /**
     * Get a list of accounts and add a default setting.
     */
    private function get_accounts() {
        if (!$this->accounts || empty($this->accounts)) {
            $this->accounts = [];
            $this->accounts[] = [
                'id' => 'default',
                'name' => _x('Default Value', 'Country settings module', 'raptus-wc-bexio-connector')
            ];
            $this->accounts = array_merge($this->accounts, Data::get_accounts());
        }
        return $this->accounts;
    }

    /**
     * Sanitize country specific fields when saving the plugin settings.
     * The input will only be saved when it is not set to the default setting
     * (which uses 'bexio_tax_shipping' / 'bexio_account_products' respectively)
     * or to the same value as the tax / account settings themselves.
     * 
     * @param $new_input The new settings to be saved to the database.
     * @param $input The currently selected settings.
     */
    public function sanitize($new_input, $input) {
        foreach ($this->get_country_list() as $code => $title) {
            // sanitize account
            $id = "bexio_country_account_{$code}";
            if (isset($input[$id])) {
                $new_input[$id] = $input[$id];
            }

            // sanitize taxes
            $classes = Data::get_tax_classes();
            $classes = array_merge([
                'shipping' => __('Shipping costs', 'raptus-wc-bexio-connector')
            ], $classes);
            foreach ($classes as $class => $title) {
                $id = "bexio_country_tax_{$code}_{$class}";
                if (isset($input[$id])) {
                    $new_input[$id] = $input[$id];
                }
            }
        }
        return $new_input;
    }

    /**
     * Set defaults for country specific account and tax rates.
     * */
    public function defaults($defaults) {
        foreach ($this->get_country_list() as $code => $title) {
            $defaults['bexio_country_account_'.$code] = 'default';

            $classes = Data::get_tax_classes();
            $classes = array_merge([
                'shipping' => __('Shipping costs', 'raptus-wc-bexio-connector')
            ], $classes);
            foreach ($classes as $class => $title) {
                $defaults['bexio_country_tax_'.$code.'_'.$class] = 'default';
            }
        }
        return $defaults;
    }

    /**
     * Set general account for synchronization.
     * */
    public function add_country_order_data($order_data, $order) {
        $default_account = Options::get('bexio_account_products');
        $this->inject_country_data($order_data, $order, 'account', $default_account);
        return $order_data;
    }

    /**
     * Override tax and account for order item.
     * 
     * @param array $data The order item.
     * @param object $item The WC_Order_Item_Product object.
     * @param object $product The WC_Product object.
     * @param object $order The order object.
     * */
    public function add_country_order_item_data($data, $item, $product, $order) {
        $default_tax = Options::get('bexio_tax_' . ($item['tax_class'] ?: 'standard'));
        $default_account = Options::get('bexio_account_products');
        $this->inject_country_data($data, $order, 'tax', $default_tax, '_' . ($item['tax_class'] ?: 'standard'));
        if(!Options::get('bexio_account_use_product_account'))
            $this->inject_country_data($data, $order, 'account', $default_account);
        return $data;
    }

    /**
     * Override tax for fee data.
     * 
     * @param array $fee The fee data.
     * @param object $order The order object.
     * */
    public function add_country_fee_data($fee, $order) {
        $default_tax = Options::get('bexio_tax_' . ($fee['tax_class'] ?: 'standard'));
        $default_account = Options::get('bexio_account_products');
        $this->inject_country_data($fee, $order, 'tax', $default_tax, '_standard');
        $this->inject_country_data($fee, $order, 'account', $default_account);
        return $fee;
    }

    /**
     * Override tax for shipping data.
     * 
     * @param array $shipping The shipping data.
     * @param object $order The order object.
     * */
    public function add_country_shipping_data($shipping, $order) {
        $default_tax = Options::get('bexio_tax_shipping');
        $this->inject_country_data($shipping, $order, 'tax', $default_tax, '_shipping');
        return $shipping;
    }

    /**
     * Override tax for discount data.
     * Do nothing if tax is not synced as a fee.
     * 
     * @param array $discount The discount data.
     * @param object $order The order object.
     * */
    public function add_country_discount_data($discount, $order) {
        if(!isset($discount['tax_id']))
            return $discount;
        $default_tax = Options::get('bexio_tax_standard');
        $this->inject_country_data($discount, $order, 'tax', $default_tax, '_standard');
        return $discount;
    }

    /**
     * Set country specific account or tax in the given order data if applicable.
     *
     * @param array $data The order data to enhance. Passed by reference.
     * @param string $order The order object.
     * @param string $type The data type (tax or account).
     * @param string $default The default value.
     * @param string $option_suffix The suffix of the option to get, like tax class.
     */
    protected function inject_country_data(&$data, $order, $type, $default, $option_suffix = '') {
        $country = $order->get_shipping_country();
        $value = Options::get("bexio_country_{$type}_{$country}{$option_suffix}");

        // abort if nothing needs to be done
        if (empty($value) || $value == $default || $value == 'default') {
            return;
        }

        $id_key = "{$type}_id";

        // only override the global setting, not the item specific ones
        if ($default && $data[$id_key] == $default) {
            $data[$id_key] = $value;
        }
    }

}