<?php

namespace Raptus\WC\BexioConnector\Modules;

defined('ABSPATH') or die;

use Raptus\WC\BexioConnector\Api;
use Raptus\WC\BexioConnector\Options;
use Raptus\WC\BexioConnector\Modules;
use \WC_Tax;

/**
 * Provides functionality to synchronize discounts differently.
 */
class DiscountSync extends Module {

    /**
     * Constructs a new instance of this class.
     */
    public function __construct() {
        parent::__construct('discountsync', _x('Discount Sync', 'DiscountSync Module', 'raptus-wc-bexio-connector'));

        // This filter needs to always be registered (not only when module is active)
        add_action('woocommerce_new_order_item', [$this, 'add_order_item_data'], 10, 3);

        add_filter('woocommerce_hidden_order_itemmeta', [$this, 'hide_order_item_meta']);
    }

    /**
     * Get the description of this module.
     *
     * @return string the description of this module
     */
    public function get_description() {
        return _x('Allows finer control over how discounts are synchronized, including discounts per position.', 'DiscountSync Module', 'raptus-wc-bexio-connector');
    }

    /**
     * Load this module.
     * This function should only be invoked if this module is active.
     */
    protected function load() {
        if (is_admin()) {
            $this->load_global_settings();
        }

        add_filter('raptus_wc_bexio_connector_order_item_data', [$this, 'calculate_discount_per_position'], 10, 4);
        add_filter('raptus_wc_bexio_connector_discount_data', [$this, 'disable_discount_position']);
    }

    /**
     * Load the global settings for configuring the discount synchronization mode.
     */
    protected function load_global_settings() {
        add_action('raptus_wp_bexio_connector_after_settings/bexio_account_discount', [$this, 'settings'], 10, 2);
        add_filter('raptus_wc_bexio_connector_sanitize_settings', [$this, 'sanitize'], 10, 2);
        add_filter('raptus_wc_bexio_connector_defaults', [$this, 'defaults'], 10, 1);
    }

    /**
     * Add the additional settings that this modules provides.
     */
    public function settings($page, $section) {
        add_settings_field(
            'bexio_discount_sync_mode',
            _x('Discount synchronization mode', 'DiscountSync Module', 'raptus-wc-bexio-connector'),
            [$this, 'bexio_discount_sync_mode_callback'],
            $page,
            $section
        );
    }

    /**
     * Callback for the additional settings that this modules provides.
     * 
     * Summarized = regular behaviour (creates a discount position)
     * Itemized = new behaviour (calculates and sets discount per position)
     */
    public function bexio_discount_sync_mode_callback() {
        echo '<select id="bexio_discount_sync_mode" name="raptus_wp_bexio_connector_options[bexio_discount_sync_mode]" class="regular-text">';
        echo '<option value="summarized" ' . selected(Options::get('bexio_discount_sync_mode'), 'summarized', false) . '>' . _x('Summarized (default)', 'DiscountSync Module', 'raptus-wc-bexio-connector') .'</option>';
        echo '<option value="itemized" ' . selected(Options::get('bexio_discount_sync_mode'), 'itemized', false) . '>' . _x('Itemized', 'DiscountSync Module', 'raptus-wc-bexio-connector') .'</option>';
        echo '</select>';
        echo '<p class="description">';
        echo _x('Sets how discounts are synchronized.
                 Summarized means, that a discount position will be created (default).
                 Itemized means, that discounts will be calculated per position based on difference between regular price and price paid in checkout (value in percent).', 'DiscountSync Module', 'raptus-wc-bexio-connector');
        echo '</p>';
        ?>
        <script type="text/javascript">
            jQuery(document).ready( function($) {
                var disable_sync_mode_cb = function() {
                    if ($('#bexio_discount_sync_mode').val() != 'summarized') {
                        $('#bexio_account_discount').closest('tr').fadeOut();
                    } else {
                        $('#bexio_account_discount').closest('tr').fadeIn();
                    }
                };
                $('#bexio_discount_sync_mode').click(disable_sync_mode_cb);
                disable_sync_mode_cb();
            });
        </script>
        <?php
    }

    /**
     * Sanitize the additional settings that this modules provides.
     */
    public function sanitize($new_input, $input) {
        if (isset($input['bexio_discount_sync_mode']) && !empty($input['bexio_discount_sync_mode'])) {
            $new_input['bexio_discount_sync_mode'] = $input['bexio_discount_sync_mode'];
        }
        return $new_input;
    }

    /**
     * Inject the default values for the additional settings that this modules provides.
     */
    public function defaults($defaults) {
        $defaults['bexio_discount_sync_mode'] = 'summarized';
        return $defaults;
    }

    /**
     * Save the regular product price (with and without tax) in the order item meta data.
     * 
     * This is needed to calculate the correct percent discount at sync, regardless of
     * whether discounts are applied on the total (like default WooCommerce coupons),
     * or whether they are already applied on the subtotal (like some B2B plugins do),
     * which would lead to no difference between subtotal & total to calculate the discount.
     * */
    public function add_order_item_data($item_id, $item, $order_id) {
        $product = null;
        if (get_class($item) == 'WC_Order_Item_Product')  // exclude shipping and tax items
            $product = $item->get_product();

        if(is_null($product) || !$product) {
            return;
        }

        // Read price directly from database to prevent falsification by B2B functionalities
        $regular_price = get_post_meta($product->get_id(), '_regular_price', true);

        // Ensure wc_get_price_excluding_tax uses the correct price
        $args['price'] = $regular_price;
        $price_without_tax = wc_get_price_excluding_tax($product, $args);

        /*
        Do not use wc_get_price_including_tax() for price with tax as that uses NumberUtil::round(),
        thus falsifying the accurate tax needed to calculate discount correctly.
        */
        $tax_rates = WC_Tax::get_rates($product->get_tax_class());
        $taxes = WC_Tax::calc_tax($price_without_tax, $tax_rates, false);
        $tax_sum = array_sum($taxes);
        $price_with_tax = $price_without_tax + $tax_sum;

        // Add regular price to order item meta data to allow discount calculation at sync
        wc_add_order_item_meta($item_id, '_price_without_tax', $price_without_tax);
        wc_add_order_item_meta($item_id, '_price_with_tax', $price_with_tax);
    }

    /**
     * Hide the regular price item meta data from the order dashboard.
     * */
    public function hide_order_item_meta($itemmeta) {
        $itemmeta[] = '_price_without_tax';
        $itemmeta[] = '_price_with_tax';
        return $itemmeta;
    }

    /**
     * Calculate discount per position if mode 'itemized' was selected.
     * 
     * Uses the difference between the regular price and the total price of the item
     * in order to calculate the percent difference, which is then set as 'discount'
     * in the item data.
     * 
     * Uses the regular price to calculate difference from, to also work if the shop
     * applies discounts differently from how WooCommerce by default does it, to be
     * able to work more independently of plugins / configurations. The tax setting
     * is also respected in the calculation.
     * 
     * @param array $data The order item data.
     * @param WC_Order_Item_Product $item The order item object.
     * @param WC_Product $product The product object.
     * @param WC_Order $order The order object.
     * @return array The order item data.
     * */
    public function calculate_discount_per_position($data, $item, $product, $order) {
        $discount_mode = Options::get('bexio_discount_sync_mode');
        if($discount_mode == 'itemized') {
            $incl_tax = get_option('woocommerce_tax_display_cart') == 'incl';

            $item_id = $item->get_id();
            $regular_price = null;
            $total = null;

            // Get product prices for calculation (with or without tax)
            if ($incl_tax) {
                $regular_price = wc_get_order_item_meta($item_id, '_price_with_tax', true);
                $tax_data = $item['line_tax_data'];
                // Use line_tax_data because line_tax gets rounded by WooCommerce, leading to wrong calculation
                $line_tax = array_sum($tax_data['total']);
                $total = ($item['line_total'] + $line_tax) / $item['qty'];
            } else {
                $regular_price = wc_get_order_item_meta($item_id, '_price_without_tax', true);
                $total = $item['line_total'] / $item['qty'];
            }

            // Cast to float value to prevent problems with potential string values
            if (is_numeric($regular_price)) {
                $regular_price = floatval($regular_price);
            }

            if (!$regular_price > 0 || !is_numeric($total)) {
                /*
                If the regular price (without any reductions) or the total is not a numeric value
                (to catch null, false & ''), or the regular price is 0 (would cause division by 0),
                we can not reliably calculate the discount, so abort at this point.
                */
                return $data;
            }

            // Calculate discount in percent
            $discount = 100 - (100 / $regular_price * $total);
            if($discount && $discount > 0) {
                // Set correct price without any discounts (as bexio will subtract discount automatically)
                $data['price'] = $regular_price;
                // Set discount per item position (if greater than 0)
                $data['discount'] = $discount;
            }
        }
        return $data;
    }

    /**
     * Do not add a discount position if 'itemized' mode was selected for sync.
     * 
     * @param array The discount data.
     * @return array The discount data (or empty array if mode is 'itemized').
     * */
    public function disable_discount_position($discount) {
        $discount_mode = Options::get('bexio_discount_sync_mode');
        if($discount_mode == 'itemized') {
            return [];
        }
        return $discount;
    }

}
