<?php

namespace Raptus\WC\BexioConnector\Modules;

defined('ABSPATH') or die;

use Raptus\WC\BexioConnector\Options;
use Raptus\WC\BexioConnector\Data;

/**
 * Provides functionality to synchronize document numbers.
 * - Order numbers
 * - Invoice numbers
 * - Delivery numbers
 *
 * This feature can only be used, if the automatic document numbering is disabled in Bexio.
 */
class DocumentNumbers extends Module {

    /**
    * The field types.
     * @var array
     */
    private $documents;

    /**
    * The document types.
     * @var array
     */
    private $types;

    /**
     * Constructs a new instance of this class.
     * Registers all required filter and action hooks.
     */
    public function __construct() {
        parent::__construct('documentnumbers', __('Document numbers', 'raptus-wc-bexio-connector'));
    }

    /**
     * Load this module.
     * This function should only be invoked if this module is active.
     */
    protected function load() {
        if (!empty($this->get_documents())) {
            add_action('admin_init', array($this, 'settings'), 10);
            add_filter('raptus_wc_bexio_connector_sanitize_settings', array($this, 'sanitize'), 10, 2);
            add_filter('raptus_wc_bexio_connector_order_data', array($this, 'order_number'), 10, 2);
            add_filter('raptus_wc_bexio_connector_invoice_data', array($this, 'invoice_number'), 10, 2);
            // TODO: enable document numbers for deliveries once supported by Bexio API
            // add_filter('raptus_wc_bexio_connector_delivery_data', array($this, 'delivery_number'), 10, 2);
            add_filter('raptus_wc_bexio_connector_defaults', array($this, 'defaults'), 10, 2);
        }
    }

    /**
     * Inject the order document numbers.
     *
     * @param  array The order data.
     * @param  \WC_Order The WooCommerc order object.
     * @return array The modified order data.
     */
    public function order_number($data, $order) {
        foreach ($this->get_documents() as $document => $documentName) {
            $data["{$document}_nr"] = $this->get_document_number($document, $order);
        }
        return $data;
    }

    /**
     * Inject the invoice document number.
     *
     * @param  array The order data.
     * @param  \WC_Order The WooCommerc order object.
     * @return array The modified order data.
     */
    public function invoice_number($data, $order) {
        $data['invoice_nr'] = $this->get_document_number('invoice', $order);
        return $data;
    }

    /**
     * Inject the delivery document number.
     *
     * @param  array The order data.
     * @param  \WC_Order The WooCommerc order object.
     * @return array The modified order data.
     */
    public function delivery_number($data, $order) {
        $data['delivery_nr'] = $this->get_document_number('delivery', $order);
        return $data;
    }

    /**
     * Get the document number for the given document.
     */
    public function get_document_number($document, $order) {
        $documentNumber = $order->get_order_number();
        $base = "bexio_document_number_{$document}";
        $prefix  = Options::get($base . '_prefix');
        $postfix = Options::get($base . '_postfix');
        if ($prefix || $postfix) {
            $documentNumber =  $prefix . $documentNumber . $postfix;
        }
        return apply_filters('raptus_wc_bexio_connector_document_number', $documentNumber, $document, $order);
    }

    /**
     * Load the additional settings that this modules provides.
     */
    public function defaults($defaults) {
        foreach ($this->get_documents() as $document => $documentName) {
            foreach ($this->get_types() as $type => $typeName) {
                $settings_key = "bexio_document_number_{$document}_{$type}";
                $defaults[$settings_key] = '';
            }
        }
        return $defaults;
    }

    /**
     * Load the additional settings that this modules provides.
     */
    public function sanitize($new_input, $input) {
        foreach ($this->get_documents() as $document => $documentName) {
            foreach ($this->get_types() as $type => $typeName) {
                $settings_key = "bexio_document_number_{$document}_{$type}";
                if (isset($input[$settings_key]) && !empty($input[$settings_key])) {
                    $new_input[$settings_key] = sanitize_text_field($input[$settings_key]);
                }
            }
        }
        return $new_input;
    }

    /**
     * Load the additional settings that this modules provides.
     */
    public function settings() {
        add_settings_field(
            'bexio_document_number_header',
            __('Document', 'raptus-wc-bexio-connector'),
            array($this, 'render_document_number_header'),
            'raptus_wp_bexio_connector_options_admin_advanced',
            'raptus_wp_bexio_connector_options_settings_bexio_advanced'
        );

        foreach ($this->get_documents() as $document => $documentName) {
            add_settings_field("bexio_document_number_{$document}", $documentName,
                array($this, 'render_document_number_field'),
                'raptus_wp_bexio_connector_options_admin_advanced',
                'raptus_wp_bexio_connector_options_settings_bexio_advanced',
                array('class' => 'wcba-table-header', 'document' => $document));
        }

        add_settings_field('bexio_document_number_footer', '',
            array($this, 'render_document_number_footer'),
            'raptus_wp_bexio_connector_options_admin_advanced',
            'raptus_wp_bexio_connector_options_settings_bexio_advanced'
        );
    }

    /**
     * Render the table header.
     */
    public function render_document_number_header() {
        echo '<table class="wcba-settings-table" style="table-layout: auto;">';
        echo '<thead>';
        echo '<tr>';
        foreach ($this->get_types() as $type => $name) {
            echo '<th>' . $name . '</th>';
            if ($type == 'prefix') {
                echo '<th style="text-align: center; width: 18%;">&nbsp;</th>';
            }
        }
        echo '</tr>';
        echo '</thead>';
        echo '</table>';
    }

    /**
     * Render document number field.
     */
    public function render_document_number_field($args) {
        echo '<table class="wcba-settings-table" style="table-layout: auto;">';
        echo '<tbody>';
        echo '<tr>';
        $document = $args['document'];
        foreach ($this->get_types() as $type => $name) {
            echo '<td>';
            $settings_key = "bexio_document_number_{$document}_{$type}";
            echo '<input class="all-options" style="width: 100%;" type="text" value="' . Options::get($settings_key)
               . '" id="' . $settings_key . '" name="raptus_wp_bexio_connector_options[' . $settings_key . ']">';
            if ($type == 'prefix') {
                echo '<td style="text-align: center; width: 18%;"><code>{order_number}</code></td>';
            }
            echo '</td>';
        }
        echo '</tr>';
        echo '</tbody>';
        echo '</table>';
    }

    /**
    * Render the table footer.
    */
    public function render_document_number_footer() {
        echo '<p class="description">' . __('Specify a number format for the different Bexio documents. This number format will only be used if automatic document numbering is disabled in Bexio.', 'raptus-wc-bexio-connector') . '</p>';
    }

    /**
     * Get the description of this module.
     *
     * @return string the description of this module
     */
    public function get_description() {
        $description  = __('Activate this module if you wish to control Bexio document numbers for orders and invoices from within WooCommerce.', 'raptus-wc-bexio-connector');
        $url = \Raptus\WC\BexioConnector\BEXIO_BASE_URL;
        $hint = sprintf(__('Be sure to disbable the <a href="%s" target="_blank">automatic document numbering</a> in Bexio for the desired document types, otherwise this function will not work.', 'raptus-wc-bexio-connector'), $url . 'admin/softwareSetting');

        return sprintf('%s<br/><small class="hint">%s</small>', $description, $hint);
    }

    /**
     * Get a list of all field types.
     *
     * @return array The field types.
     */
    protected function get_types() {
        if (!$this->types) {
            $this->types = array(
                'prefix'  => __('Prefix', 'raptus-wc-bexio-connector'),
                'postfix' => __('Postfix', 'raptus-wc-bexio-connector'),
            );
        }
        return $this->types;
    }

    /**
     * Get a list of all document types.
     *
     * @return array The document types.
     */
    protected function get_documents() {
        if (!$this->documents) {
            $config = Data::get_document_settings();
            $config_key = 'use_automatic_enumeration';
            $this->documents = Data::get_sync_types();

            foreach($this->documents as $document => $documentName) {
                // TODO: enable document numbers for deliveries once supported by Bexio API
                if ($document == 'delivery')
                    unset($this->documents[$document]);

                if (!isset($config[$document]) || (isset($config[$document][$config_key]) && $config[$document][$config_key])) {
                    unset($this->documents[$document]);
                }
            }
        }
        return $this->documents;
    }
}