<?php

// Namespace declaration
namespace Raptus\WC\BexioConnector\Modules;

// Exit if accessed directly
defined('ABSPATH') or die();

use Raptus\WC\BexioConnector\Options;
use Raptus\WC\BexioConnector\Tasks;


/**
 * Extended state synchronization
 */
class ExtendedStateSync extends Module {


    private static $cache;


    /**
     * Constructs a new instance of this class.
     * Registers all required filter and action hooks.
     */
    public function __construct() {
        parent::__construct('extendedstatesync', __('Extended state synchronization', 'raptus-wc-bexio-connector'));
    }

    /**
     * Get the description of this module.
     *
     * @return string the description of this module
     */
    public function get_description() {
        return __('With this additional module the order status is synchronized with the invoice status in Bexio. Changes of the invoice status are transposed to the order status. For example if an invoice is marked as paid in Bexio, the order status in WooCommerce is marked as completed.', 'raptus-wc-bexio-connector');
    }

    /**
     * Load this module.
     * This function should only be invoked if this module is active.
     */
    protected function load() {
        add_filter('raptus_wc_bexio_connector_tasks', [$this, 'register_cron'], 10, 1);
        add_action('init', [$this, 'init']);
        if (Options::get('bexio_extendedstatesync_mode') == null )
            Options::set('bexio_extendedstatesync_mode', 'none');
    }

    /*
     * Initialize the filter to deactivate the payment sync rules in functions.php.
     * Remove sync_payment rules from settings and deactivate them
     * Hooks into WordPress init hook
     */
    public function init() {
        if (is_admin()) {
            $this->load_global_settings();
        }
        $var = apply_filters('raptus-wc-bexio-connector-deactivate-sync-payment-rules', true);
        if ($var && $this->is_extendedstate_sync_enabled()){
            add_filter('raptus-wc-bexio-connector-sync-payment',array($this, 'sync_payment'), 10, 1);
            add_action('admin_head',  array($this, 'add_CSS'));   
        }
    }

    /*
     * Deactivate module and task
     */
    public function deactivate() {
        parent::deactivate();
        $task = new \Raptus\WC\BexioConnector\Tasks\SyncInvoiceState();
        $task->deactivate();
    }

    /**
     * Register the additonal cron task to synchronize invoice state with order state periodically.
     */
    public function register_cron($tasks = array()) {
        $tasks[] = \Raptus\WC\BexioConnector\Tasks\SyncInvoiceState::class;
        return $tasks;
    }

    /*
     * Deactivate payment sync rule
     */
    public function sync_payment($can_sync, $order = null){
        return false;
    }
    
    /*
     * Hide payment sync rules
     */
    public function add_CSS(){
        ?>
            <style>
                #raptus_wp_bexio_connector_sync_rules .wcba-settings-table th:nth-child(4),
                #raptus_wp_bexio_connector_sync_rules .wcba-settings-table td:nth-child(4)
                    { display: none; }
            </style>
        <?php
    }

    /**
     * Set the extendedstate fetch mode for the given order item.
     */
    public function is_extendedstate_sync_enabled($is_enabled = true) {
        return $is_enabled && Options::get('bexio_extendedstatesync_mode') == 'invoicesync';
    }
  
    /**
     * Load the global settings for configuring the extended state sync.
     */
    protected function load_global_settings() {
        add_filter('raptus_wp_bexio_connector_register_setting/bexio_sync_order_states', '__return_false');
            add_action('raptus_wp_bexio_connector_after_settings/bexio_sync_order_states', [$this, 'settings'], 10, 2);
        add_filter('raptus_wc_bexio_connector_sanitize_settings', array($this, 'sanitize'), 10, 2);
        add_filter('raptus_wc_bexio_connector_defaults', array($this, 'defaults'), 10, 1);
    }

    /**
     * Load the additionally needed settings that this modules provides.
     */
    public function settings($page, $section) {
        add_settings_field(
            'bexio_extendedstatesync_mode',
            __('Extended State Synchronization', 'raptus-wc-bexio-connector'),
            array($this, 'settings_field_extendedsync_mode'),
            $page,
            $section
        );
    }

    /**
     * Get the extended state synchronization modes for the module Extended State Sync
     *
     * @return array The state synchronization modes.
     */
    public function get_extendedstatesync_mode() {
        if (!isset(static::$cache) || empty(static::$cache)) {
            static::$cache = array(
                'none' => array(
                    'name' => __('None', 'raptus-wc-bexio-connector'),
                    'desc' => __('No order state synchronization', 'raptus-wc-bexio-connector'),
                    'default' => true
                ), 
                'ordersync' => array(
                    'name' => __('Order state', 'raptus-wc-bexio-connector'),
                    'desc' => __('Synchronize order states', 'raptus-wc-bexio-connector')
                ), 
                'invoicesync' => array(
                    'name' => __('Invoice state', 'raptus-wc-bexio-connector'),
                    'desc' => __('Synchronize order states through invoice states', 'raptus-wc-bexio-connector')
                )
            );
        }
        return static::$cache;
    }

    /**
     * Render the settings field for selecting a state synchronization mode.
     */
    public function settings_field_extendedsync_mode() {
        echo '<select id="bexio_extendedstatesync_mode" name="raptus_wp_bexio_connector_options[bexio_extendedstatesync_mode]" class="regular-text">';
        foreach ($this->get_extendedstatesync_mode() as $id => $item) {
            printf('<option value="%s" %s>%s</option>',
                $id, selected(Options::get('bexio_extendedstatesync_mode'), $id, false), $item['name']
            );
        }

        echo '</select>';
        echo '<p class="description">';
        echo $this->get_extendedstatesync_description();
        echo  __('When "Invoice State" is selected the invoice state in Bexio is synchronized with the WooCommerce order, f.e. if you set the invoice in Bexio to paid, the order in WooCommerce set to completed. The synchronization is periodically repeated each hour.', 'raptus-wc-bexio-connector');
        echo '</p>';

    }


    /**
     * Inject the default values for the additional settings that this modules provides.
     *
     * @param  array $defaults The defaults.
     * @return array The modified defaults.
     */
    public function defaults($defaults) {
        foreach ( $this->get_extendedstatesync_mode() as $id => $item) {
            if (isset($item['default']) && $item['default']) {
                $defaults['bexio_extendedstatesync_mode'] = $id;
            }
        }
        return $defaults;
    }


    /**
     * Sanitize the additional settings that this modules provides.
     */
    public function sanitize($new_input, $input) {
        if (isset($input['bexio_extendedstatesync_mode'])) {
            $new_input['bexio_extendedstatesync_mode'] = $input['bexio_extendedstatesync_mode'];
        }
        return $new_input;
    }


    /**
     * Get the extended state sync mode description.
     *
     * @return string The localized descrtiption.
     */
    private function get_extendedstatesync_description() {
        $description = __('Select the extended state sync mode:', 'raptus-wc-bexio-connector') . '<br/>';
        foreach ($this->get_extendedstatesync_mode() as $id => $item) {
            $description .= sprintf('- <b>%s:</b> %s<br/>', $item['name'], $item['desc']);
        }
        return $description;
    }

}
