<?php

namespace Raptus\WC\BexioConnector\Modules;

defined('ABSPATH') or die;

use Raptus\WC\BexioConnector\Data;
use Raptus\WC\BexioConnector\Options;
use Raptus\WC\BexioConnector\Activatable;

/**
 * Abstract module implementation that provides basic common functinality.
 */
abstract class Module implements Activatable {

    /**
     * The id of this module.
     * @var string
     */
    private $id;

    /**
     * The name of this module.
     * @var string
     */
    private $name;

    /**
     * Status whether this module is activated or not.
     * @var boolean
     */
    protected $is_activated = false;

    /**
     * Constructs a new instance of this class.
     * Registers all required filter and action hooks.
     */
    public function __construct($id, $name) {
        $this->id = $id;
        $this->name = $name;
    }

    /**
     * Get the id of this module.
     *
     * @return string the id of this module
     */
    public function get_id() {
        return $this->id;
    }

    /**
     * Get the name of this module.
     *
     * @return string the name of this module
     */
    public function get_name() {
        return $this->name;
    }

    /**
     * Get the description of this module.
     *
     * @return string the description of this module
     */
    public function get_description() {
        return $this->name;
    }

    /**
     * Get the activated state of this module
     * 
     * @return boolean true if this module is activated, false otherwise.
     */
    public function get_activated() {
        return $this->is_activated;
    }

    /**
     * Test if this module is available.
     * This is the case if the part of the currently active plan, determined by the remote service.
     *
     * @return boolean true if this module is available, false otherwise.
     */
    public function available() {
        if (!Options::get('isauthorized') && Options::get('wizard') != 'done') {
            return false;
        }
        $status = Data::get_client_status();
        $modules = is_array($status) && isset($status['modules']) ? $status['modules'] : [];
        return in_array($this->get_id(), $modules);
    }

    /**
     * Test if this module is active.
     * This is the case if the module is available and enabled by means of module configuration.
     *
     * @return boolean true if this module is active, false otherwise.
     */
    public function active() {
        $modules = Options::get('modules');
        return $this->available() && is_array($modules) && in_array($this->get_id(), $modules);
    }

    /**
     * Activate this module.
     * This action will effectively load the module and all it's dependencies.
     * Nothing will be done if the module is not active.
     *
     * @return boolean true if this module has been activated, false otherwise.
     */
    public function activate() {
        if ($this->active() && !$this->is_activated) {
            $this->load();
            $this->is_activated = true;
        }
        return $this->is_activated;
    }

    /**
     * Deactivate this module.
     *
     * @return boolean `true` if this module has been deactivated, `false` otherwise.
     */
    public function deactivate() {
        // Do nothing by default
    }

    /**
     * Load this module.
     * This function should only be invoked if this module is active.
     */
    protected abstract function load();
}