<?php

namespace Raptus\WC\BexioConnector\Modules;

defined('ABSPATH') or die;

use Raptus\WC\BexioConnector\Options;
use Raptus\WC\BexioConnector\Data;
use Raptus\WC\BexioConnector\Settings;
use Raptus\WC\BexioConnector\Order;
use Raptus\WC\BexioConnector\Api;
use Raptus\WC\BexioConnector\Email;
use WC_Order;
use WC_Emails;

/**
 * Add bexio invoices as PDF attachments to WooCommerce emails.
 */
class PDFAttachments extends Module {

    /**
     * Constructs a new instance of this class.
     */
    public function __construct() {
        parent::__construct('pdfattachments', __('PDF Attachments', 'raptus-wc-bexio-connector'));
    }

    /**
     * Load this module.
     * This function should only be invoked if this module is active.
     */
    protected function load() {

        // Settings
        add_action('raptus_wc_bexio_connector_meta_boxes', array($this, 'add_meta_box'), 10, 1);
        add_action('admin_init', array($this, 'settings'), 10);
        add_filter('raptus_wc_bexio_connector_sanitize_settings', array($this, 'sanitize'), 10, 2);
        add_filter('raptus_wc_bexio_connector_defaults', array($this, 'defaults'), 10, 2);

        // Handle attachment
        add_filter('woocommerce_email_attachments', array($this, 'attachments'), 10, 3);
        add_filter('raptus_wp_bexio_connector_edit_link_invoice', array($this, 'order_column_invoice'), 10, 2);
    }

    /**
     * Get the description of this module.
     *
     * @return string the description of this module
     */
    public function get_description() {
        return __('Attach PDF invoices from Bexio to WooCommerce Emails that are sent to customers.', 'raptus-wc-bexio-connector');
    }

    /**
     * Add the bexio invoice attachments.
     *
     * @param  array $attachments The attachment data.
     * @param  string $type The email type.
     * @param  mixed $context The email context (product, order, etc).
     * @return array The email attachments array.
     */
    public function attachments($attachments, $type, $context) {
        if ($context instanceof WC_Order) {
            $order = $context;
            $invoice = $this->download_invoice($order);
            if ($this->can_send_invoice($type, $order)) {
                if (!is_array($attachments)) {
                    if (empty($attachments)) {
                        $attachments = array();
                    } else {
                        $attachments = array($attachments);
                    }
                }
                $attachments[] = $invoice;
            }
        }
        return $attachments;
    }

    /**
     * Download the PDF invoice from bexio.
     *
     * @param \WC_Order $order The order to download the invoice for.
     * @return string The file path to the pdf invoice or `false`.
     */
    private function download_invoice($order) {
        $order_id = method_exists($order, 'get_id') ? $order->get_id() : $order->id;
        $bexio_invoice_id = get_post_meta($order_id, '_bexio_invoice_id', true);
        if (!$bexio_invoice_id)
            return false;
        $bexio_invoice_url = get_post_meta($order_id, '_bexio_invoice_pdf_url', true);
        $upload_dir = wp_upload_dir();
        if ($bexio_invoice_url) {
            $bexio_invoice_path = str_replace($upload_dir['baseurl'], $upload_dir['basedir'], $bexio_invoice_url);
            if (file_exists($bexio_invoice_path))
                return $bexio_invoice_path;
        }
        $pdf = Api::instance()->get_invoice_pdf($bexio_invoice_id);
        if ($pdf) {
            $pdf_path = apply_filters('raptus_wp_bexio_connector_invoice_pdf_upload_directory', '/invoices/');
            $dest = $upload_dir['basedir'] . $pdf_path;
            if (!file_exists($dest)) {
                mkdir($dest);
            }
            $filename = apply_filters('raptus_wp_bexio_connector_invoice_pdf_file_name', $pdf['name'], $dest, $pdf);
            $output = $dest . $filename;
            $content = base64_decode($pdf['content']);
            $bytes = file_put_contents($output, $content);
            if ($bytes !== false) {
                $pdf_url = $upload_dir['baseurl'] . $pdf_path . $filename;
                update_post_meta($order_id, '_bexio_invoice_pdf_url', $pdf_url);
                return $output;
            }
        } else {
            $error = Api::instance()->get_error_message();
            Email::instance()->send_pdf_download_failed_email($order_id, $error);
            $message = 'WCBA - Download of invoice PDF failed for order ' . $order_id;
            error_log($message);
            error_log($error);
            Api::instance()->log_to_remote($message, 'error', array('error_message' => $error));
        }
        return false;
    }

    /**
     * Append a link to the downloaded PDF after the invoice in the bexio details.
     */
    public function order_column_invoice($html, $order_id) {
        $invoice_pdf_url = get_post_meta($order_id, '_bexio_invoice_pdf_url', true);
        if (!$invoice_pdf_url) {
            global $pagenow;
            $order = wc_get_order($order_id);
            $status = $order->get_status();
            /**
             * Auto download PDFs if
             * - status is not cancelled
             * - status is not pending
             * - status is not completed or we are on the order detail
             **/
            if ($status != 'cancelled' && $status != 'pending' && ($status != 'completed' || $pagenow == 'post.php' && $_REQUEST['post'] == $order_id)) {
                $this->download_invoice($order);
                $invoice_pdf_url = get_post_meta($order_id, '_bexio_invoice_pdf_url', true);
            }
        }
        if ($invoice_pdf_url) {
            $html .= ', <a href="' . $invoice_pdf_url . '" target="_blank">';
            $html .= __('Link to the PDF', 'raptus-wc-bexio-connector');
            $html .= '</a>';
        }
        return $html;
    }

    /**
     * Test if an invoice can be sent based on the configured rules.
     *
     * @param string $type The email type.
     * @param \WC_Order $order The order to send the invoice for.
     * @return boolean `true` if the invoice can be sent, `false` otherwise.
     */
    protected function can_send_invoice($type, $order) {
        $gateway_id = Order::instance()->get_payment_method($order);
        $id = $this->settings_id($gateway_id);
        $email = Options::get($id);
        $can_send = $email == $type;

        return apply_filters('raptus_wp_bexio_connector_can_send_invoice', $can_send, $type, $order);
    }

    /**
     * Inject the default values for the additional settings that this modules provides.
     */
    public function defaults($defaults) {
        foreach (Data::get_payment_gateways() as $gateway_id => $gateway) {
            $id = $this->settings_id($gateway_id);
            $defaults[$id] = 0;
        }
        return $defaults;
    }

    /**
     * Sanitize the additional settings that this modules provides.
     */
    public function sanitize($new_input, $input) {
        foreach (Data::get_payment_gateways() as $gateway_id => $gateway) {
            $id = $this->settings_id($gateway_id);
            if (isset($input[$id])) {
                $new_input[$id] = $input[$id];
            }
        }
        return $new_input;
    }

    /**
     * Register the additional settings meta boxe.
     */
    public function add_meta_box($page) {
        if (Options::get('isauthorized')) {
            add_meta_box('raptus_wp_bexio_connector_' . $this->get_id(),
                $this->get_name(),
                array($this, 'do_settings_section'),
                $page,
                'normal',
                'core'
            );
        }
    }

    /**
     * Render the additional settings sections.
     */
    public function do_settings_section() {
        do_settings_sections(Settings::SETTINGS_PAGE_ID . '_' . $this->get_id());
    }

    /**
     * Load the additional section and settings that this modules provides.
     */
    public function settings() {
        $settings_page_id = Settings::SETTINGS_PAGE_ID . '_' . $this->get_id();
        $settings_section_id = 'raptus_wp_bexio_connector_options_settings_' . $this->get_id();

        // Add section
        add_settings_section($settings_section_id, '',
            array($this, 'print_section_info'),
            $settings_page_id
        );

        // Add table header
        add_settings_field(
            $this->settings_id('header'),
            __('Payment method', 'raptus-wc-bexio-connector'),
            array($this, 'setting_header'),
            $settings_page_id,
            $settings_section_id
        );

        // Add settings field for each payment gateway
        foreach (Data::get_payment_gateways() as $gateway_id => $gateway) {
            $id = $this->settings_id($gateway_id);
            add_settings_field($id, $gateway->get_method_title(),
                array($this, 'setting_email'),
                $settings_page_id,
                $settings_section_id,
                array('class' => 'wcba-table-header', 'id' => $id)
            );
        }
    }

    /**
     * Print the section introduction.
     */
    public function print_section_info() {
        _e('Select the emails where you wish to attach a PDF invoice.', 'raptus-wc-bexio-connector');
        echo ' ' . sprintf(__('If a email is mising, make sure that is is enabled in the <a href="%s">WooCommerce email settings</a>.', 'raptus-wc-bexio-connector'),
            admin_url('admin.php') . '?page=wc-settings&tab=email');
    }

    /**
     * Print the table header.
     */
    public function setting_header() {
        echo '<strong>' . __('WooCommerce email', 'raptus-wc-bexio-connector') . '</strong>';
    }

    /**
     * Render the table email setting select field.
     */
    public function setting_email($args) {
        $selected = Options::get($args['id']);
        echo '<select id="' . $args['id'] . '" name="raptus_wp_bexio_connector_options[' . $args['id'] . ']" class="regular-text" class="regular-text">';
        printf('<option value="%s" %s>%s</option>', 0, selected($selected, 0, false), __('Don\'t send', 'raptus-wc-bexio-connector'));
        foreach ($this->emails() as $email => $title) {
            printf('<option value="%s" %s>%s</option>', $email, selected($selected, $email, false), $title);
        }
        echo "</select>";
    }

    /**
     * Get a list of all available WooCommerce emails that are currently active.
     * Including manual ones.
     */
    protected function emails() {
        $emails = WC_Emails::instance();
        $types = array();
        foreach($emails->get_emails() as $class => $instance) {
            if ($instance->is_enabled() || $instance->is_manual()) {
                $types[$instance->id] = $instance->title;
            }
        }
        return $types;
    }

    /**
     * Get the setting id for the given payment gateway.
     */
    private function settings_id($gateway_id) {
        return 'bexio_' . $this->get_id() . '_rule_' . $gateway_id;
    }
}