<?php

// Namespace declaration
namespace Raptus\WC\BexioConnector\Modules;

// Exit if accessed directly
defined('ABSPATH') or die();

use Raptus\WC\BexioConnector\Options;
use Raptus\WC\BexioConnector\Data;

/**
 * Product and product category data enhancement with bexio accounts.
 */
class ProductAccounts extends Module {

    /**
     * Constructs a new instance of this class.
     * Registers all required filter and action hooks.
     */
    public function __construct() {
        parent::__construct('productaccounts', __('Product accounts', 'raptus-wc-bexio-connector'));
    }

    /**
     * Get the description of this module.
     *
     * @return string the description of this module
     */
    public function get_description() {
        return __('This additional module allows you to set Bexio accounts on products and product categories for individual accounting.', 'raptus-wc-bexio-connector');
    }

    /**
     * Test if this module is active.
     * This is the case if the module is available and enabled by means of module configuration.
     *
     * @return boolean true if this module is active, false otherwise.
     */
    public function active() {
        // TODO check if the module was already active before (upgrade)
        return parent::active();
    }

    /**
     * Load this module.
     * This function should only be invoked if this module is active.
     */
    protected function load() {
        add_filter('raptus_wc_bexio_connector_order_item_data', array($this, 'set_account_id'), 10, 3);

        if (is_admin()) {
            // Display field on product pages
            add_action('woocommerce_product_options_general_product_data', array($this,'display_bexio_account_product'), 10);

            // Save field for products pages
            add_action('woocommerce_process_product_meta', array($this, 'save_bexio_account_product'));

            // Display field on product categories
            add_action('product_cat_add_form_fields', array($this, 'display_bexio_account_add_product_cat'), 10);
            add_action('product_cat_edit_form_fields', array($this, 'display_bexio_account_edit_product_cat'), 10, 1);

            // Save field for product categories
            add_action('edited_product_cat', array($this, 'save_bexio_account_product_cat'), 10, 1);
            add_action('create_product_cat', array($this, 'save_bexio_account_product_cat'), 10, 1);
        }

        add_action('admin_init', array($this, 'settings'), 10);
        add_filter('raptus_wc_bexio_connector_sanitize_settings', array($this, 'sanitize'), 10, 2);
        add_filter('raptus_wc_bexio_connector_defaults', array($this, 'defaults'), 10, 1);
        add_filter('raptus_wc_bexio_connector_order_data', array($this, 'order_data'), 10, 2);
    }

    /**
     * Display the custom text field for selecting a bexio account on product pages.
     */
    public function display_bexio_account_product() {
        woocommerce_wp_select(array(
            'id' => 'bexio_account_product',
            'label' => __('Bexio account', 'raptus-wc-bexio-connector'),
            'desc_tip' => true,
            'description' => __('Select a revenue account to be used when creating a product position within a Bexio order.
                     Revenue accounts defined directly on products take precendence over all other settings.
                     Leave empty to use the account defined on the product category of this product,
                     or the globally defined account setting.', 'raptus-wc-bexio-connector'),
            'options' => $this->get_account_options(),
        ));
    }

    /**
     * Display the extra taxonomy field for selecting a Bexio account,
     * when creating a new product category.
     */
    public function display_bexio_account_add_product_cat() {
        echo '<div class="form-field">';
        echo '<label for="bexio_account_product_cat">' . __('Bexio account', 'raptus-wc-bexio-connector') . '</label>';
        $this->print_bexio_account_select_field($term);
        echo '</div>';
    }

    /**
     * Display the extra taxonomy field for selecting a Bexio account,
     * when editing an existing product category.
     *
     * @param \WC_Term $term The current term instance.
     */
    public function display_bexio_account_edit_product_cat($term) {
        echo '<tr class="form-field">';
        echo '  <th scope="row" valign="top">';
        echo '    <label for="bexio_account_product_cat">' . __('Bexio account', 'raptus-wc-bexio-connector') . '</label>';
        echo '  </th>';
        echo '  <td>';
        $this->print_bexio_account_select_field($term);
        echo '  </td>';
        echo '</tr>';
    }

    /**
     * Display the extra taxonomy field for selecting a Bexio account.
     *
     * @param \WC_Term $term The current term instance. Optional.
     */
    public function print_bexio_account_select_field($term = null) {
        $selected = $term ? get_term_meta($term->term_id, 'bexio_account_product_cat', true) : 0;
        echo '<select id="bexio_account_product_cat" name="bexio_account_product_cat">';
        foreach ($this->get_account_options() as $id => $label) {
            printf('<option value="%s" %s>%s</option>',
                $id, selected($selected, $id, false), $label);
        }
        echo '</select>';
        echo '<p class="description">';
        echo __('Select a revenue account to be used when creating a product position within a Bexio order.
                 Revenue accounts defined directly on products take precendence over this setting.
                 Leave empty to use the globally defined account setting instead.', 'raptus-wc-bexio-connector');
        echo '</p>';
    }

    /**
     * Save the selected Bexio account of the current product category.
     *
     * @param int $term_id The current product category ID.
     */
    public function save_bexio_account_product_cat($term_id) {
        $account = filter_input(INPUT_POST, 'bexio_account_product_cat');
        update_term_meta($term_id, 'bexio_account_product_cat', $account);
    }

    /**
     *Save the selected Bexio account of the current product.
     *
     * @param int $post_id The current product ID.
     */
    public function save_bexio_account_product($post_id) {
        $product = wc_get_product($post_id);
        $account = filter_input(INPUT_POST, 'bexio_account_product');
        update_post_meta($post_id, 'bexio_account_product', $account);
    }

    /**
     * Set the account ID for the given order item.
     */
    public function set_account_id($data, $item, $product) {
        if ($account_id = $this->get_account_id($product)) {
            $data['account_id'] = $account_id;
        }
        return $data;
    }

    /**
     * Get the account ID for the given product.
     * The account ID can be set o product level, product category (parent) or global scope.
     *
     * The function will return
     * 1. the account set on the given product if available
     * 2. otherwise, the account from any of the product's categories if available
     * 3. otherwise, the account from any parent category if available (recursivley)
     * 4. finally, the globally defined account setting
     *
     * @param  \WC_Product $product The product to get the account id for.
     * @return int The acccording account id for the product.
     */
    public function get_account_id($product) {
        if ($product) {
            $product_id = method_exists($product, 'get_id') ? $product->get_id() : $product->id;

            if ($product->is_type('variation')) {
                $product_id = $product->get_parent_id();
            }

            // 1. Test if an account is set on the product
            if ($account = get_post_meta($product_id, 'bexio_account_product', true)) {
                return $account;
            }

            // 2. Test if an account is set on any of the product's categories
            $terms = get_the_terms($product_id, 'product_cat') ?: array();
            foreach ($terms as $term) {
                if ($account = get_term_meta($term->term_id, 'bexio_account_product_cat', true)) {
                    return $account;
                }
            }

            // 3. Test if an account is set on any of the products parent categories (recursivley)
            foreach ($terms as $term) {
                $parent_id = $term->parent;
                while ($parent_id != 0) {
                    if ($account = get_term_meta($parent_id, 'bexio_account_product_cat', true)) {
                        return $account;
                    }
                    $parent_term = get_term_by('id', $parent_id, 'product_cat');
                    $parent_id = $parent_term->parent;
                }
            }
        }

        // 4. Return globally set bexio account
        return Options::get('bexio_account_products');
    }

    /**
     * Get a list of bexio accounts, where the key is the ID and the value is the name.
     *
     * @return array The account option list.
     */
    private function get_account_options() {
        $accounts = array(0 => __('Select a Bexio account', 'raptus-wc-bexio-connector') . '...');
        foreach (Data::get_accounts() as $item) {
            $accounts[$item['id']] = $item['account_no'] . ': ' . $item['name'];
        }
        return $accounts;
    }

    /**
     * Fetch all orders that have been synchronized with Bexio.
     *
     * @return array An array containing the following fields for each order:
     *         `id` The id of the WooCommerce order/post
     *         `order_id` The id of the remote Bexio order
     */
    protected function get_synced_bexio_orders() {
        global $wpdb;
        return $wpdb->get_results(
            "SELECT post_id AS id, meta_value AS order_id FROM $wpdb->postmeta WHERE meta_key = '_bexio_order_id'",
        ARRAY_A);
    }

    /**
     * Add the additional settings that this modules provides.
     */
    public function settings() {
        add_settings_field(
            'bexio_account_use_product_account',
            __('Source of accounts on product positions', 'raptus-wc-bexio-connector'),
            array($this, 'bexio_account_use_product_account_callback'),
            'raptus_wp_bexio_connector_options_admin_advanced',
            'raptus_wp_bexio_connector_options_settings_bexio_advanced'
        );
    }

    /**
     * Callback for the additional settings that this modules provides.
     */
    public function bexio_account_use_product_account_callback() {
        echo '<select id="bexio_account_use_product_account" name="raptus_wp_bexio_connector_options[bexio_account_use_product_account]" class="regular-text">';
        echo '<option value="0" ' . selected(Options::get('bexio_account_use_product_account'), 0, false) . '>' . __('Use Bexio Automator plugin settings', 'raptus-wc-bexio-connector') .'</option>';
        echo '<option value="1" ' . selected(Options::get('bexio_account_use_product_account'), 1, false) . '>' . __('Use Bexio-product settings', 'raptus-wc-bexio-connector') .'</option>';
        echo '</select>';
        echo '<p class="description">';
        echo __('Normally the accounts configured in the Bexio Automator settings are used when creating product positions. Alternatively you can set this to use the account on the Bexio product for product positions.', 'raptus-wc-bexio-connector');
        echo '</p>';
    }

    /**
     * Sanitize the additional settings that this modules provides.
     */
    public function sanitize($new_input, $input) {
        if (isset($input['bexio_account_use_product_account'])) {
            $new_input['bexio_account_use_product_account'] = $input['bexio_account_use_product_account'];
        }
        return $new_input;
    }

    /**
     * Inject the default values for the additional settings that this modules provides.
     */
    public function defaults($defaults) {
        $defaults['bexio_account_use_product_account'] = 0;
        return $defaults;
    }

    /**
     * Set the usage of product accounts in the order data
     */
    public function order_data($order_data, $order) {
        $order_data['use_product_account'] = Options::get('bexio_account_use_product_account');
        return $order_data;
    }
}
