<?php

namespace Raptus\WC\BexioConnector\Modules;

defined('ABSPATH') or die;

use Raptus\WC\BexioConnector\Api;
use Raptus\WC\BexioConnector\Options;
use Raptus\WC\BexioConnector\Modules;
use Raptus\WC\BexioConnector\Tasks;
use Raptus\WC\BexioConnector\Modules\StockSync;
use Raptus\WC\BexioConnector\Utils;

/**
 * Provides functionality to import products from Bexio.
 */
class ProductImport extends Module {

    /**
     * Constructs a new instance of this class.
     */
    public function __construct() {
        parent::__construct('productimport', __('Product Import', 'raptus-wc-bexio-connector'));
    }

    /**
     * Load this module.
     * This function should only be invoked if this module is active.
     */
    protected function load() {
        add_action('admin_enqueue_scripts', [$this, 'admin_scripts'], 100);
        add_action('admin_init', [$this, 'init']);

        add_action('admin_init', [$this, 'settings'], 10);
        add_filter('raptus_wc_bexio_connector_sanitize_settings', [$this, 'sanitize'], 10, 2);
        add_filter('raptus_wc_bexio_connector_defaults', [$this, 'defaults'], 10, 1);

        add_filter('bulk_actions-edit-product', [$this, 'add_product_bulk_actions'], 10, 1);
        add_filter('handle_bulk_actions-edit-product', [$this, 'handle_product_bulk_actions'], 10, 3);
        add_action('admin_notices', [$this, 'check_product_sync_status'], 10);

        add_filter('raptus_wc_bexio_connector_tasks', [$this, 'register_cron'], 10, 1);
    }

    /**
     * Ensure scheduled task is removed when module is deactivated.
     * */
    public function deactivate() {
        parent::deactivate();
        $task = new \Raptus\WC\BexioConnector\Tasks\AsyncPrice();
        $task->deactivate();
    }

    /**
     * Get the description of this module.
     *
     * @return string the description of this module
     */
    public function get_description() {
        return __('Import product data from Bexio. New products are created as draft, existing products (sku already exists in WooCommerce) are updated.', 'raptus-wc-bexio-connector');
    }

    public function admin_scripts() {
        global $pagenow;
        if (($pagenow == 'edit.php') && ($_GET['post_type'] == 'product')) {
            wp_register_script('product_import', \Raptus\WC\BexioConnector\ASSETS_URL . 'js/productimport.js', ['jquery', 'woocommerce_admin'], \Raptus\WC\BexioConnector\PLUGIN_VERSION);
            wp_enqueue_script('product_import');

            $params = array(
                'strings' => array(
                    'import_bexio_products' => __('Import from Bexio', 'raptus-wc-bexio-connector'),
                ),
                'urls' => array(
                    'import_bexio_products' => esc_url_raw(admin_url('edit.php?post_type=product&import_products_from_bexio=1')),
                ),
            );

            wp_localize_script('product_import', 'product_import', $params);
        }
    }

    public function init() {
        if (isset($_GET['import_products_from_bexio'])) {
            // import product data from bexio
            $result = $this->import_products();

            // Remove the import query argument to prevent another import when reloading the page
            $redirect_url = remove_query_arg('import_products_from_bexio');

            // Add created / updated count for admin notice
            $redirect_url = add_query_arg(array(
                'product_import_complete' => 1,
                'products_updated' => $result['updated'],
                'products_created' => $result['created'],
                'products_failed' => $result['failed'],
            ), $redirect_url);

            wp_redirect($redirect_url);
            exit();
        }
    }

    /**
     * Add the additional settings that this modules provides.
     */
    public function settings() {
        add_settings_field(
            'bexio_sync_product_price',
            __('Synchronize product prices', 'raptus-wc-bexio-connector'),
            [$this, 'bexio_sync_product_price'],
            'raptus_wp_bexio_connector_options_admin_advanced',
            'raptus_wp_bexio_connector_options_settings_bexio_advanced'
        );
    }

    /**
     * Callback for the additional settings that this modules provides.
     */
    public function bexio_sync_product_price() {
        echo '<label>';
        echo '<input type="checkbox" id="bexio_sync_product_price" name="raptus_wp_bexio_connector_options[bexio_sync_product_price]" '
           . checked(Options::get('bexio_sync_product_price'), true, false) . '>';
        echo __('Synchronize product prices', 'raptus-wc-bexio-connector') . '</label>';
        echo '<p class="description">';
        echo __('Enable the product price synchronization if you wish to pull product price information from Bexio.
            If enabled, the product price from Bexio will be periodically fetched. By default, this occurs twice a day, but this can be adjusted via filters.', 'raptus-wc-bexio-connector');
        echo '</p>';
    }

    /**
     * Sanitize the additional settings that this modules provides.
     */
    public function sanitize($new_input, $input) {
        if (isset($input['bexio_sync_product_price'])) {
            $new_input['bexio_sync_product_price'] = $input['bexio_sync_product_price'] ? 1 : 0;
        } else {
            $new_input['bexio_sync_product_price'] = 0;
        }
        return $new_input;
    }

    /**
     * Inject the default values for the additional settings that this modules provides.
     */
    public function defaults($defaults) {
        $defaults['bexio_sync_product_price'] = false;
        return $defaults;
    }

    private function import_products($skus = array()) {
        error_log('WCBA - Beginning product-import from Bexio');
        $result = Api::instance()->get_products(implode(',', $skus));

        $stocksync_module_is_active = false;
        $count_updated = 0;
        $count_created = 0;
        $count_failed = 0;

        foreach (Modules::get() as $module) {
            if ($module->get_id() == 'stocksync' && $module->active()) {
                $stocksync_module_is_active = true;
            }
        }

        if (!empty($result)) {
            foreach ($result as $product) {
                try {
                    // get ID of WooCommerce Product
                    $product_id = $this->get_product_id_by_sku($product['sku']);

                    if ($product_id) {
                        // product with SKU already exists - update

                        // skip products in trash (do not update)
                        if(get_post_status($product_id) == 'trash')
                            continue;

                        // get WC_Product object
                        $wc_product = wc_get_product($product_id);

                        if (!apply_filters('raptus_wc_bexio_connector_import_product', true, $product, $wc_product))
                            continue;

                        if ($wc_product) {
                            $current_price = $wc_product->get_regular_price();

                            if($product['sale_price'] != $current_price) {
                                $wc_product->set_regular_price($product['sale_price']);
                            }

                            // allow further modification of product data before saving
                            $wc_product = apply_filters('raptus_wc_bexio_connector_import_product_before_save', $wc_product, $product);

                            $wc_product->save();
                            $count_updated++;
                        }

                        unset($wc_product);
                    } else {
                        // product with SKU does not yet exists - create new product as draft

                        if (!apply_filters('raptus_wc_bexio_connector_import_product', true, $product, null))
                            continue;

                        $post_id = wp_insert_post([
                            'post_type' => 'product',
                            'post_status' => 'draft',
                            'post_title' => $product['name'],
                            'post_content' => $product['description'],
                        ]);

                        $wc_product = wc_get_product($post_id);
                        $wc_product->set_sku($product['sku']);
                        $wc_product->set_regular_price($product['sale_price']);

                        // convert units as Bexio uses fixed units (mm & g) and WC allows different units each
                        $wc_product->set_width($this->convert_size($product['width']));
                        $wc_product->set_height($this->convert_size($product['height']));
                        $wc_product->set_weight($this->convert_weight($product['weight']));

                        if ($stocksync_module_is_active) {
                            $wc_product->set_manage_stock(true);
                            $stock_fetch_mode = Options::get('bexio_stock_fetch_mode');
                            switch ($stock_fetch_mode) {
                                case 'inactive':
                                    break;

                                case 'effective':
                                    $wc_product->set_stock($product['stock_nr']);
                                    break;

                                case 'reserved':
                                    $wc_product->set_stock($product['stock_reserved_nr']);
                                    break;

                                case 'picked':
                                    $wc_product->set_stock($product['stock_picked_nr']);
                                    break;

                                case 'unpicked':
                                    $wc_product->set_stock($product['stock_nr'] - $product['stock_picked_nr']);
                                    break;

                                case 'available':
                                    $wc_product->set_stock($product['stock_nr'] - $product['stock_reserved_nr'] - $product['stock_picked_nr']);
                                    break;

                                default:
                                    break;
                            }
                            $wc_product->set_low_stock_amount($product['stock_min_nr']);
                        } elseif (!Options::get('bexio_disable_stock_synchronization')) {
                            $wc_product->set_manage_stock(true);
                            $wc_product->set_stock($product['stock_nr']);
                            $wc_product->set_low_stock_amount($product['stock_min_nr']);
                        }

                        // allow further modification of product data before saving
                        $wc_product = apply_filters('raptus_wc_bexio_connector_create_product_before_save', $wc_product, $product);

                        $wc_product->save();
                        $count_created++;

                        unset($wc_product);
                    }
                } catch(\Exception $e) {
                    $count_failed++;
                    // Log the error
                    error_log(sprintf('Import of product with SKU %s failed. Error: %s', $product['sku'], $e->getMessage()));
                    // Proceed import with the next product
                    continue;
                }
            }

            error_log('WCBA - product import updated '.$count_updated.' product(s)');
            error_log('WCBA - product import created '.$count_created.' product(s)');
            error_log('WCBA - product import failed to import / skipped '.$count_failed.' product(s)');
        }

        return [
            'updated' => $count_updated,
            'created' => $count_created,
            'failed' => $count_failed
        ];
    }

    /**
     * Get the ID of a WooCommerce Product by its SKU.
     *
     * Use $wpdb instead of get_posts(), as get_posts() by default will ignore
     * post-types when using 'any', if the post type is not set to public (which some are).
     * Also do not depend on wc_get_product_types() as some types 'might' not
     * be included there. Therefore, directly query via $wpdb and only look for
     * '_sku' to find an existing product (regardless of type or status).
     *
     * @param $sku The sku of the product to search for.
     */
    private function get_product_id_by_sku($sku) {
        // get ID of product matching the SKU, including those in trash (to not create duplicates)
        global $wpdb;
        $products = $wpdb->get_col($wpdb->prepare(
            "SELECT post_id from {$wpdb->prefix}postmeta
            WHERE meta_key = '_sku'
            AND meta_value = %s", $sku
        ));
        return empty($products) ? null : $products[0];
    }

    /**
     * Convert the size from Bexio (in mm) to the corresponding size unit selected in WooCommerce.
     * Factors are based on values according to https://www.unitconverters.net/length/.
     * 
     * @param $size The size as set on the Bexio product.
     * @return The converted size
     * */
    private function convert_size($size) {
        if(!$size || $size == 0)
            return $size;

        $size_unit = get_option('woocommerce_dimension_unit');

        switch($size_unit) {
            case 'in':
                return $size * 0.0393700787;

            case 'yd':
                return $size * 0.0010936133;

            case 'm':
                return $size / 1000;

            case 'mm':
                return $size;

            case 'cm':
            default:
                return $size / 10;
        }
    }

    /**
     * Convert the weight from Bexio (in g) to the corresponding weight unit selected in WooCommerce.
     * Factors are based on values according to https://www.unitconverters.net/weight-and-mass/.
     * 
     * @param $weight The weight as set on the Bexio product.
     * */
    private function convert_weight($weight) {
        if(!$weight || $weight == 0)
            return $weight;

        $weight_unit = get_option('woocommerce_weight_unit');

        switch($weight_unit) {
            case 'oz':
                return $weight * 0.0352739619;

            case 'lbs':
                return $weight * 0.0022046226;

            case 'g':
                return $weight;

            case 'kg':
            default:
                return $weight / 1000;
        }
    }

    public function add_product_bulk_actions($actions) {
        $actions['product_bexio_sync'] = __('Synchronize Product with Bexio', 'raptus-wc-bexio-connector');
        $actions['product_price_bexio_sync'] = __('Synchronize Product price with Bexio', 'raptus-wc-bexio-connector');
        return $actions;
    }

    public function handle_product_bulk_actions($redirect_to, $doaction, $post_ids) {
        if ($doaction == 'product_bexio_sync' ) {
            $ids = implode(',', $post_ids);
            global $wpdb;
            $skus = $wpdb->get_col("SELECT pm.meta_value FROM $wpdb->posts AS p
                RIGHT JOIN $wpdb->postmeta AS pm ON p.ID = pm.post_id
                WHERE p.post_type IN ('product', 'product_variation')
                AND (pm.meta_key = '_sku' AND pm.meta_value != '')
                AND pm.post_id IN ($ids)
            ");
            $result = $this->import_products($skus);
            $redirect_to = add_query_arg(array(
                'product_import_complete' => 1,
                'products_updated' => $result['updated'],
                'products_created' => $result['created'],
                'products_failed' => $result['failed'],
            ), $redirect_to);
        } elseif ($doaction == 'product_price_bexio_sync') {
            $result = $this->sync_product_price($post_ids);
            $prefix = 'product_price_bexio_sync';
            $redirect_to = add_query_arg(array(
                "{$prefix}_complete" => $result['count_updated'],
                "{$prefix}_skipped" => $result['count_skipped'],
            ), $redirect_to);
        }
        return $redirect_to;
    }

    /**
     * Checks the Bexio synchronization status of the current order after a synchronization has been performed.
     *
     * - Prints a success message if the price is now synchronized.
     */
    public function check_product_sync_status($type = '', $msg = '') {
        if (isset($_GET['product_price_bexio_sync_complete'])) {
            if ($_GET['product_price_bexio_sync_complete'] == 0 && isset($_GET['product_price_bexio_sync_skipped']) && $_GET['product_price_bexio_sync_skipped'] == 0) {
                $type = 'error';
                $msg = __('The price synchronization failed, the price of no prouducts was updated.', 'raptus-wc-bexio-connector');
            }
            if ($_GET['product_price_bexio_sync_complete'] > 0) {
                $complete = $_GET['product_price_bexio_sync_complete'];
                $type = 'success';
                $msg = sprintf(_n('The price of %d product was successfully synchronized with Bexio.', 'The prices of %d products were successfully synchronized with Bexio.', $complete, 'raptus-wc-bexio-connector'), $complete);
            }
            if (isset($_GET['product_price_bexio_sync_skipped']) && $_GET['product_price_bexio_sync_skipped'] > 0) {
                $skipped = $_GET['product_price_bexio_sync_skipped'];
                if ($_GET['product_price_bexio_sync_complete'] > 0)
                    $msg .= '<br/>';
                $msg .= '<em>' . sprintf(_n('%d product price-update was skipped.', '%d product price-updates were skipped.', $skipped, 'raptus-wc-bexio-connector'), $skipped) . '</em>';
            }
            Utils::print_notice($type, $msg);
        }
        if (isset($_GET['product_import_complete'])) {
            $total = 0;
            $msg = _x('Import of products from Bexio finished:', 'Admin notice for product import', 'raptus-wc-bexio-connector');
            if ($_GET['products_updated'] > 0) {
                $updated = $_GET['products_updated'];
                $total += $updated;
                $msg .= '<br>'.sprintf(_nx('%d product was updated', '%d products were updated', $updated, 'Admin notice for product import', 'raptus-wc-bexio-connector'), $updated);
            }
            if ($_GET['products_created'] > 0) {
                $created = $_GET['products_created'];
                $total += $created;
                $msg .= '<br>'.sprintf(_nx('%d product was created', '%d products were created', $created, 'Admin notice for product import', 'raptus-wc-bexio-connector'), $created);
            }
            $type = $total > 0 ? 'success' : 'info';
            Utils::print_notice($type, $msg);
            if ($_GET['products_failed'] > 0) {
                $failed = $_GET['products_failed'];
                $failed_msg = sprintf(_nx('Import of Bexio products failed to import %d product.', 'Import of Bexio products failed to import %d products.', $failed, 'Admin notice for product import', 'raptus-wc-bexio-connector'), $failed);
                $failed_msg .= '<br>' . _x('This usually happens because there was an error while trying to update / create the product. Check your error log for more details.', 'Admin notice for product import', 'raptus-wc-bexio-connector');
                Utils::print_notice('warning', $failed_msg);
            }
        }
    }

    /**
     * Register the additonal cron task to synchronize product price periodically.
     */
    public function register_cron($tasks = array()) {
        $tasks[] = \Raptus\WC\BexioConnector\Tasks\AsyncPrice::class;
        return $tasks;
    }

    /**
     * Synchronize the price of the given products (or all, if none specified).
     */
    public function sync_product_price($product_ids = array()) {
        error_log('WCBA - Beginning import of product prices from Bexio');
        if($product_ids) {
            global $wpdb;
            $skus = $wpdb->get_col("SELECT meta_value FROM {$wpdb->postmeta} WHERE meta_key = '_sku' AND meta_value != '' AND post_id IN (".join(',', $product_ids).")");
        }
        $result = Api::instance()->get_prices($skus ?? null);

        if (!empty($result)) {
            $count_updated = 0;
            $count_skipped = 0;

            foreach ($result as $product) {
                $product_id = $this->get_product_id_by_sku($product['sku']);
                if ($product_id) {
                    // skip products in trash (do not update)
                    if(get_post_status($product_id) == 'trash')
                        continue;

                    // get WC_Product object
                    $wc_product = wc_get_product($product_id);

                    if ($wc_product) {
                        $current_price = $wc_product->get_regular_price();

                        if($product['price'] != $current_price && !empty($product['price'])) {
                            $wc_product->set_regular_price($product['price']);
                            $wc_product->save();
                            $count_updated++;
                        }
                    }

                    unset($wc_product);
                }
            }

            $count_skipped = count($result) - $count_updated;

            if ($count_updated)
                error_log('WCBA - updated product price of '.$count_updated.' product(s)');
            if ($count_skipped)
                error_log('WCBA - skipped price-update of '.$count_skipped.' product(s)');
        }

        return compact('count_updated', 'count_skipped');
    }

}
