<?php

// Namespace declaration
namespace Raptus\WC\BexioConnector\Modules;

// Exit if accessed directly
defined('ABSPATH') or die();

use Raptus\WC\BexioConnector\Options;
use Raptus\WC\BexioConnector\Data;
use Raptus\WC\BexioConnector\Settings;

/**
 * Stock synchronization extensions that adds extra features.
 */
class StockSync extends Module {

    /**
     * Constructs a new instance of this class.
     * Registers all required filter and action hooks.
     */
    public function __construct() {
        parent::__construct('stocksync', __('Extended stock synchronization', 'raptus-wc-bexio-connector'));
    }

    /**
     * Get the description of this module.
     *
     * @return string the description of this module
     */
    public function get_description() {
        return __('Extend the product stock synchronization with stock fetch modes, stock settings per prdoduct and stock synchronization via cron job.', 'raptus-wc-bexio-connector');
    }

    /**
     * Load this module.
     * This function should only be invoked if this module is active.
     */
    protected function load() {
        add_filter('raptus_wc_bexio_connector_tasks', array($this, 'register_cron'), 10, 1);

        // Disable stock sync if cron job is active
        add_filter('raptus_wc_bexio_connector_enable_stock_synchronization', array($this, 'is_stock_sync_enabled'), 10, 1);
        add_filter('raptus_wc_bexio_connector_stock_fetch_mode', array($this, 'get_fetch_mode_global'), 10, 1);
        add_filter('raptus_wc_bexio_connector_stock_products', array($this, 'set_fetch_mode_on_product'), 10, 2);

        if (is_admin()) {
            $this->load_global_settings();
            if ($this->is_stock_sync_enabled()) {
                $this->load_product_options();
            }
        }
    }

    /**
     * Ensure scheduled task is removed when module is deactivated.
     * */
    public function deactivate() {
        parent::deactivate();
        $task = new \Raptus\WC\BexioConnector\Tasks\AsyncStock();
        $task->deactivate();
    }


    /**
     * Register the additonal cron task to synchronize product stock periodically.
     */
    public function register_cron($tasks = array()) {
        $tasks[] = \Raptus\WC\BexioConnector\Tasks\AsyncStock::class;
        return $tasks;
    }

    /**
     * Set the stock fetch mode for the given order item.
     */
    public function is_stock_sync_enabled($is_enabled = true) {
        return $is_enabled && Options::get('bexio_stock_fetch_mode') != 'inactive';
    }

    /**
     * Set the stock fetch mode for the given order item.
     */
    public function get_fetch_mode_global($mode = null) {
        return Options::get('bexio_stock_fetch_mode');
    }

    /**
     * Set the stock fetch mode for the given list of product skus.
     * Remove products where the stock synchronization has been disabled.
     *
     * @param array $skus The product SKU list.
     * @param array The modified product list, optionally with fetch modes set.
     */
    public function set_fetch_mode_on_product($modes = array(), $skus = array()) {
        $default_fetch_mode = $this->get_fetch_mode_global();
        foreach ($skus as $i => $sku) {
            $product_id = wc_get_product_id_by_sku($sku);
            $product = wc_get_product($product_id);
            $fetch_mode = $this->get_fetch_mode_product($product);
            // Only inject product fetch mode if it differs from global setting to reduce request payload
            // The remote instance will anyway use the global setting by default
            if ($fetch_mode != $default_fetch_mode && $fetch_mode != 'inactive') {
                $modes[$sku] = $fetch_mode;
            }
        }
        return $modes;
    }

    /**
     * Load the global settings for configuring the product stock fetch mode.
     */
    protected function load_global_settings() {
        add_filter('raptus_wp_bexio_connector_register_setting/bexio_disable_stock_synchronization', '__return_false');
        add_action('raptus_wp_bexio_connector_after_settings/bexio_disable_stock_synchronization', [$this, 'settings'], 10, 2);

        add_filter('raptus_wc_bexio_connector_sanitize_settings', array($this, 'sanitize'), 10, 2);
        add_filter('raptus_wc_bexio_connector_defaults', array($this, 'defaults'), 10, 1);
    }

    /**
     * Load the additionally needed settings that this modules provides.
     */
    public function settings($page, $section) {
        add_settings_field(
            'bexio_stock_fetch_mode',
            __('Stock fetch mode', 'raptus-wc-bexio-connector'),
            array($this, 'settings_field_fetch_mode'),
            $page,
            $section
        );

        add_settings_field(
            'bexio_async_stock_sync',
            __('Stock synchronization mode', 'raptus-wc-bexio-connector'),
            array($this, 'settings_field_async_stock'),
            $page,
            $section
        );
    }

    /**
     * Render the settings field for selecting a product stock fetch mode.
     */
    public function settings_field_fetch_mode() {
        echo '<select id="bexio_stock_fetch_mode" name="raptus_wp_bexio_connector_options[bexio_stock_fetch_mode]" class="regular-text">';
        foreach (Data::get_stock_fetch_modes() as $id => $item) {
            printf('<option value="%s" %s>%s</option>',
                $id, selected(Options::get('bexio_stock_fetch_mode'), $id, false), $item['name']
            );
        }

        echo '</select>';
        echo '<p class="description">';
        echo $this->get_stock_fetch_mode_description();
        echo '<br/>' . __('You can override this global setting on each product or product category if necessary.', 'raptus-wc-bexio-connector');
        echo '</p>';


        ?>
        <script type="text/javascript">
            jQuery(document).ready( function($) {
                var disable_stock_settings_cb = function() {
                    if ($('#bexio_stock_fetch_mode').val() == 'inactive') {
                        $('#bexio_async_stock_sync').closest('tr').fadeOut();
                    } else {
                        $('#bexio_async_stock_sync').closest('tr').fadeIn();
                    }
                };
                $('#bexio_stock_fetch_mode').change(disable_stock_settings_cb);
                disable_stock_settings_cb();
            });
        </script>
        <?php
    }

    /**
     * Render the settings field for selecting stock sync mode (synchronous or asynchronous).
     */
    public function settings_field_async_stock() {
        echo '<select id="bexio_async_stock_sync" name="raptus_wp_bexio_connector_options[bexio_async_stock_sync]" class="regular-text">';
        foreach (Data::get_stock_synchronization_modes() as $item) {
            printf('<option value="%s" %s>%s</option>',
                $item['id'], selected(Options::get('bexio_async_stock_sync'), $item['id'], false), $item['name']
            );
        }
        echo '</select>';
        echo '<p class="description">';
        echo __('Select the stock synchronization mode.
                 <b>Regular</b> means that the stock is synchronized to bexio based on events.
                 <b>Asynchronous</b> means that stock is synchronized to bexio once per hour.', 'raptus-wc-bexio-connector');
        echo '</p>';
    }

    /**
     * Inject the default values for the additional settings that this modules provides.
     *
     * @param  array $defaults The defaults.
     * @return array The modified defaults.
     */
    public function defaults($defaults) {
        $defaults['bexio_stock_fetch_mode'] = Options::get('bexio_disable_stock_synchronization') ? 'inactive' : 'available';
        $defaults['bexio_async_stock_sync'] = 'regular';
        return $defaults;
    }

    /**
     * Sanitize the additional settings that this modules provides.
     */
    public function sanitize($new_input, $input) {
        if (isset($input['bexio_stock_fetch_mode'])) {
            $new_input['bexio_stock_fetch_mode'] = $input['bexio_stock_fetch_mode'];
        }
        if (isset($input['bexio_async_stock_sync'])) {
            $new_input['bexio_async_stock_sync'] = $input['bexio_async_stock_sync'];
        }
        return $new_input;
    }

    /**
     * Load the additional product options provided by this module.
     */
    protected function load_product_options() {
        // Display field on product pages
        add_action('woocommerce_product_options_general_product_data', array($this,'display_bexio_stock_fetch_mode_product'), 9);

        // Save field for products pages
        add_action('woocommerce_process_product_meta', array($this, 'save_bexio_stock_fetch_mode_product'));

        // Display field on product categories
        add_action('product_cat_add_form_fields', array($this, 'display_bexio_stock_fetch_mode_add_product_cat'), 9);
        add_action('product_cat_edit_form_fields', array($this, 'display_bexio_stock_fetch_mode_edit_product_cat'), 9, 1);

        // Save field for product categories
        add_action('edited_product_cat', array($this, 'save_bexio_stock_fetch_mode_product_cat'), 9, 1);
        add_action('create_product_cat', array($this, 'save_bexio_stock_fetch_mode_product_cat'), 9, 1);
    }

    /**
     * Display the custom text field for selecting a stock fetch mode on product pages.
     */
    public function display_bexio_stock_fetch_mode_product() {
        woocommerce_wp_select(array(
            'id' => 'bexio_stock_fetch_mode_product',
            'label' => __('Stock fetch mode', 'raptus-wc-bexio-connector'),
            'desc_tip' => true,
            'description' => $this->get_stock_fetch_mode_description(),
            'options' => $this->get_stock_fetch_modes(),
        ));
    }

    /**
     * Display the extra taxonomy field for selecting a stock fetch mode,
     * when creating a new product category.
     */
    public function display_bexio_stock_fetch_mode_add_product_cat() {
        echo '<div class="form-field">';
        echo '<label for="bexio_stock_fetch_mode_product_cat">' . __('Stock fetch mode', 'raptus-wc-bexio-connector') . '</label>';
        $this->print_bexio_stock_fetch_mode_select_field($term);
        echo '</div>';
    }

    /**
     * Display the extra taxonomy field for selecting a stock fetch mode,
     * when editing an existing product category.
     *
     * @param \WC_Term $term The current term instance.
     */
    public function display_bexio_stock_fetch_mode_edit_product_cat($term) {
        echo '<tr class="form-field">';
        echo '  <th scope="row" valign="top">';
        echo '    <label for="bexio_stock_fetch_mode_product_cat">' . __('Stock fetch mode', 'raptus-wc-bexio-connector') . '</label>';
        echo '  </th>';
        echo '  <td>';
        $this->print_bexio_stock_fetch_mode_select_field($term);
        echo '  </td>';
        echo '</tr>';
    }

    /**
     * Display the extra taxonomy field for selecting a stock fetch mode.
     *
     * @param \WC_Term $term The current term instance. Optional.
     */
    public function print_bexio_stock_fetch_mode_select_field($term = null) {
        $selected = $term ? get_term_meta($term->term_id, 'bexio_stock_fetch_mode_product_cat', true) : 0;
        echo '<select id="bexio_stock_fetch_mode_product_cat" name="bexio_stock_fetch_mode_product_cat">';
        foreach ($this->get_stock_fetch_modes() as $id => $label) {
            printf('<option value="%s" %s>%s</option>',
                $id, selected($selected, $id, false), $label);
        }
        echo '</select>';
        echo '<p class="description">';
        echo $this->get_stock_fetch_mode_description();
        echo '</p>';
    }

    /**
     * Save the selected stock fetch mode of the current product category.
     *
     * @param int $term_id The current product category ID.
     */
    public function save_bexio_stock_fetch_mode_product_cat($term_id) {
        $fetch_mode = filter_input(INPUT_POST, 'bexio_stock_fetch_mode_product_cat');
        update_term_meta($term_id, 'bexio_stock_fetch_mode_product_cat', $fetch_mode);
    }

    /**
     * Save the selected stock fetch mode of the current product.
     *
     * @param int $post_id The current product ID.
     */
    public function save_bexio_stock_fetch_mode_product($post_id) {
        $product = wc_get_product($post_id);
        $fetch_mode = filter_input(INPUT_POST, 'bexio_stock_fetch_mode_product');
        update_post_meta($post_id, 'bexio_stock_fetch_mode_product', $fetch_mode);
    }

    /**
     * Get the fetch mode for the given product.
     * The fetch mode can be set o product level, product category (parent) or global scope.
     *
     * The function will return
     * 1. the fetch mode set on the given product if available
     * 2. otherwise, the fetch mode from any of the product's categories if available
     * 3. otherwise, the fetch mode from any parent category if available (recursivley)
     * 4. finally, the globally defined fetch mode setting
     *
     * @param  \WC_Product $product The product to get the fetch mode id for.
     * @return int The acccording fetch mode id for the product.
     */
    public function get_fetch_mode_product($product) {
        if ($product) {
            $product_id = method_exists($product, 'get_id') ? $product->get_id() : $product->id;

            if ($product->is_type('variation')) {
                $product_id = $product->get_parent_id();
            }

            // 1. Test if a fetch mode is set on the product
            if ($fetch_mode = get_post_meta($product_id, 'bexio_stock_fetch_mode_product', true)) {
                return $fetch_mode;
            }

            // 2. Test if a fetch mode is set on any of the product's categories
            $terms = get_the_terms($product_id, 'product_cat') ?: array();
            foreach ($terms as $term) {
                if ($fetch_mode = get_term_meta($term->term_id, 'bexio_stock_fetch_mode_product_cat', true)) {
                    return $fetch_mode;
                }
            }

            // 3. Test if a fetch mode is set on any of the products parent categories (recursivley)
            foreach ($terms as $term) {
                $parent_id = $term->parent;
                while ($parent_id != 0) {
                    if ($fetch_mode = get_term_meta($parent_id, 'bexio_stock_fetch_mode_product_cat', true)) {
                        return $fetch_mode;
                    }
                    $parent_term = get_term_by('id', $parent_id, 'product_cat');
                    $parent_id = $parent_term->parent;
                }
            }
        }

        // 4. Return globally set stock fetch mode
        return Options::get('bexio_stock_fetch_mode');
    }

    /**
     * Get a list of fetch modes, where the key is the ID and the value is the name.
     *
     * @return array The account option list.
     */
    private function get_stock_fetch_modes() {
        $fetch_modes = array(0 => __('Select a stock fetch mode', 'raptus-wc-bexio-connector') . '&hellip;');
        foreach (Data::get_stock_fetch_modes() as $id => $item) {
            $fetch_modes[$id] = $item['name'];
        }
        return $fetch_modes;
    }

    /**
     * Get the stock fetch mode option description.
     *
     * @return string The localized descrtiption.
     */
    private function get_stock_fetch_mode_description() {
        $description = __('Select the stock fetch mode.', 'raptus-wc-bexio-connector') . '<br/>';
        foreach (Data::get_stock_fetch_modes() as $id => $item) {
            $description .= sprintf('- <b>%s:</b> %s<br/>', $item['name'], $item['desc']);
        }
        return $description;
    }
}
