<?php

// Namespace declaration
namespace Raptus\WC\BexioConnector\Tasks;

// Exit if accessed directly
defined('ABSPATH') or die();

use Raptus\WC\BexioConnector\Options;
use Raptus\WC\BexioConnector\Order;

/**
 * Synchronize WooCommerce orders with Bexio.
 */
class AsyncNewOrders extends Task {

    /**
     * The time at which this option was activated.
     * @var string
     */
    private $activation_time;

    /**
     * The name of the option to store the activation time.
     * @var string
     */
    private $activation_option = 'wcba_asyncorders_activation';

    /**
     * Creates a new cron task to synchronize 'new' orders every 5 minutes.
     *
     * As 'new orders' are considered orders with no '_bexio_order_id'.
     * Prioritizes sync of 'new' orders (every 5 minutes), instead of only 1x / hour.
     * After a 'new' order is synced, the regular sync task (1x / hour) will suffice for further sync.
     *
     * If the asynchronous synchronization is enabled, the task schedule is set up.
     * If it is set to regular, all prevously registered tasks are cleared.
     */
    public function __construct() {
        parent::__construct(
            'raptus_sync_new_bexio_orders',
            '5min',
            300 // 300s = 5 minutes
        );
    }

    /**
     * Test if this task is active.
     */
    public function active() {
        return apply_filters('raptus_wc_bexio_connector_seperate_increased_interval_sync_new_orders', Options::get('bexio_async_order_creation') == 'async');
    }

    /**
     * Initialize the task scheduling.
     */
    public function init() {
        parent::init();
        $this->activation();
    }

    /**
     * Create activation timestamp.
     * Only orders after this will be synced asynchronously.
     *
     * @return string The activation time.
     */
    private function activation() {
        if (!$this->activation_time) {
            $activation_time = get_option($this->activation_option);
            if (!$activation_time) {
                $activation_time = date('Y-m-d h:i:s');
                add_option($this->activation_option, $activation_time);
            }
            $this->activation_time = $activation_time;
        }
        return $this->activation_time;
    }

    /**
     * Exectute the cron task.
     */
    public function execute() {
        $this->sync_orders();
    }

    /**
     * Unschedule the cron task.
     */
    public function unschedule() {
        parent::unschedule();
        delete_option($this->activation_option);
    }

    /**
     * Transition all WooCommerce order states to the corresponding state
     * of the associated remote Bexio order.
     */
    protected function sync_orders() {
        $orders = $this->get_unsynced_new_orders();
        foreach ($orders as $order_id) {
            Order::instance()->sync_all($order_id);
        }
    }

    /**
     * Fetch all 'new' orders that have not yet had an order synchronized with Bexio.
     * Only regard orders that were made after the asynchronous option was selected
     * (do not synchronize previous orders without _bexio_order_id).
     *
     * @return array An array containing the following fields for each order:
     *         `id` The id of the WooCommerce order/post
     */
    protected function get_unsynced_new_orders() {
        return get_posts(apply_filters('raptus_wc_bexio_connector_unsynced_new_orders_query', [
            'post_type' => 'shop_order',
            'posts_per_page' => -1,
            'post_status' => array_keys(wc_get_order_statuses()),
            'fields' => 'ids',
            'date_query' => [[
                'column' => 'post_date_gmt',
                'after' => $this->activation(),
            ]],
            'meta_query' => [
                [
                    'key' => '_bexio_order_id',
                    'compare' => 'NOT EXISTS'
                ],
            ]
        ]));
    }
}