<?php

// Namespace declaration
namespace Raptus\WC\BexioConnector\Tasks;

// Exit if accessed directly
defined('ABSPATH') or die();

use Raptus\WC\BexioConnector\Options;
use Raptus\WC\BexioConnector\Api;
use Raptus\WC\BexioConnector\Modules;
use Raptus\WC\BexioConnector\Tasks;

/**
 * Synchronize Bexio invoice states back to WooCommerce orders.
 */
class SyncInvoiceState extends Task {

    /**
     * Creates a new cron task to synchronize invoice states with order states hourly.
     */
    public function __construct() {
         parent::__construct('raptus_sync_bexio_invoice_states');
    }

    /**
     * Test if this task is active.
     */
    public function active() {
        $active = false;
        $modules = Modules::get();
        if($modules['extendedstatesync']->active() && Options::get('bexio_extendedstatesync_mode') == 'invoicesync') {
            $active = true;
        }
        return $active;
    }

    /**
     * Execute the cron task.
     */
    public function execute() {
        $this->transition_order_states();
    }

    /**
     * Unschedule the cron task.
     */
    public function unschedule() {
        parent::unschedule();
    }

    /**
     * Transition all WooCommerce order states from the corresponding state
     * of the associated remote Bexio invoice.
     */
    protected function transition_order_states() {
        $orders = $this->get_synced_bexio_orders();
        $invoice_ids = array_column($orders, 'invoice_id');
        $order_states = Api::instance()->get_invoice_states($invoice_ids);
        $valid_wc_order_states = $this->get_valid_wc_order_states();

        foreach ($orders as $order_data) {
            if (!array_key_exists($order_data['invoice_id'], $order_states))
                continue;

            $status = $order_states[$order_data['invoice_id']];
            if (!in_array($status, $valid_wc_order_states))
                continue;

            $order = wc_get_order($order_data['id']);
            if (empty($order) || !$this->validate_order_status_transition($order, $status))
                continue;

            if (!apply_filters('raptus_wc_bexio_connector_transition_invoice_state', true, $order))
                continue;

            try {
                $order->update_status($status);
            } catch (\Exception $ex) {
                $message = 'Could not change the status of order #' . $order_data['id'] . ' to "' . $status . '"';
                error_log($message);
                error_log($ex->getMessage());
                API::instance()->log_to_remote($message, 'error', array('error_message' => $ex->getMessage()));
            }
        }
    }

    /**
     * Validate the specified order status transition.
     * The transition is invalid if the new status is equal to the current one.
     *
     * @param \WC_Order $order The order to be transitioned.
     * @param string $new_status The new order status.
     * @param boolean `true` if it is a valid status transition, `false` otherwise.
     */
    protected function validate_order_status_transition($order, $new_status) {
        $status = $order->get_status();
        $endstates = array('cancelled', 'completed', 'refunded');
        if ($status == $new_status ||
            (in_array($status, $endstates) && in_array($new_status, $endstates)))
        {
            return false;
        }
        return true;
    }

    /**
     * Get a list of all valid WooCommerce order states.
     *
     * @return array The list of all valid WooCommerce order states.
     */
    protected function get_valid_wc_order_states() {
        $wc_states = array();
        $prefix = 'wc-';
        foreach (array_keys(\wc_get_order_statuses()) as $state) {
            if (substr($state, 0, strlen($prefix)) == $prefix) {
                $wc_states[] = substr($state, strlen($prefix));
            }
        }
        return $wc_states;
    }

    /**
     * Fetch all orders that have been synchronized with Bexio.
     *
     * @return array An array containing the following fields for each order:
     *         `id` The id of the WooCommerce order/post
     *         `invoice_id` The id of the remote Bexio invoice
     */
    protected function get_synced_bexio_orders() {
        global $wpdb;
        return $wpdb->get_results(apply_filters('raptus_wc_bexio_connector_synced_bexio_invoices_query',
            "SELECT post_id AS id, meta_value AS invoice_id FROM $wpdb->postmeta WHERE meta_key = '_bexio_invoice_id'"
        ), ARRAY_A);
    }
}