<?php

// Namespace declaration
namespace Raptus\WC\BexioConnector\Tasks;

// Exit if accessed directly
defined('ABSPATH') or die();

use Raptus\WC\BexioConnector\Options;
use Raptus\WC\BexioConnector\Api;
use Raptus\WC\BexioConnector\Modules;
use Raptus\WC\BexioConnector\Tasks;

/**
 * Synchronize Bexio order states back to WooCommerce.
 */
class SyncOrderStates extends Task {

    /**
     * Creates a new cron task to synchronize order states hourly.
     * If the order state synchronization is enabled, the task schedule is setup.
     * If the order state synchronization is disabled, all prevously registered tasks are cleared.
     */
    public function __construct() {
        parent::__construct('raptus_sync_bexio_order_states');
    }

    /**
     * Test if this task is active.
     */
    public function active() {
        $active = false;
        if(Options::get('bexio_sync_order_states')) {
            $active = true;
        }
        $modules = Modules::get();
        if($modules['extendedstatesync']->get_activated() && Options::get('bexio_extendedstatesync_mode') == 'ordersync') {
            $active = true;
        }
        return $active;
    }

    /**
     * Execute the cron task.
     */
    public function execute() {
        $this->transition_order_states();
    }

    /**
     * Transition all WooCommerce order states to the corresponding state
     * of the associated remote Bexio order.
     */
    protected function transition_order_states() {
        $orders = $this->get_synced_bexio_orders();
        $order_ids = array_column($orders, 'order_id');
        $order_states = Api::instance()->get_order_states($order_ids);
        $valid_wc_order_states = $this->get_valid_wc_order_states();

        foreach ($orders as $order_data) {
            if (!array_key_exists($order_data['order_id'], $order_states))
                continue;

            $status = $order_states[$order_data['order_id']];
            if (!in_array($status, $valid_wc_order_states))
                continue;

            $order = wc_get_order($order_data['id']);
            if (empty($order) || !$this->validate_order_status_transition($order, $status))
                continue;

            if (!apply_filters('raptus_wc_bexio_connector_transition_order_state', true, $order))
                continue;

            try {
                $order->update_status($status);
            } catch (\Exception $ex) {
                $message = 'Could not change the status of order #' . $order_data['id'] . ' to "' . $status . '"';
                error_log($message);
                error_log($ex->getMessage());
                API::instance()->log_to_remote($message, 'error', array('error_message' => $ex->getMessage()));
            }
        }
    }

    /**
     * Validate the specified order status transition.
     * The transition is invalid if the new status is equal to the current one.
     *
     * Futhermore, transitioning from pending/processing to on-hold is prohibited.
     * There reason is that paid orders might still be marked as open in Bexio,
     * if the optional invoice creation and payment is disabled.
     *
     * Also, we ignore the remote stauts "processing" since this could otherwise
     * lead to an unforntunate series of events that mark unpaid invoices as paid.
     * Aussming we have a new unpaid order with status "pending" (e.g. via invoice),
     * the following would happend if we don't ignore this status:
     *
     * 1. Create remote order and invoice (unpaid)
     * 2. The created invoice marks the remote order as "processing"
     * 3. Sync order states back to WooCommerce and transition the order to "processing"
     * 4. The WooCommerce order is now marked as paid, thus we pay the remote invoice
     *
     * Therefore we ingore the status, which effectivley prevents this problem.
     * The only down-side is, that the status "processing" is not reflected in the shop.
     *
     * @param \WC_Order $order The order to be transitioned.
     * @param string $new_status The new order status.
     * @param boolean `true` if it is a valid status transition, `false` otherwise.
     */
    protected function validate_order_status_transition($order, $new_status) {
        $status = $order->get_status();
        $endstates = array('cancelled', 'completed', 'refunded');
        if ($status == $new_status || $new_status == 'processing') {
            return false;
        } else if (in_array($status, array('pending', 'processing')) && $new_status == 'on-hold') {
            return false;
        } else if (in_array($status, $endstates) && in_array($new_status, $endstates)) {
            return false;
        }
        return true;
    }

    /**
     * Get a list of all valid WooCommerce order states.
     *
     * @return array The list of all valid WooCommerce order states.
     */
    protected function get_valid_wc_order_states() {
        $wc_states = array();
        $prefix = 'wc-';
        foreach (array_keys(\wc_get_order_statuses()) as $state) {
            if (substr($state, 0, strlen($prefix)) == $prefix) {
                $wc_states[] = substr($state, strlen($prefix));
            }
        }
        return $wc_states;
    }

    /**
     * Fetch all orders that have been synchronized with Bexio.
     *
     * @return array An array containing the following fields for each order:
     *         `id` The id of the WooCommerce order/post
     *         `order_id` The id of the remote Bexio order
     */
    protected function get_synced_bexio_orders() {
        global $wpdb;
        return $wpdb->get_results(apply_filters('raptus_wc_bexio_connector_synced_bexio_orders_query',
            "SELECT post_id AS id, meta_value AS order_id FROM $wpdb->postmeta WHERE meta_key = '_bexio_order_id'"
        ), ARRAY_A);
    }
}