<?php

// Namespace declaration
namespace Raptus\WC\BexioConnector\Tasks;

// Exit if accessed directly
defined('ABSPATH') or die();

use Raptus\WC\BexioConnector\Activatable;

/**
 * Abstract task implementation for registering and scheduling recurring Tasks
 * using the WP Cron framework.
 */
abstract class Task implements Activatable {

    /**
     * The name of this task.
     * @var string
     */
    private $task_name;

    /**
     * The recurrence of this task.
     * @var string
     */
     private $recurrence;

    /**
     * The interval of this task.
     * @var int
     */
    private $interval;

    /**
     * Creates a new cron task to synchronize order states hourly.
     * If the asynchronous synchronization is enabled, the task schedule is set up.
     * If it is set to regular, all prevously registered tasks are cleared.
     */
    public function __construct($task_name, $recurrence = 'hourly', $interval = 3600) {
        $this->task_name = $task_name;
        $this->recurrence = $recurrence;
        $this->interval = $interval;
    }

    /**
     * Get the id of this task.
     *
     * @return string the id of this task
     */
    public function get_id() {
        return $this->task_name;
    }

    /**
     * Activate this task.
     *
     * @return boolean `true` if the task has been activated, `false` otherwise.
     */
    public function activate() {
        if (apply_filters('raptus_wc_bexio_connector_activate_task_'.$this->task_name, $this->active())) {
            $this->init();
            return true;
        } else {
            $this->unschedule();
        }
    }

    /**
     * Deactivate this task.
     * This is an alias for `Task::unschedule`.
     *
     * @return boolean `true` if the task has been deactivated, `false` otherwise.
     */
    public function deactivate() {
        return $this->unschedule();
    }

    /**
     * Initialize the task scheduling.
     */
    public function init() {
        add_action($this->task_name, array($this, 'execute'));
        add_action('init', array($this, 'schedule'));
    }

    /**
    * Schedule the task.
    */
    public function schedule() {
        if (!wp_next_scheduled($this->task_name)) {
            wp_schedule_event(time() + $this->interval, $this->recurrence, $this->task_name);
        }
    }

    /**
    * Unschedule the task.
    */
    public function unschedule() {
        if (wp_next_scheduled($this->task_name)) {
            wp_clear_scheduled_hook($this->task_name);
            return true;
        }
    }

    /**
     * Test if this task is active.
     * @return boolean
     */
    public abstract function active();

    /**
     * Exectute the cron task.
     */
    public abstract function execute();
}